/*
 * FreeSec: libcrypt for NetBSD
 *
 * Copyright (c) 1994 David Burren
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This is an original implementation of the DES and the crypt(3) interfaces
 * by David Burren <davidb@werj.com.au>.
 *
 * An excellent reference on the underlying algorithm (and related
 * algorithms) is:
 *
 *	B. Schneier, Applied Cryptography: protocols, algorithms,
 *	and source code in C, John Wiley & Sons, 1994.
 *
 * Note that in that book's description of DES the lookups for the initial,
 * pbox, and final permutations are inverted (this has been brought to the
 * attention of the author).  A list of errata for this book has been
 * posted to the sci.crypt newsgroup by the author and is available for FTP.
 *
 * ARCHITECTURE ASSUMPTIONS:
 *	It is assumed that the 8-byte arrays passed by reference can be
 *	addressed as arrays of uint32_t's (ie. the CPU is not picky about
 *	alignment).
 */

#include <stdint.h>
#include "libc.h"

static int i64c(int i)
{
	i &= 0x3f;
	if (i == 0)
		return '.';
	if (i == 1)
		return '/';
	if (i < 12)
		return ('0' - 2 + i);
	if (i < 38)
		return ('A' - 12 + i);
	return ('a' - 38 + i);
}

/* Generated by const_des_init() (removed) */

static const uint8_t final_perm[64] = {
	0x39, 0x31, 0x29, 0x21, 0x19, 0x11, 0x09, 0x01, 0x3b, 0x33, 0x2b, 0x23, 0x1b, 0x13, 0x0b, 0x03,
	0x3d, 0x35, 0x2d, 0x25, 0x1d, 0x15, 0x0d, 0x05, 0x3f, 0x37, 0x2f, 0x27, 0x1f, 0x17, 0x0f, 0x07,
	0x38, 0x30, 0x28, 0x20, 0x18, 0x10, 0x08, 0x00, 0x3a, 0x32, 0x2a, 0x22, 0x1a, 0x12, 0x0a, 0x02,
	0x3c, 0x34, 0x2c, 0x24, 0x1c, 0x14, 0x0c, 0x04, 0x3e, 0x36, 0x2e, 0x26, 0x1e, 0x16, 0x0e, 0x06,
};
static const uint8_t m_sbox[4][4096] = {
	{
		0xef, 0xe3, 0xe1, 0xed, 0xe8, 0xe4, 0xee, 0xe7,
		0xe6, 0xef, 0xeb, 0xe2, 0xe3, 0xe8, 0xe4, 0xee,
		0xe9, 0xec, 0xe7, 0xe0, 0xe2, 0xe1, 0xed, 0xea,
		0xec, 0xe6, 0xe0, 0xe9, 0xe5, 0xeb, 0xea, 0xe5,
		0xe0, 0xed, 0xee, 0xe8, 0xe7, 0xea, 0xeb, 0xe1,
		0xea, 0xe3, 0xe4, 0xef, 0xed, 0xe4, 0xe1, 0xe2,
		0xe5, 0xeb, 0xe8, 0xe6, 0xec, 0xe7, 0xe6, 0xec,
		0xe9, 0xe0, 0xe3, 0xe5, 0xe2, 0xee, 0xef, 0xe9,
		0x0f, 0x03, 0x01, 0x0d, 0x08, 0x04, 0x0e, 0x07,
		0x06, 0x0f, 0x0b, 0x02, 0x03, 0x08, 0x04, 0x0e,
		0x09, 0x0c, 0x07, 0x00, 0x02, 0x01, 0x0d, 0x0a,
		0x0c, 0x06, 0x00, 0x09, 0x05, 0x0b, 0x0a, 0x05,
		0x00, 0x0d, 0x0e, 0x08, 0x07, 0x0a, 0x0b, 0x01,
		0x0a, 0x03, 0x04, 0x0f, 0x0d, 0x04, 0x01, 0x02,
		0x05, 0x0b, 0x08, 0x06, 0x0c, 0x07, 0x06, 0x0c,
		0x09, 0x00, 0x03, 0x05, 0x02, 0x0e, 0x0f, 0x09,
		0x4f, 0x43, 0x41, 0x4d, 0x48, 0x44, 0x4e, 0x47,
		0x46, 0x4f, 0x4b, 0x42, 0x43, 0x48, 0x44, 0x4e,
		0x49, 0x4c, 0x47, 0x40, 0x42, 0x41, 0x4d, 0x4a,
		0x4c, 0x46, 0x40, 0x49, 0x45, 0x4b, 0x4a, 0x45,
		0x40, 0x4d, 0x4e, 0x48, 0x47, 0x4a, 0x4b, 0x41,
		0x4a, 0x43, 0x44, 0x4f, 0x4d, 0x44, 0x41, 0x42,
		0x45, 0x4b, 0x48, 0x46, 0x4c, 0x47, 0x46, 0x4c,
		0x49, 0x40, 0x43, 0x45, 0x42, 0x4e, 0x4f, 0x49,
		0xff, 0xf3, 0xf1, 0xfd, 0xf8, 0xf4, 0xfe, 0xf7,
		0xf6, 0xff, 0xfb, 0xf2, 0xf3, 0xf8, 0xf4, 0xfe,
		0xf9, 0xfc, 0xf7, 0xf0, 0xf2, 0xf1, 0xfd, 0xfa,
		0xfc, 0xf6, 0xf0, 0xf9, 0xf5, 0xfb, 0xfa, 0xf5,
		0xf0, 0xfd, 0xfe, 0xf8, 0xf7, 0xfa, 0xfb, 0xf1,
		0xfa, 0xf3, 0xf4, 0xff, 0xfd, 0xf4, 0xf1, 0xf2,
		0xf5, 0xfb, 0xf8, 0xf6, 0xfc, 0xf7, 0xf6, 0xfc,
		0xf9, 0xf0, 0xf3, 0xf5, 0xf2, 0xfe, 0xff, 0xf9,
		0xdf, 0xd3, 0xd1, 0xdd, 0xd8, 0xd4, 0xde, 0xd7,
		0xd6, 0xdf, 0xdb, 0xd2, 0xd3, 0xd8, 0xd4, 0xde,
		0xd9, 0xdc, 0xd7, 0xd0, 0xd2, 0xd1, 0xdd, 0xda,
		0xdc, 0xd6, 0xd0, 0xd9, 0xd5, 0xdb, 0xda, 0xd5,
		0xd0, 0xdd, 0xde, 0xd8, 0xd7, 0xda, 0xdb, 0xd1,
		0xda, 0xd3, 0xd4, 0xdf, 0xdd, 0xd4, 0xd1, 0xd2,
		0xd5, 0xdb, 0xd8, 0xd6, 0xdc, 0xd7, 0xd6, 0xdc,
		0xd9, 0xd0, 0xd3, 0xd5, 0xd2, 0xde, 0xdf, 0xd9,
		0x7f, 0x73, 0x71, 0x7d, 0x78, 0x74, 0x7e, 0x77,
		0x76, 0x7f, 0x7b, 0x72, 0x73, 0x78, 0x74, 0x7e,
		0x79, 0x7c, 0x77, 0x70, 0x72, 0x71, 0x7d, 0x7a,
		0x7c, 0x76, 0x70, 0x79, 0x75, 0x7b, 0x7a, 0x75,
		0x70, 0x7d, 0x7e, 0x78, 0x77, 0x7a, 0x7b, 0x71,
		0x7a, 0x73, 0x74, 0x7f, 0x7d, 0x74, 0x71, 0x72,
		0x75, 0x7b, 0x78, 0x76, 0x7c, 0x77, 0x76, 0x7c,
		0x79, 0x70, 0x73, 0x75, 0x72, 0x7e, 0x7f, 0x79,
		0x1f, 0x13, 0x11, 0x1d, 0x18, 0x14, 0x1e, 0x17,
		0x16, 0x1f, 0x1b, 0x12, 0x13, 0x18, 0x14, 0x1e,
		0x19, 0x1c, 0x17, 0x10, 0x12, 0x11, 0x1d, 0x1a,
		0x1c, 0x16, 0x10, 0x19, 0x15, 0x1b, 0x1a, 0x15,
		0x10, 0x1d, 0x1e, 0x18, 0x17, 0x1a, 0x1b, 0x11,
		0x1a, 0x13, 0x14, 0x1f, 0x1d, 0x14, 0x11, 0x12,
		0x15, 0x1b, 0x18, 0x16, 0x1c, 0x17, 0x16, 0x1c,
		0x19, 0x10, 0x13, 0x15, 0x12, 0x1e, 0x1f, 0x19,
		0x4f, 0x43, 0x41, 0x4d, 0x48, 0x44, 0x4e, 0x47,
		0x46, 0x4f, 0x4b, 0x42, 0x43, 0x48, 0x44, 0x4e,
		0x49, 0x4c, 0x47, 0x40, 0x42, 0x41, 0x4d, 0x4a,
		0x4c, 0x46, 0x40, 0x49, 0x45, 0x4b, 0x4a, 0x45,
		0x40, 0x4d, 0x4e, 0x48, 0x47, 0x4a, 0x4b, 0x41,
		0x4a, 0x43, 0x44, 0x4f, 0x4d, 0x44, 0x41, 0x42,
		0x45, 0x4b, 0x48, 0x46, 0x4c, 0x47, 0x46, 0x4c,
		0x49, 0x40, 0x43, 0x45, 0x42, 0x4e, 0x4f, 0x49,
		0x2f, 0x23, 0x21, 0x2d, 0x28, 0x24, 0x2e, 0x27,
		0x26, 0x2f, 0x2b, 0x22, 0x23, 0x28, 0x24, 0x2e,
		0x29, 0x2c, 0x27, 0x20, 0x22, 0x21, 0x2d, 0x2a,
		0x2c, 0x26, 0x20, 0x29, 0x25, 0x2b, 0x2a, 0x25,
		0x20, 0x2d, 0x2e, 0x28, 0x27, 0x2a, 0x2b, 0x21,
		0x2a, 0x23, 0x24, 0x2f, 0x2d, 0x24, 0x21, 0x22,
		0x25, 0x2b, 0x28, 0x26, 0x2c, 0x27, 0x26, 0x2c,
		0x29, 0x20, 0x23, 0x25, 0x22, 0x2e, 0x2f, 0x29,
		0xef, 0xe3, 0xe1, 0xed, 0xe8, 0xe4, 0xee, 0xe7,
		0xe6, 0xef, 0xeb, 0xe2, 0xe3, 0xe8, 0xe4, 0xee,
		0xe9, 0xec, 0xe7, 0xe0, 0xe2, 0xe1, 0xed, 0xea,
		0xec, 0xe6, 0xe0, 0xe9, 0xe5, 0xeb, 0xea, 0xe5,
		0xe0, 0xed, 0xee, 0xe8, 0xe7, 0xea, 0xeb, 0xe1,
		0xea, 0xe3, 0xe4, 0xef, 0xed, 0xe4, 0xe1, 0xe2,
		0xe5, 0xeb, 0xe8, 0xe6, 0xec, 0xe7, 0xe6, 0xec,
		0xe9, 0xe0, 0xe3, 0xe5, 0xe2, 0xee, 0xef, 0xe9,
		0xff, 0xf3, 0xf1, 0xfd, 0xf8, 0xf4, 0xfe, 0xf7,
		0xf6, 0xff, 0xfb, 0xf2, 0xf3, 0xf8, 0xf4, 0xfe,
		0xf9, 0xfc, 0xf7, 0xf0, 0xf2, 0xf1, 0xfd, 0xfa,
		0xfc, 0xf6, 0xf0, 0xf9, 0xf5, 0xfb, 0xfa, 0xf5,
		0xf0, 0xfd, 0xfe, 0xf8, 0xf7, 0xfa, 0xfb, 0xf1,
		0xfa, 0xf3, 0xf4, 0xff, 0xfd, 0xf4, 0xf1, 0xf2,
		0xf5, 0xfb, 0xf8, 0xf6, 0xfc, 0xf7, 0xf6, 0xfc,
		0xf9, 0xf0, 0xf3, 0xf5, 0xf2, 0xfe, 0xff, 0xf9,
		0x2f, 0x23, 0x21, 0x2d, 0x28, 0x24, 0x2e, 0x27,
		0x26, 0x2f, 0x2b, 0x22, 0x23, 0x28, 0x24, 0x2e,
		0x29, 0x2c, 0x27, 0x20, 0x22, 0x21, 0x2d, 0x2a,
		0x2c, 0x26, 0x20, 0x29, 0x25, 0x2b, 0x2a, 0x25,
		0x20, 0x2d, 0x2e, 0x28, 0x27, 0x2a, 0x2b, 0x21,
		0x2a, 0x23, 0x24, 0x2f, 0x2d, 0x24, 0x21, 0x22,
		0x25, 0x2b, 0x28, 0x26, 0x2c, 0x27, 0x26, 0x2c,
		0x29, 0x20, 0x23, 0x25, 0x22, 0x2e, 0x2f, 0x29,
		0xbf, 0xb3, 0xb1, 0xbd, 0xb8, 0xb4, 0xbe, 0xb7,
		0xb6, 0xbf, 0xbb, 0xb2, 0xb3, 0xb8, 0xb4, 0xbe,
		0xb9, 0xbc, 0xb7, 0xb0, 0xb2, 0xb1, 0xbd, 0xba,
		0xbc, 0xb6, 0xb0, 0xb9, 0xb5, 0xbb, 0xba, 0xb5,
		0xb0, 0xbd, 0xbe, 0xb8, 0xb7, 0xba, 0xbb, 0xb1,
		0xba, 0xb3, 0xb4, 0xbf, 0xbd, 0xb4, 0xb1, 0xb2,
		0xb5, 0xbb, 0xb8, 0xb6, 0xbc, 0xb7, 0xb6, 0xbc,
		0xb9, 0xb0, 0xb3, 0xb5, 0xb2, 0xbe, 0xbf, 0xb9,
		0xdf, 0xd3, 0xd1, 0xdd, 0xd8, 0xd4, 0xde, 0xd7,
		0xd6, 0xdf, 0xdb, 0xd2, 0xd3, 0xd8, 0xd4, 0xde,
		0xd9, 0xdc, 0xd7, 0xd0, 0xd2, 0xd1, 0xdd, 0xda,
		0xdc, 0xd6, 0xd0, 0xd9, 0xd5, 0xdb, 0xda, 0xd5,
		0xd0, 0xdd, 0xde, 0xd8, 0xd7, 0xda, 0xdb, 0xd1,
		0xda, 0xd3, 0xd4, 0xdf, 0xdd, 0xd4, 0xd1, 0xd2,
		0xd5, 0xdb, 0xd8, 0xd6, 0xdc, 0xd7, 0xd6, 0xdc,
		0xd9, 0xd0, 0xd3, 0xd5, 0xd2, 0xde, 0xdf, 0xd9,
		0x8f, 0x83, 0x81, 0x8d, 0x88, 0x84, 0x8e, 0x87,
		0x86, 0x8f, 0x8b, 0x82, 0x83, 0x88, 0x84, 0x8e,
		0x89, 0x8c, 0x87, 0x80, 0x82, 0x81, 0x8d, 0x8a,
		0x8c, 0x86, 0x80, 0x89, 0x85, 0x8b, 0x8a, 0x85,
		0x80, 0x8d, 0x8e, 0x88, 0x87, 0x8a, 0x8b, 0x81,
		0x8a, 0x83, 0x84, 0x8f, 0x8d, 0x84, 0x81, 0x82,
		0x85, 0x8b, 0x88, 0x86, 0x8c, 0x87, 0x86, 0x8c,
		0x89, 0x80, 0x83, 0x85, 0x82, 0x8e, 0x8f, 0x89,
		0x1f, 0x13, 0x11, 0x1d, 0x18, 0x14, 0x1e, 0x17,
		0x16, 0x1f, 0x1b, 0x12, 0x13, 0x18, 0x14, 0x1e,
		0x19, 0x1c, 0x17, 0x10, 0x12, 0x11, 0x1d, 0x1a,
		0x1c, 0x16, 0x10, 0x19, 0x15, 0x1b, 0x1a, 0x15,
		0x10, 0x1d, 0x1e, 0x18, 0x17, 0x1a, 0x1b, 0x11,
		0x1a, 0x13, 0x14, 0x1f, 0x1d, 0x14, 0x11, 0x12,
		0x15, 0x1b, 0x18, 0x16, 0x1c, 0x17, 0x16, 0x1c,
		0x19, 0x10, 0x13, 0x15, 0x12, 0x1e, 0x1f, 0x19,
		0x3f, 0x33, 0x31, 0x3d, 0x38, 0x34, 0x3e, 0x37,
		0x36, 0x3f, 0x3b, 0x32, 0x33, 0x38, 0x34, 0x3e,
		0x39, 0x3c, 0x37, 0x30, 0x32, 0x31, 0x3d, 0x3a,
		0x3c, 0x36, 0x30, 0x39, 0x35, 0x3b, 0x3a, 0x35,
		0x30, 0x3d, 0x3e, 0x38, 0x37, 0x3a, 0x3b, 0x31,
		0x3a, 0x33, 0x34, 0x3f, 0x3d, 0x34, 0x31, 0x32,
		0x35, 0x3b, 0x38, 0x36, 0x3c, 0x37, 0x36, 0x3c,
		0x39, 0x30, 0x33, 0x35, 0x32, 0x3e, 0x3f, 0x39,
		0xaf, 0xa3, 0xa1, 0xad, 0xa8, 0xa4, 0xae, 0xa7,
		0xa6, 0xaf, 0xab, 0xa2, 0xa3, 0xa8, 0xa4, 0xae,
		0xa9, 0xac, 0xa7, 0xa0, 0xa2, 0xa1, 0xad, 0xaa,
		0xac, 0xa6, 0xa0, 0xa9, 0xa5, 0xab, 0xaa, 0xa5,
		0xa0, 0xad, 0xae, 0xa8, 0xa7, 0xaa, 0xab, 0xa1,
		0xaa, 0xa3, 0xa4, 0xaf, 0xad, 0xa4, 0xa1, 0xa2,
		0xa5, 0xab, 0xa8, 0xa6, 0xac, 0xa7, 0xa6, 0xac,
		0xa9, 0xa0, 0xa3, 0xa5, 0xa2, 0xae, 0xaf, 0xa9,
		0xaf, 0xa3, 0xa1, 0xad, 0xa8, 0xa4, 0xae, 0xa7,
		0xa6, 0xaf, 0xab, 0xa2, 0xa3, 0xa8, 0xa4, 0xae,
		0xa9, 0xac, 0xa7, 0xa0, 0xa2, 0xa1, 0xad, 0xaa,
		0xac, 0xa6, 0xa0, 0xa9, 0xa5, 0xab, 0xaa, 0xa5,
		0xa0, 0xad, 0xae, 0xa8, 0xa7, 0xaa, 0xab, 0xa1,
		0xaa, 0xa3, 0xa4, 0xaf, 0xad, 0xa4, 0xa1, 0xa2,
		0xa5, 0xab, 0xa8, 0xa6, 0xac, 0xa7, 0xa6, 0xac,
		0xa9, 0xa0, 0xa3, 0xa5, 0xa2, 0xae, 0xaf, 0xa9,
		0x6f, 0x63, 0x61, 0x6d, 0x68, 0x64, 0x6e, 0x67,
		0x66, 0x6f, 0x6b, 0x62, 0x63, 0x68, 0x64, 0x6e,
		0x69, 0x6c, 0x67, 0x60, 0x62, 0x61, 0x6d, 0x6a,
		0x6c, 0x66, 0x60, 0x69, 0x65, 0x6b, 0x6a, 0x65,
		0x60, 0x6d, 0x6e, 0x68, 0x67, 0x6a, 0x6b, 0x61,
		0x6a, 0x63, 0x64, 0x6f, 0x6d, 0x64, 0x61, 0x62,
		0x65, 0x6b, 0x68, 0x66, 0x6c, 0x67, 0x66, 0x6c,
		0x69, 0x60, 0x63, 0x65, 0x62, 0x6e, 0x6f, 0x69,
		0x6f, 0x63, 0x61, 0x6d, 0x68, 0x64, 0x6e, 0x67,
		0x66, 0x6f, 0x6b, 0x62, 0x63, 0x68, 0x64, 0x6e,
		0x69, 0x6c, 0x67, 0x60, 0x62, 0x61, 0x6d, 0x6a,
		0x6c, 0x66, 0x60, 0x69, 0x65, 0x6b, 0x6a, 0x65,
		0x60, 0x6d, 0x6e, 0x68, 0x67, 0x6a, 0x6b, 0x61,
		0x6a, 0x63, 0x64, 0x6f, 0x6d, 0x64, 0x61, 0x62,
		0x65, 0x6b, 0x68, 0x66, 0x6c, 0x67, 0x66, 0x6c,
		0x69, 0x60, 0x63, 0x65, 0x62, 0x6e, 0x6f, 0x69,
		0xcf, 0xc3, 0xc1, 0xcd, 0xc8, 0xc4, 0xce, 0xc7,
		0xc6, 0xcf, 0xcb, 0xc2, 0xc3, 0xc8, 0xc4, 0xce,
		0xc9, 0xcc, 0xc7, 0xc0, 0xc2, 0xc1, 0xcd, 0xca,
		0xcc, 0xc6, 0xc0, 0xc9, 0xc5, 0xcb, 0xca, 0xc5,
		0xc0, 0xcd, 0xce, 0xc8, 0xc7, 0xca, 0xcb, 0xc1,
		0xca, 0xc3, 0xc4, 0xcf, 0xcd, 0xc4, 0xc1, 0xc2,
		0xc5, 0xcb, 0xc8, 0xc6, 0xcc, 0xc7, 0xc6, 0xcc,
		0xc9, 0xc0, 0xc3, 0xc5, 0xc2, 0xce, 0xcf, 0xc9,
		0xcf, 0xc3, 0xc1, 0xcd, 0xc8, 0xc4, 0xce, 0xc7,
		0xc6, 0xcf, 0xcb, 0xc2, 0xc3, 0xc8, 0xc4, 0xce,
		0xc9, 0xcc, 0xc7, 0xc0, 0xc2, 0xc1, 0xcd, 0xca,
		0xcc, 0xc6, 0xc0, 0xc9, 0xc5, 0xcb, 0xca, 0xc5,
		0xc0, 0xcd, 0xce, 0xc8, 0xc7, 0xca, 0xcb, 0xc1,
		0xca, 0xc3, 0xc4, 0xcf, 0xcd, 0xc4, 0xc1, 0xc2,
		0xc5, 0xcb, 0xc8, 0xc6, 0xcc, 0xc7, 0xc6, 0xcc,
		0xc9, 0xc0, 0xc3, 0xc5, 0xc2, 0xce, 0xcf, 0xc9,
		0xbf, 0xb3, 0xb1, 0xbd, 0xb8, 0xb4, 0xbe, 0xb7,
		0xb6, 0xbf, 0xbb, 0xb2, 0xb3, 0xb8, 0xb4, 0xbe,
		0xb9, 0xbc, 0xb7, 0xb0, 0xb2, 0xb1, 0xbd, 0xba,
		0xbc, 0xb6, 0xb0, 0xb9, 0xb5, 0xbb, 0xba, 0xb5,
		0xb0, 0xbd, 0xbe, 0xb8, 0xb7, 0xba, 0xbb, 0xb1,
		0xba, 0xb3, 0xb4, 0xbf, 0xbd, 0xb4, 0xb1, 0xb2,
		0xb5, 0xbb, 0xb8, 0xb6, 0xbc, 0xb7, 0xb6, 0xbc,
		0xb9, 0xb0, 0xb3, 0xb5, 0xb2, 0xbe, 0xbf, 0xb9,
		0x5f, 0x53, 0x51, 0x5d, 0x58, 0x54, 0x5e, 0x57,
		0x56, 0x5f, 0x5b, 0x52, 0x53, 0x58, 0x54, 0x5e,
		0x59, 0x5c, 0x57, 0x50, 0x52, 0x51, 0x5d, 0x5a,
		0x5c, 0x56, 0x50, 0x59, 0x55, 0x5b, 0x5a, 0x55,
		0x50, 0x5d, 0x5e, 0x58, 0x57, 0x5a, 0x5b, 0x51,
		0x5a, 0x53, 0x54, 0x5f, 0x5d, 0x54, 0x51, 0x52,
		0x55, 0x5b, 0x58, 0x56, 0x5c, 0x57, 0x56, 0x5c,
		0x59, 0x50, 0x53, 0x55, 0x52, 0x5e, 0x5f, 0x59,
		0x9f, 0x93, 0x91, 0x9d, 0x98, 0x94, 0x9e, 0x97,
		0x96, 0x9f, 0x9b, 0x92, 0x93, 0x98, 0x94, 0x9e,
		0x99, 0x9c, 0x97, 0x90, 0x92, 0x91, 0x9d, 0x9a,
		0x9c, 0x96, 0x90, 0x99, 0x95, 0x9b, 0x9a, 0x95,
		0x90, 0x9d, 0x9e, 0x98, 0x97, 0x9a, 0x9b, 0x91,
		0x9a, 0x93, 0x94, 0x9f, 0x9d, 0x94, 0x91, 0x92,
		0x95, 0x9b, 0x98, 0x96, 0x9c, 0x97, 0x96, 0x9c,
		0x99, 0x90, 0x93, 0x95, 0x92, 0x9e, 0x9f, 0x99,
		0x9f, 0x93, 0x91, 0x9d, 0x98, 0x94, 0x9e, 0x97,
		0x96, 0x9f, 0x9b, 0x92, 0x93, 0x98, 0x94, 0x9e,
		0x99, 0x9c, 0x97, 0x90, 0x92, 0x91, 0x9d, 0x9a,
		0x9c, 0x96, 0x90, 0x99, 0x95, 0x9b, 0x9a, 0x95,
		0x90, 0x9d, 0x9e, 0x98, 0x97, 0x9a, 0x9b, 0x91,
		0x9a, 0x93, 0x94, 0x9f, 0x9d, 0x94, 0x91, 0x92,
		0x95, 0x9b, 0x98, 0x96, 0x9c, 0x97, 0x96, 0x9c,
		0x99, 0x90, 0x93, 0x95, 0x92, 0x9e, 0x9f, 0x99,
		0x5f, 0x53, 0x51, 0x5d, 0x58, 0x54, 0x5e, 0x57,
		0x56, 0x5f, 0x5b, 0x52, 0x53, 0x58, 0x54, 0x5e,
		0x59, 0x5c, 0x57, 0x50, 0x52, 0x51, 0x5d, 0x5a,
		0x5c, 0x56, 0x50, 0x59, 0x55, 0x5b, 0x5a, 0x55,
		0x50, 0x5d, 0x5e, 0x58, 0x57, 0x5a, 0x5b, 0x51,
		0x5a, 0x53, 0x54, 0x5f, 0x5d, 0x54, 0x51, 0x52,
		0x55, 0x5b, 0x58, 0x56, 0x5c, 0x57, 0x56, 0x5c,
		0x59, 0x50, 0x53, 0x55, 0x52, 0x5e, 0x5f, 0x59,
		0x0f, 0x03, 0x01, 0x0d, 0x08, 0x04, 0x0e, 0x07,
		0x06, 0x0f, 0x0b, 0x02, 0x03, 0x08, 0x04, 0x0e,
		0x09, 0x0c, 0x07, 0x00, 0x02, 0x01, 0x0d, 0x0a,
		0x0c, 0x06, 0x00, 0x09, 0x05, 0x0b, 0x0a, 0x05,
		0x00, 0x0d, 0x0e, 0x08, 0x07, 0x0a, 0x0b, 0x01,
		0x0a, 0x03, 0x04, 0x0f, 0x0d, 0x04, 0x01, 0x02,
		0x05, 0x0b, 0x08, 0x06, 0x0c, 0x07, 0x06, 0x0c,
		0x09, 0x00, 0x03, 0x05, 0x02, 0x0e, 0x0f, 0x09,
		0x3f, 0x33, 0x31, 0x3d, 0x38, 0x34, 0x3e, 0x37,
		0x36, 0x3f, 0x3b, 0x32, 0x33, 0x38, 0x34, 0x3e,
		0x39, 0x3c, 0x37, 0x30, 0x32, 0x31, 0x3d, 0x3a,
		0x3c, 0x36, 0x30, 0x39, 0x35, 0x3b, 0x3a, 0x35,
		0x30, 0x3d, 0x3e, 0x38, 0x37, 0x3a, 0x3b, 0x31,
		0x3a, 0x33, 0x34, 0x3f, 0x3d, 0x34, 0x31, 0x32,
		0x35, 0x3b, 0x38, 0x36, 0x3c, 0x37, 0x36, 0x3c,
		0x39, 0x30, 0x33, 0x35, 0x32, 0x3e, 0x3f, 0x39,
		0x7f, 0x73, 0x71, 0x7d, 0x78, 0x74, 0x7e, 0x77,
		0x76, 0x7f, 0x7b, 0x72, 0x73, 0x78, 0x74, 0x7e,
		0x79, 0x7c, 0x77, 0x70, 0x72, 0x71, 0x7d, 0x7a,
		0x7c, 0x76, 0x70, 0x79, 0x75, 0x7b, 0x7a, 0x75,
		0x70, 0x7d, 0x7e, 0x78, 0x77, 0x7a, 0x7b, 0x71,
		0x7a, 0x73, 0x74, 0x7f, 0x7d, 0x74, 0x71, 0x72,
		0x75, 0x7b, 0x78, 0x76, 0x7c, 0x77, 0x76, 0x7c,
		0x79, 0x70, 0x73, 0x75, 0x72, 0x7e, 0x7f, 0x79,
		0x8f, 0x83, 0x81, 0x8d, 0x88, 0x84, 0x8e, 0x87,
		0x86, 0x8f, 0x8b, 0x82, 0x83, 0x88, 0x84, 0x8e,
		0x89, 0x8c, 0x87, 0x80, 0x82, 0x81, 0x8d, 0x8a,
		0x8c, 0x86, 0x80, 0x89, 0x85, 0x8b, 0x8a, 0x85,
		0x80, 0x8d, 0x8e, 0x88, 0x87, 0x8a, 0x8b, 0x81,
		0x8a, 0x83, 0x84, 0x8f, 0x8d, 0x84, 0x81, 0x82,
		0x85, 0x8b, 0x88, 0x86, 0x8c, 0x87, 0x86, 0x8c,
		0x89, 0x80, 0x83, 0x85, 0x82, 0x8e, 0x8f, 0x89,
		0x4f, 0x43, 0x41, 0x4d, 0x48, 0x44, 0x4e, 0x47,
		0x46, 0x4f, 0x4b, 0x42, 0x43, 0x48, 0x44, 0x4e,
		0x49, 0x4c, 0x47, 0x40, 0x42, 0x41, 0x4d, 0x4a,
		0x4c, 0x46, 0x40, 0x49, 0x45, 0x4b, 0x4a, 0x45,
		0x40, 0x4d, 0x4e, 0x48, 0x47, 0x4a, 0x4b, 0x41,
		0x4a, 0x43, 0x44, 0x4f, 0x4d, 0x44, 0x41, 0x42,
		0x45, 0x4b, 0x48, 0x46, 0x4c, 0x47, 0x46, 0x4c,
		0x49, 0x40, 0x43, 0x45, 0x42, 0x4e, 0x4f, 0x49,
		0xff, 0xf3, 0xf1, 0xfd, 0xf8, 0xf4, 0xfe, 0xf7,
		0xf6, 0xff, 0xfb, 0xf2, 0xf3, 0xf8, 0xf4, 0xfe,
		0xf9, 0xfc, 0xf7, 0xf0, 0xf2, 0xf1, 0xfd, 0xfa,
		0xfc, 0xf6, 0xf0, 0xf9, 0xf5, 0xfb, 0xfa, 0xf5,
		0xf0, 0xfd, 0xfe, 0xf8, 0xf7, 0xfa, 0xfb, 0xf1,
		0xfa, 0xf3, 0xf4, 0xff, 0xfd, 0xf4, 0xf1, 0xf2,
		0xf5, 0xfb, 0xf8, 0xf6, 0xfc, 0xf7, 0xf6, 0xfc,
		0xf9, 0xf0, 0xf3, 0xf5, 0xf2, 0xfe, 0xff, 0xf9,
		0x1f, 0x13, 0x11, 0x1d, 0x18, 0x14, 0x1e, 0x17,
		0x16, 0x1f, 0x1b, 0x12, 0x13, 0x18, 0x14, 0x1e,
		0x19, 0x1c, 0x17, 0x10, 0x12, 0x11, 0x1d, 0x1a,
		0x1c, 0x16, 0x10, 0x19, 0x15, 0x1b, 0x1a, 0x15,
		0x10, 0x1d, 0x1e, 0x18, 0x17, 0x1a, 0x1b, 0x11,
		0x1a, 0x13, 0x14, 0x1f, 0x1d, 0x14, 0x11, 0x12,
		0x15, 0x1b, 0x18, 0x16, 0x1c, 0x17, 0x16, 0x1c,
		0x19, 0x10, 0x13, 0x15, 0x12, 0x1e, 0x1f, 0x19,
		0xcf, 0xc3, 0xc1, 0xcd, 0xc8, 0xc4, 0xce, 0xc7,
		0xc6, 0xcf, 0xcb, 0xc2, 0xc3, 0xc8, 0xc4, 0xce,
		0xc9, 0xcc, 0xc7, 0xc0, 0xc2, 0xc1, 0xcd, 0xca,
		0xcc, 0xc6, 0xc0, 0xc9, 0xc5, 0xcb, 0xca, 0xc5,
		0xc0, 0xcd, 0xce, 0xc8, 0xc7, 0xca, 0xcb, 0xc1,
		0xca, 0xc3, 0xc4, 0xcf, 0xcd, 0xc4, 0xc1, 0xc2,
		0xc5, 0xcb, 0xc8, 0xc6, 0xcc, 0xc7, 0xc6, 0xcc,
		0xc9, 0xc0, 0xc3, 0xc5, 0xc2, 0xce, 0xcf, 0xc9,
		0xef, 0xe3, 0xe1, 0xed, 0xe8, 0xe4, 0xee, 0xe7,
		0xe6, 0xef, 0xeb, 0xe2, 0xe3, 0xe8, 0xe4, 0xee,
		0xe9, 0xec, 0xe7, 0xe0, 0xe2, 0xe1, 0xed, 0xea,
		0xec, 0xe6, 0xe0, 0xe9, 0xe5, 0xeb, 0xea, 0xe5,
		0xe0, 0xed, 0xee, 0xe8, 0xe7, 0xea, 0xeb, 0xe1,
		0xea, 0xe3, 0xe4, 0xef, 0xed, 0xe4, 0xe1, 0xe2,
		0xe5, 0xeb, 0xe8, 0xe6, 0xec, 0xe7, 0xe6, 0xec,
		0xe9, 0xe0, 0xe3, 0xe5, 0xe2, 0xee, 0xef, 0xe9,
		0x8f, 0x83, 0x81, 0x8d, 0x88, 0x84, 0x8e, 0x87,
		0x86, 0x8f, 0x8b, 0x82, 0x83, 0x88, 0x84, 0x8e,
		0x89, 0x8c, 0x87, 0x80, 0x82, 0x81, 0x8d, 0x8a,
		0x8c, 0x86, 0x80, 0x89, 0x85, 0x8b, 0x8a, 0x85,
		0x80, 0x8d, 0x8e, 0x88, 0x87, 0x8a, 0x8b, 0x81,
		0x8a, 0x83, 0x84, 0x8f, 0x8d, 0x84, 0x81, 0x82,
		0x85, 0x8b, 0x88, 0x86, 0x8c, 0x87, 0x86, 0x8c,
		0x89, 0x80, 0x83, 0x85, 0x82, 0x8e, 0x8f, 0x89,
		0x8f, 0x83, 0x81, 0x8d, 0x88, 0x84, 0x8e, 0x87,
		0x86, 0x8f, 0x8b, 0x82, 0x83, 0x88, 0x84, 0x8e,
		0x89, 0x8c, 0x87, 0x80, 0x82, 0x81, 0x8d, 0x8a,
		0x8c, 0x86, 0x80, 0x89, 0x85, 0x8b, 0x8a, 0x85,
		0x80, 0x8d, 0x8e, 0x88, 0x87, 0x8a, 0x8b, 0x81,
		0x8a, 0x83, 0x84, 0x8f, 0x8d, 0x84, 0x81, 0x82,
		0x85, 0x8b, 0x88, 0x86, 0x8c, 0x87, 0x86, 0x8c,
		0x89, 0x80, 0x83, 0x85, 0x82, 0x8e, 0x8f, 0x89,
		0x2f, 0x23, 0x21, 0x2d, 0x28, 0x24, 0x2e, 0x27,
		0x26, 0x2f, 0x2b, 0x22, 0x23, 0x28, 0x24, 0x2e,
		0x29, 0x2c, 0x27, 0x20, 0x22, 0x21, 0x2d, 0x2a,
		0x2c, 0x26, 0x20, 0x29, 0x25, 0x2b, 0x2a, 0x25,
		0x20, 0x2d, 0x2e, 0x28, 0x27, 0x2a, 0x2b, 0x21,
		0x2a, 0x23, 0x24, 0x2f, 0x2d, 0x24, 0x21, 0x22,
		0x25, 0x2b, 0x28, 0x26, 0x2c, 0x27, 0x26, 0x2c,
		0x29, 0x20, 0x23, 0x25, 0x22, 0x2e, 0x2f, 0x29,
		0xdf, 0xd3, 0xd1, 0xdd, 0xd8, 0xd4, 0xde, 0xd7,
		0xd6, 0xdf, 0xdb, 0xd2, 0xd3, 0xd8, 0xd4, 0xde,
		0xd9, 0xdc, 0xd7, 0xd0, 0xd2, 0xd1, 0xdd, 0xda,
		0xdc, 0xd6, 0xd0, 0xd9, 0xd5, 0xdb, 0xda, 0xd5,
		0xd0, 0xdd, 0xde, 0xd8, 0xd7, 0xda, 0xdb, 0xd1,
		0xda, 0xd3, 0xd4, 0xdf, 0xdd, 0xd4, 0xd1, 0xd2,
		0xd5, 0xdb, 0xd8, 0xd6, 0xdc, 0xd7, 0xd6, 0xdc,
		0xd9, 0xd0, 0xd3, 0xd5, 0xd2, 0xde, 0xdf, 0xd9,
		0x4f, 0x43, 0x41, 0x4d, 0x48, 0x44, 0x4e, 0x47,
		0x46, 0x4f, 0x4b, 0x42, 0x43, 0x48, 0x44, 0x4e,
		0x49, 0x4c, 0x47, 0x40, 0x42, 0x41, 0x4d, 0x4a,
		0x4c, 0x46, 0x40, 0x49, 0x45, 0x4b, 0x4a, 0x45,
		0x40, 0x4d, 0x4e, 0x48, 0x47, 0x4a, 0x4b, 0x41,
		0x4a, 0x43, 0x44, 0x4f, 0x4d, 0x44, 0x41, 0x42,
		0x45, 0x4b, 0x48, 0x46, 0x4c, 0x47, 0x46, 0x4c,
		0x49, 0x40, 0x43, 0x45, 0x42, 0x4e, 0x4f, 0x49,
		0x6f, 0x63, 0x61, 0x6d, 0x68, 0x64, 0x6e, 0x67,
		0x66, 0x6f, 0x6b, 0x62, 0x63, 0x68, 0x64, 0x6e,
		0x69, 0x6c, 0x67, 0x60, 0x62, 0x61, 0x6d, 0x6a,
		0x6c, 0x66, 0x60, 0x69, 0x65, 0x6b, 0x6a, 0x65,
		0x60, 0x6d, 0x6e, 0x68, 0x67, 0x6a, 0x6b, 0x61,
		0x6a, 0x63, 0x64, 0x6f, 0x6d, 0x64, 0x61, 0x62,
		0x65, 0x6b, 0x68, 0x66, 0x6c, 0x67, 0x66, 0x6c,
		0x69, 0x60, 0x63, 0x65, 0x62, 0x6e, 0x6f, 0x69,
		0x9f, 0x93, 0x91, 0x9d, 0x98, 0x94, 0x9e, 0x97,
		0x96, 0x9f, 0x9b, 0x92, 0x93, 0x98, 0x94, 0x9e,
		0x99, 0x9c, 0x97, 0x90, 0x92, 0x91, 0x9d, 0x9a,
		0x9c, 0x96, 0x90, 0x99, 0x95, 0x9b, 0x9a, 0x95,
		0x90, 0x9d, 0x9e, 0x98, 0x97, 0x9a, 0x9b, 0x91,
		0x9a, 0x93, 0x94, 0x9f, 0x9d, 0x94, 0x91, 0x92,
		0x95, 0x9b, 0x98, 0x96, 0x9c, 0x97, 0x96, 0x9c,
		0x99, 0x90, 0x93, 0x95, 0x92, 0x9e, 0x9f, 0x99,
		0x2f, 0x23, 0x21, 0x2d, 0x28, 0x24, 0x2e, 0x27,
		0x26, 0x2f, 0x2b, 0x22, 0x23, 0x28, 0x24, 0x2e,
		0x29, 0x2c, 0x27, 0x20, 0x22, 0x21, 0x2d, 0x2a,
		0x2c, 0x26, 0x20, 0x29, 0x25, 0x2b, 0x2a, 0x25,
		0x20, 0x2d, 0x2e, 0x28, 0x27, 0x2a, 0x2b, 0x21,
		0x2a, 0x23, 0x24, 0x2f, 0x2d, 0x24, 0x21, 0x22,
		0x25, 0x2b, 0x28, 0x26, 0x2c, 0x27, 0x26, 0x2c,
		0x29, 0x20, 0x23, 0x25, 0x22, 0x2e, 0x2f, 0x29,
		0x1f, 0x13, 0x11, 0x1d, 0x18, 0x14, 0x1e, 0x17,
		0x16, 0x1f, 0x1b, 0x12, 0x13, 0x18, 0x14, 0x1e,
		0x19, 0x1c, 0x17, 0x10, 0x12, 0x11, 0x1d, 0x1a,
		0x1c, 0x16, 0x10, 0x19, 0x15, 0x1b, 0x1a, 0x15,
		0x10, 0x1d, 0x1e, 0x18, 0x17, 0x1a, 0x1b, 0x11,
		0x1a, 0x13, 0x14, 0x1f, 0x1d, 0x14, 0x11, 0x12,
		0x15, 0x1b, 0x18, 0x16, 0x1c, 0x17, 0x16, 0x1c,
		0x19, 0x10, 0x13, 0x15, 0x12, 0x1e, 0x1f, 0x19,
		0xbf, 0xb3, 0xb1, 0xbd, 0xb8, 0xb4, 0xbe, 0xb7,
		0xb6, 0xbf, 0xbb, 0xb2, 0xb3, 0xb8, 0xb4, 0xbe,
		0xb9, 0xbc, 0xb7, 0xb0, 0xb2, 0xb1, 0xbd, 0xba,
		0xbc, 0xb6, 0xb0, 0xb9, 0xb5, 0xbb, 0xba, 0xb5,
		0xb0, 0xbd, 0xbe, 0xb8, 0xb7, 0xba, 0xbb, 0xb1,
		0xba, 0xb3, 0xb4, 0xbf, 0xbd, 0xb4, 0xb1, 0xb2,
		0xb5, 0xbb, 0xb8, 0xb6, 0xbc, 0xb7, 0xb6, 0xbc,
		0xb9, 0xb0, 0xb3, 0xb5, 0xb2, 0xbe, 0xbf, 0xb9,
		0x7f, 0x73, 0x71, 0x7d, 0x78, 0x74, 0x7e, 0x77,
		0x76, 0x7f, 0x7b, 0x72, 0x73, 0x78, 0x74, 0x7e,
		0x79, 0x7c, 0x77, 0x70, 0x72, 0x71, 0x7d, 0x7a,
		0x7c, 0x76, 0x70, 0x79, 0x75, 0x7b, 0x7a, 0x75,
		0x70, 0x7d, 0x7e, 0x78, 0x77, 0x7a, 0x7b, 0x71,
		0x7a, 0x73, 0x74, 0x7f, 0x7d, 0x74, 0x71, 0x72,
		0x75, 0x7b, 0x78, 0x76, 0x7c, 0x77, 0x76, 0x7c,
		0x79, 0x70, 0x73, 0x75, 0x72, 0x7e, 0x7f, 0x79,
		0xff, 0xf3, 0xf1, 0xfd, 0xf8, 0xf4, 0xfe, 0xf7,
		0xf6, 0xff, 0xfb, 0xf2, 0xf3, 0xf8, 0xf4, 0xfe,
		0xf9, 0xfc, 0xf7, 0xf0, 0xf2, 0xf1, 0xfd, 0xfa,
		0xfc, 0xf6, 0xf0, 0xf9, 0xf5, 0xfb, 0xfa, 0xf5,
		0xf0, 0xfd, 0xfe, 0xf8, 0xf7, 0xfa, 0xfb, 0xf1,
		0xfa, 0xf3, 0xf4, 0xff, 0xfd, 0xf4, 0xf1, 0xf2,
		0xf5, 0xfb, 0xf8, 0xf6, 0xfc, 0xf7, 0xf6, 0xfc,
		0xf9, 0xf0, 0xf3, 0xf5, 0xf2, 0xfe, 0xff, 0xf9,
		0x5f, 0x53, 0x51, 0x5d, 0x58, 0x54, 0x5e, 0x57,
		0x56, 0x5f, 0x5b, 0x52, 0x53, 0x58, 0x54, 0x5e,
		0x59, 0x5c, 0x57, 0x50, 0x52, 0x51, 0x5d, 0x5a,
		0x5c, 0x56, 0x50, 0x59, 0x55, 0x5b, 0x5a, 0x55,
		0x50, 0x5d, 0x5e, 0x58, 0x57, 0x5a, 0x5b, 0x51,
		0x5a, 0x53, 0x54, 0x5f, 0x5d, 0x54, 0x51, 0x52,
		0x55, 0x5b, 0x58, 0x56, 0x5c, 0x57, 0x56, 0x5c,
		0x59, 0x50, 0x53, 0x55, 0x52, 0x5e, 0x5f, 0x59,
		0xcf, 0xc3, 0xc1, 0xcd, 0xc8, 0xc4, 0xce, 0xc7,
		0xc6, 0xcf, 0xcb, 0xc2, 0xc3, 0xc8, 0xc4, 0xce,
		0xc9, 0xcc, 0xc7, 0xc0, 0xc2, 0xc1, 0xcd, 0xca,
		0xcc, 0xc6, 0xc0, 0xc9, 0xc5, 0xcb, 0xca, 0xc5,
		0xc0, 0xcd, 0xce, 0xc8, 0xc7, 0xca, 0xcb, 0xc1,
		0xca, 0xc3, 0xc4, 0xcf, 0xcd, 0xc4, 0xc1, 0xc2,
		0xc5, 0xcb, 0xc8, 0xc6, 0xcc, 0xc7, 0xc6, 0xcc,
		0xc9, 0xc0, 0xc3, 0xc5, 0xc2, 0xce, 0xcf, 0xc9,
		0xbf, 0xb3, 0xb1, 0xbd, 0xb8, 0xb4, 0xbe, 0xb7,
		0xb6, 0xbf, 0xbb, 0xb2, 0xb3, 0xb8, 0xb4, 0xbe,
		0xb9, 0xbc, 0xb7, 0xb0, 0xb2, 0xb1, 0xbd, 0xba,
		0xbc, 0xb6, 0xb0, 0xb9, 0xb5, 0xbb, 0xba, 0xb5,
		0xb0, 0xbd, 0xbe, 0xb8, 0xb7, 0xba, 0xbb, 0xb1,
		0xba, 0xb3, 0xb4, 0xbf, 0xbd, 0xb4, 0xb1, 0xb2,
		0xb5, 0xbb, 0xb8, 0xb6, 0xbc, 0xb7, 0xb6, 0xbc,
		0xb9, 0xb0, 0xb3, 0xb5, 0xb2, 0xbe, 0xbf, 0xb9,
		0x9f, 0x93, 0x91, 0x9d, 0x98, 0x94, 0x9e, 0x97,
		0x96, 0x9f, 0x9b, 0x92, 0x93, 0x98, 0x94, 0x9e,
		0x99, 0x9c, 0x97, 0x90, 0x92, 0x91, 0x9d, 0x9a,
		0x9c, 0x96, 0x90, 0x99, 0x95, 0x9b, 0x9a, 0x95,
		0x90, 0x9d, 0x9e, 0x98, 0x97, 0x9a, 0x9b, 0x91,
		0x9a, 0x93, 0x94, 0x9f, 0x9d, 0x94, 0x91, 0x92,
		0x95, 0x9b, 0x98, 0x96, 0x9c, 0x97, 0x96, 0x9c,
		0x99, 0x90, 0x93, 0x95, 0x92, 0x9e, 0x9f, 0x99,
		0x3f, 0x33, 0x31, 0x3d, 0x38, 0x34, 0x3e, 0x37,
		0x36, 0x3f, 0x3b, 0x32, 0x33, 0x38, 0x34, 0x3e,
		0x39, 0x3c, 0x37, 0x30, 0x32, 0x31, 0x3d, 0x3a,
		0x3c, 0x36, 0x30, 0x39, 0x35, 0x3b, 0x3a, 0x35,
		0x30, 0x3d, 0x3e, 0x38, 0x37, 0x3a, 0x3b, 0x31,
		0x3a, 0x33, 0x34, 0x3f, 0x3d, 0x34, 0x31, 0x32,
		0x35, 0x3b, 0x38, 0x36, 0x3c, 0x37, 0x36, 0x3c,
		0x39, 0x30, 0x33, 0x35, 0x32, 0x3e, 0x3f, 0x39,
		0x7f, 0x73, 0x71, 0x7d, 0x78, 0x74, 0x7e, 0x77,
		0x76, 0x7f, 0x7b, 0x72, 0x73, 0x78, 0x74, 0x7e,
		0x79, 0x7c, 0x77, 0x70, 0x72, 0x71, 0x7d, 0x7a,
		0x7c, 0x76, 0x70, 0x79, 0x75, 0x7b, 0x7a, 0x75,
		0x70, 0x7d, 0x7e, 0x78, 0x77, 0x7a, 0x7b, 0x71,
		0x7a, 0x73, 0x74, 0x7f, 0x7d, 0x74, 0x71, 0x72,
		0x75, 0x7b, 0x78, 0x76, 0x7c, 0x77, 0x76, 0x7c,
		0x79, 0x70, 0x73, 0x75, 0x72, 0x7e, 0x7f, 0x79,
		0xef, 0xe3, 0xe1, 0xed, 0xe8, 0xe4, 0xee, 0xe7,
		0xe6, 0xef, 0xeb, 0xe2, 0xe3, 0xe8, 0xe4, 0xee,
		0xe9, 0xec, 0xe7, 0xe0, 0xe2, 0xe1, 0xed, 0xea,
		0xec, 0xe6, 0xe0, 0xe9, 0xe5, 0xeb, 0xea, 0xe5,
		0xe0, 0xed, 0xee, 0xe8, 0xe7, 0xea, 0xeb, 0xe1,
		0xea, 0xe3, 0xe4, 0xef, 0xed, 0xe4, 0xe1, 0xe2,
		0xe5, 0xeb, 0xe8, 0xe6, 0xec, 0xe7, 0xe6, 0xec,
		0xe9, 0xe0, 0xe3, 0xe5, 0xe2, 0xee, 0xef, 0xe9,
		0x3f, 0x33, 0x31, 0x3d, 0x38, 0x34, 0x3e, 0x37,
		0x36, 0x3f, 0x3b, 0x32, 0x33, 0x38, 0x34, 0x3e,
		0x39, 0x3c, 0x37, 0x30, 0x32, 0x31, 0x3d, 0x3a,
		0x3c, 0x36, 0x30, 0x39, 0x35, 0x3b, 0x3a, 0x35,
		0x30, 0x3d, 0x3e, 0x38, 0x37, 0x3a, 0x3b, 0x31,
		0x3a, 0x33, 0x34, 0x3f, 0x3d, 0x34, 0x31, 0x32,
		0x35, 0x3b, 0x38, 0x36, 0x3c, 0x37, 0x36, 0x3c,
		0x39, 0x30, 0x33, 0x35, 0x32, 0x3e, 0x3f, 0x39,
		0xaf, 0xa3, 0xa1, 0xad, 0xa8, 0xa4, 0xae, 0xa7,
		0xa6, 0xaf, 0xab, 0xa2, 0xa3, 0xa8, 0xa4, 0xae,
		0xa9, 0xac, 0xa7, 0xa0, 0xa2, 0xa1, 0xad, 0xaa,
		0xac, 0xa6, 0xa0, 0xa9, 0xa5, 0xab, 0xaa, 0xa5,
		0xa0, 0xad, 0xae, 0xa8, 0xa7, 0xaa, 0xab, 0xa1,
		0xaa, 0xa3, 0xa4, 0xaf, 0xad, 0xa4, 0xa1, 0xa2,
		0xa5, 0xab, 0xa8, 0xa6, 0xac, 0xa7, 0xa6, 0xac,
		0xa9, 0xa0, 0xa3, 0xa5, 0xa2, 0xae, 0xaf, 0xa9,
		0xaf, 0xa3, 0xa1, 0xad, 0xa8, 0xa4, 0xae, 0xa7,
		0xa6, 0xaf, 0xab, 0xa2, 0xa3, 0xa8, 0xa4, 0xae,
		0xa9, 0xac, 0xa7, 0xa0, 0xa2, 0xa1, 0xad, 0xaa,
		0xac, 0xa6, 0xa0, 0xa9, 0xa5, 0xab, 0xaa, 0xa5,
		0xa0, 0xad, 0xae, 0xa8, 0xa7, 0xaa, 0xab, 0xa1,
		0xaa, 0xa3, 0xa4, 0xaf, 0xad, 0xa4, 0xa1, 0xa2,
		0xa5, 0xab, 0xa8, 0xa6, 0xac, 0xa7, 0xa6, 0xac,
		0xa9, 0xa0, 0xa3, 0xa5, 0xa2, 0xae, 0xaf, 0xa9,
		0x0f, 0x03, 0x01, 0x0d, 0x08, 0x04, 0x0e, 0x07,
		0x06, 0x0f, 0x0b, 0x02, 0x03, 0x08, 0x04, 0x0e,
		0x09, 0x0c, 0x07, 0x00, 0x02, 0x01, 0x0d, 0x0a,
		0x0c, 0x06, 0x00, 0x09, 0x05, 0x0b, 0x0a, 0x05,
		0x00, 0x0d, 0x0e, 0x08, 0x07, 0x0a, 0x0b, 0x01,
		0x0a, 0x03, 0x04, 0x0f, 0x0d, 0x04, 0x01, 0x02,
		0x05, 0x0b, 0x08, 0x06, 0x0c, 0x07, 0x06, 0x0c,
		0x09, 0x00, 0x03, 0x05, 0x02, 0x0e, 0x0f, 0x09,
		0x5f, 0x53, 0x51, 0x5d, 0x58, 0x54, 0x5e, 0x57,
		0x56, 0x5f, 0x5b, 0x52, 0x53, 0x58, 0x54, 0x5e,
		0x59, 0x5c, 0x57, 0x50, 0x52, 0x51, 0x5d, 0x5a,
		0x5c, 0x56, 0x50, 0x59, 0x55, 0x5b, 0x5a, 0x55,
		0x50, 0x5d, 0x5e, 0x58, 0x57, 0x5a, 0x5b, 0x51,
		0x5a, 0x53, 0x54, 0x5f, 0x5d, 0x54, 0x51, 0x52,
		0x55, 0x5b, 0x58, 0x56, 0x5c, 0x57, 0x56, 0x5c,
		0x59, 0x50, 0x53, 0x55, 0x52, 0x5e, 0x5f, 0x59,
		0x6f, 0x63, 0x61, 0x6d, 0x68, 0x64, 0x6e, 0x67,
		0x66, 0x6f, 0x6b, 0x62, 0x63, 0x68, 0x64, 0x6e,
		0x69, 0x6c, 0x67, 0x60, 0x62, 0x61, 0x6d, 0x6a,
		0x6c, 0x66, 0x60, 0x69, 0x65, 0x6b, 0x6a, 0x65,
		0x60, 0x6d, 0x6e, 0x68, 0x67, 0x6a, 0x6b, 0x61,
		0x6a, 0x63, 0x64, 0x6f, 0x6d, 0x64, 0x61, 0x62,
		0x65, 0x6b, 0x68, 0x66, 0x6c, 0x67, 0x66, 0x6c,
		0x69, 0x60, 0x63, 0x65, 0x62, 0x6e, 0x6f, 0x69,
		0x0f, 0x03, 0x01, 0x0d, 0x08, 0x04, 0x0e, 0x07,
		0x06, 0x0f, 0x0b, 0x02, 0x03, 0x08, 0x04, 0x0e,
		0x09, 0x0c, 0x07, 0x00, 0x02, 0x01, 0x0d, 0x0a,
		0x0c, 0x06, 0x00, 0x09, 0x05, 0x0b, 0x0a, 0x05,
		0x00, 0x0d, 0x0e, 0x08, 0x07, 0x0a, 0x0b, 0x01,
		0x0a, 0x03, 0x04, 0x0f, 0x0d, 0x04, 0x01, 0x02,
		0x05, 0x0b, 0x08, 0x06, 0x0c, 0x07, 0x06, 0x0c,
		0x09, 0x00, 0x03, 0x05, 0x02, 0x0e, 0x0f, 0x09,
		0xdf, 0xd3, 0xd1, 0xdd, 0xd8, 0xd4, 0xde, 0xd7,
		0xd6, 0xdf, 0xdb, 0xd2, 0xd3, 0xd8, 0xd4, 0xde,
		0xd9, 0xdc, 0xd7, 0xd0, 0xd2, 0xd1, 0xdd, 0xda,
		0xdc, 0xd6, 0xd0, 0xd9, 0xd5, 0xdb, 0xda, 0xd5,
		0xd0, 0xdd, 0xde, 0xd8, 0xd7, 0xda, 0xdb, 0xd1,
		0xda, 0xd3, 0xd4, 0xdf, 0xdd, 0xd4, 0xd1, 0xd2,
		0xd5, 0xdb, 0xd8, 0xd6, 0xdc, 0xd7, 0xd6, 0xdc,
		0xd9, 0xd0, 0xd3, 0xd5, 0xd2, 0xde, 0xdf, 0xd9,
	},{
		0xa7, 0xad, 0xad, 0xa8, 0xae, 0xab, 0xa3, 0xa5,
		0xa0, 0xa6, 0xa6, 0xaf, 0xa9, 0xa0, 0xaa, 0xa3,
		0xa1, 0xa4, 0xa2, 0xa7, 0xa8, 0xa2, 0xa5, 0xac,
		0xab, 0xa1, 0xac, 0xaa, 0xa4, 0xae, 0xaf, 0xa9,
		0xaa, 0xa3, 0xa6, 0xaf, 0xa9, 0xa0, 0xa0, 0xa6,
		0xac, 0xaa, 0xab, 0xa1, 0xa7, 0xad, 0xad, 0xa8,
		0xaf, 0xa9, 0xa1, 0xa4, 0xa3, 0xa5, 0xae, 0xab,
		0xa5, 0xac, 0xa2, 0xa7, 0xa8, 0xa2, 0xa4, 0xae,
		0xd7, 0xdd, 0xdd, 0xd8, 0xde, 0xdb, 0xd3, 0xd5,
		0xd0, 0xd6, 0xd6, 0xdf, 0xd9, 0xd0, 0xda, 0xd3,
		0xd1, 0xd4, 0xd2, 0xd7, 0xd8, 0xd2, 0xd5, 0xdc,
		0xdb, 0xd1, 0xdc, 0xda, 0xd4, 0xde, 0xdf, 0xd9,
		0xda, 0xd3, 0xd6, 0xdf, 0xd9, 0xd0, 0xd0, 0xd6,
		0xdc, 0xda, 0xdb, 0xd1, 0xd7, 0xdd, 0xdd, 0xd8,
		0xdf, 0xd9, 0xd1, 0xd4, 0xd3, 0xd5, 0xde, 0xdb,
		0xd5, 0xdc, 0xd2, 0xd7, 0xd8, 0xd2, 0xd4, 0xde,
		0x07, 0x0d, 0x0d, 0x08, 0x0e, 0x0b, 0x03, 0x05,
		0x00, 0x06, 0x06, 0x0f, 0x09, 0x00, 0x0a, 0x03,
		0x01, 0x04, 0x02, 0x07, 0x08, 0x02, 0x05, 0x0c,
		0x0b, 0x01, 0x0c, 0x0a, 0x04, 0x0e, 0x0f, 0x09,
		0x0a, 0x03, 0x06, 0x0f, 0x09, 0x00, 0x00, 0x06,
		0x0c, 0x0a, 0x0b, 0x01, 0x07, 0x0d, 0x0d, 0x08,
		0x0f, 0x09, 0x01, 0x04, 0x03, 0x05, 0x0e, 0x0b,
		0x05, 0x0c, 0x02, 0x07, 0x08, 0x02, 0x04, 0x0e,
		0x77, 0x7d, 0x7d, 0x78, 0x7e, 0x7b, 0x73, 0x75,
		0x70, 0x76, 0x76, 0x7f, 0x79, 0x70, 0x7a, 0x73,
		0x71, 0x74, 0x72, 0x77, 0x78, 0x72, 0x75, 0x7c,
		0x7b, 0x71, 0x7c, 0x7a, 0x74, 0x7e, 0x7f, 0x79,
		0x7a, 0x73, 0x76, 0x7f, 0x79, 0x70, 0x70, 0x76,
		0x7c, 0x7a, 0x7b, 0x71, 0x77, 0x7d, 0x7d, 0x78,
		0x7f, 0x79, 0x71, 0x74, 0x73, 0x75, 0x7e, 0x7b,
		0x75, 0x7c, 0x72, 0x77, 0x78, 0x72, 0x74, 0x7e,
		0x97, 0x9d, 0x9d, 0x98, 0x9e, 0x9b, 0x93, 0x95,
		0x90, 0x96, 0x96, 0x9f, 0x99, 0x90, 0x9a, 0x93,
		0x91, 0x94, 0x92, 0x97, 0x98, 0x92, 0x95, 0x9c,
		0x9b, 0x91, 0x9c, 0x9a, 0x94, 0x9e, 0x9f, 0x99,
		0x9a, 0x93, 0x96, 0x9f, 0x99, 0x90, 0x90, 0x96,
		0x9c, 0x9a, 0x9b, 0x91, 0x97, 0x9d, 0x9d, 0x98,
		0x9f, 0x99, 0x91, 0x94, 0x93, 0x95, 0x9e, 0x9b,
		0x95, 0x9c, 0x92, 0x97, 0x98, 0x92, 0x94, 0x9e,
		0x07, 0x0d, 0x0d, 0x08, 0x0e, 0x0b, 0x03, 0x05,
		0x00, 0x06, 0x06, 0x0f, 0x09, 0x00, 0x0a, 0x03,
		0x01, 0x04, 0x02, 0x07, 0x08, 0x02, 0x05, 0x0c,
		0x0b, 0x01, 0x0c, 0x0a, 0x04, 0x0e, 0x0f, 0x09,
		0x0a, 0x03, 0x06, 0x0f, 0x09, 0x00, 0x00, 0x06,
		0x0c, 0x0a, 0x0b, 0x01, 0x07, 0x0d, 0x0d, 0x08,
		0x0f, 0x09, 0x01, 0x04, 0x03, 0x05, 0x0e, 0x0b,
		0x05, 0x0c, 0x02, 0x07, 0x08, 0x02, 0x04, 0x0e,
		0xe7, 0xed, 0xed, 0xe8, 0xee, 0xeb, 0xe3, 0xe5,
		0xe0, 0xe6, 0xe6, 0xef, 0xe9, 0xe0, 0xea, 0xe3,
		0xe1, 0xe4, 0xe2, 0xe7, 0xe8, 0xe2, 0xe5, 0xec,
		0xeb, 0xe1, 0xec, 0xea, 0xe4, 0xee, 0xef, 0xe9,
		0xea, 0xe3, 0xe6, 0xef, 0xe9, 0xe0, 0xe0, 0xe6,
		0xec, 0xea, 0xeb, 0xe1, 0xe7, 0xed, 0xed, 0xe8,
		0xef, 0xe9, 0xe1, 0xe4, 0xe3, 0xe5, 0xee, 0xeb,
		0xe5, 0xec, 0xe2, 0xe7, 0xe8, 0xe2, 0xe4, 0xee,
		0x97, 0x9d, 0x9d, 0x98, 0x9e, 0x9b, 0x93, 0x95,
		0x90, 0x96, 0x96, 0x9f, 0x99, 0x90, 0x9a, 0x93,
		0x91, 0x94, 0x92, 0x97, 0x98, 0x92, 0x95, 0x9c,
		0x9b, 0x91, 0x9c, 0x9a, 0x94, 0x9e, 0x9f, 0x99,
		0x9a, 0x93, 0x96, 0x9f, 0x99, 0x90, 0x90, 0x96,
		0x9c, 0x9a, 0x9b, 0x91, 0x97, 0x9d, 0x9d, 0x98,
		0x9f, 0x99, 0x91, 0x94, 0x93, 0x95, 0x9e, 0x9b,
		0x95, 0x9c, 0x92, 0x97, 0x98, 0x92, 0x94, 0x9e,
		0x67, 0x6d, 0x6d, 0x68, 0x6e, 0x6b, 0x63, 0x65,
		0x60, 0x66, 0x66, 0x6f, 0x69, 0x60, 0x6a, 0x63,
		0x61, 0x64, 0x62, 0x67, 0x68, 0x62, 0x65, 0x6c,
		0x6b, 0x61, 0x6c, 0x6a, 0x64, 0x6e, 0x6f, 0x69,
		0x6a, 0x63, 0x66, 0x6f, 0x69, 0x60, 0x60, 0x66,
		0x6c, 0x6a, 0x6b, 0x61, 0x67, 0x6d, 0x6d, 0x68,
		0x6f, 0x69, 0x61, 0x64, 0x63, 0x65, 0x6e, 0x6b,
		0x65, 0x6c, 0x62, 0x67, 0x68, 0x62, 0x64, 0x6e,
		0x37, 0x3d, 0x3d, 0x38, 0x3e, 0x3b, 0x33, 0x35,
		0x30, 0x36, 0x36, 0x3f, 0x39, 0x30, 0x3a, 0x33,
		0x31, 0x34, 0x32, 0x37, 0x38, 0x32, 0x35, 0x3c,
		0x3b, 0x31, 0x3c, 0x3a, 0x34, 0x3e, 0x3f, 0x39,
		0x3a, 0x33, 0x36, 0x3f, 0x39, 0x30, 0x30, 0x36,
		0x3c, 0x3a, 0x3b, 0x31, 0x37, 0x3d, 0x3d, 0x38,
		0x3f, 0x39, 0x31, 0x34, 0x33, 0x35, 0x3e, 0x3b,
		0x35, 0x3c, 0x32, 0x37, 0x38, 0x32, 0x34, 0x3e,
		0x37, 0x3d, 0x3d, 0x38, 0x3e, 0x3b, 0x33, 0x35,
		0x30, 0x36, 0x36, 0x3f, 0x39, 0x30, 0x3a, 0x33,
		0x31, 0x34, 0x32, 0x37, 0x38, 0x32, 0x35, 0x3c,
		0x3b, 0x31, 0x3c, 0x3a, 0x34, 0x3e, 0x3f, 0x39,
		0x3a, 0x33, 0x36, 0x3f, 0x39, 0x30, 0x30, 0x36,
		0x3c, 0x3a, 0x3b, 0x31, 0x37, 0x3d, 0x3d, 0x38,
		0x3f, 0x39, 0x31, 0x34, 0x33, 0x35, 0x3e, 0x3b,
		0x35, 0x3c, 0x32, 0x37, 0x38, 0x32, 0x34, 0x3e,
		0x47, 0x4d, 0x4d, 0x48, 0x4e, 0x4b, 0x43, 0x45,
		0x40, 0x46, 0x46, 0x4f, 0x49, 0x40, 0x4a, 0x43,
		0x41, 0x44, 0x42, 0x47, 0x48, 0x42, 0x45, 0x4c,
		0x4b, 0x41, 0x4c, 0x4a, 0x44, 0x4e, 0x4f, 0x49,
		0x4a, 0x43, 0x46, 0x4f, 0x49, 0x40, 0x40, 0x46,
		0x4c, 0x4a, 0x4b, 0x41, 0x47, 0x4d, 0x4d, 0x48,
		0x4f, 0x49, 0x41, 0x44, 0x43, 0x45, 0x4e, 0x4b,
		0x45, 0x4c, 0x42, 0x47, 0x48, 0x42, 0x44, 0x4e,
		0xf7, 0xfd, 0xfd, 0xf8, 0xfe, 0xfb, 0xf3, 0xf5,
		0xf0, 0xf6, 0xf6, 0xff, 0xf9, 0xf0, 0xfa, 0xf3,
		0xf1, 0xf4, 0xf2, 0xf7, 0xf8, 0xf2, 0xf5, 0xfc,
		0xfb, 0xf1, 0xfc, 0xfa, 0xf4, 0xfe, 0xff, 0xf9,
		0xfa, 0xf3, 0xf6, 0xff, 0xf9, 0xf0, 0xf0, 0xf6,
		0xfc, 0xfa, 0xfb, 0xf1, 0xf7, 0xfd, 0xfd, 0xf8,
		0xff, 0xf9, 0xf1, 0xf4, 0xf3, 0xf5, 0xfe, 0xfb,
		0xf5, 0xfc, 0xf2, 0xf7, 0xf8, 0xf2, 0xf4, 0xfe,
		0x67, 0x6d, 0x6d, 0x68, 0x6e, 0x6b, 0x63, 0x65,
		0x60, 0x66, 0x66, 0x6f, 0x69, 0x60, 0x6a, 0x63,
		0x61, 0x64, 0x62, 0x67, 0x68, 0x62, 0x65, 0x6c,
		0x6b, 0x61, 0x6c, 0x6a, 0x64, 0x6e, 0x6f, 0x69,
		0x6a, 0x63, 0x66, 0x6f, 0x69, 0x60, 0x60, 0x66,
		0x6c, 0x6a, 0x6b, 0x61, 0x67, 0x6d, 0x6d, 0x68,
		0x6f, 0x69, 0x61, 0x64, 0x63, 0x65, 0x6e, 0x6b,
		0x65, 0x6c, 0x62, 0x67, 0x68, 0x62, 0x64, 0x6e,
		0x57, 0x5d, 0x5d, 0x58, 0x5e, 0x5b, 0x53, 0x55,
		0x50, 0x56, 0x56, 0x5f, 0x59, 0x50, 0x5a, 0x53,
		0x51, 0x54, 0x52, 0x57, 0x58, 0x52, 0x55, 0x5c,
		0x5b, 0x51, 0x5c, 0x5a, 0x54, 0x5e, 0x5f, 0x59,
		0x5a, 0x53, 0x56, 0x5f, 0x59, 0x50, 0x50, 0x56,
		0x5c, 0x5a, 0x5b, 0x51, 0x57, 0x5d, 0x5d, 0x58,
		0x5f, 0x59, 0x51, 0x54, 0x53, 0x55, 0x5e, 0x5b,
		0x55, 0x5c, 0x52, 0x57, 0x58, 0x52, 0x54, 0x5e,
		0xa7, 0xad, 0xad, 0xa8, 0xae, 0xab, 0xa3, 0xa5,
		0xa0, 0xa6, 0xa6, 0xaf, 0xa9, 0xa0, 0xaa, 0xa3,
		0xa1, 0xa4, 0xa2, 0xa7, 0xa8, 0xa2, 0xa5, 0xac,
		0xab, 0xa1, 0xac, 0xaa, 0xa4, 0xae, 0xaf, 0xa9,
		0xaa, 0xa3, 0xa6, 0xaf, 0xa9, 0xa0, 0xa0, 0xa6,
		0xac, 0xaa, 0xab, 0xa1, 0xa7, 0xad, 0xad, 0xa8,
		0xaf, 0xa9, 0xa1, 0xa4, 0xa3, 0xa5, 0xae, 0xab,
		0xa5, 0xac, 0xa2, 0xa7, 0xa8, 0xa2, 0xa4, 0xae,
		0x17, 0x1d, 0x1d, 0x18, 0x1e, 0x1b, 0x13, 0x15,
		0x10, 0x16, 0x16, 0x1f, 0x19, 0x10, 0x1a, 0x13,
		0x11, 0x14, 0x12, 0x17, 0x18, 0x12, 0x15, 0x1c,
		0x1b, 0x11, 0x1c, 0x1a, 0x14, 0x1e, 0x1f, 0x19,
		0x1a, 0x13, 0x16, 0x1f, 0x19, 0x10, 0x10, 0x16,
		0x1c, 0x1a, 0x1b, 0x11, 0x17, 0x1d, 0x1d, 0x18,
		0x1f, 0x19, 0x11, 0x14, 0x13, 0x15, 0x1e, 0x1b,
		0x15, 0x1c, 0x12, 0x17, 0x18, 0x12, 0x14, 0x1e,
		0x27, 0x2d, 0x2d, 0x28, 0x2e, 0x2b, 0x23, 0x25,
		0x20, 0x26, 0x26, 0x2f, 0x29, 0x20, 0x2a, 0x23,
		0x21, 0x24, 0x22, 0x27, 0x28, 0x22, 0x25, 0x2c,
		0x2b, 0x21, 0x2c, 0x2a, 0x24, 0x2e, 0x2f, 0x29,
		0x2a, 0x23, 0x26, 0x2f, 0x29, 0x20, 0x20, 0x26,
		0x2c, 0x2a, 0x2b, 0x21, 0x27, 0x2d, 0x2d, 0x28,
		0x2f, 0x29, 0x21, 0x24, 0x23, 0x25, 0x2e, 0x2b,
		0x25, 0x2c, 0x22, 0x27, 0x28, 0x22, 0x24, 0x2e,
		0xd7, 0xdd, 0xdd, 0xd8, 0xde, 0xdb, 0xd3, 0xd5,
		0xd0, 0xd6, 0xd6, 0xdf, 0xd9, 0xd0, 0xda, 0xd3,
		0xd1, 0xd4, 0xd2, 0xd7, 0xd8, 0xd2, 0xd5, 0xdc,
		0xdb, 0xd1, 0xdc, 0xda, 0xd4, 0xde, 0xdf, 0xd9,
		0xda, 0xd3, 0xd6, 0xdf, 0xd9, 0xd0, 0xd0, 0xd6,
		0xdc, 0xda, 0xdb, 0xd1, 0xd7, 0xdd, 0xdd, 0xd8,
		0xdf, 0xd9, 0xd1, 0xd4, 0xd3, 0xd5, 0xde, 0xdb,
		0xd5, 0xdc, 0xd2, 0xd7, 0xd8, 0xd2, 0xd4, 0xde,
		0x87, 0x8d, 0x8d, 0x88, 0x8e, 0x8b, 0x83, 0x85,
		0x80, 0x86, 0x86, 0x8f, 0x89, 0x80, 0x8a, 0x83,
		0x81, 0x84, 0x82, 0x87, 0x88, 0x82, 0x85, 0x8c,
		0x8b, 0x81, 0x8c, 0x8a, 0x84, 0x8e, 0x8f, 0x89,
		0x8a, 0x83, 0x86, 0x8f, 0x89, 0x80, 0x80, 0x86,
		0x8c, 0x8a, 0x8b, 0x81, 0x87, 0x8d, 0x8d, 0x88,
		0x8f, 0x89, 0x81, 0x84, 0x83, 0x85, 0x8e, 0x8b,
		0x85, 0x8c, 0x82, 0x87, 0x88, 0x82, 0x84, 0x8e,
		0xc7, 0xcd, 0xcd, 0xc8, 0xce, 0xcb, 0xc3, 0xc5,
		0xc0, 0xc6, 0xc6, 0xcf, 0xc9, 0xc0, 0xca, 0xc3,
		0xc1, 0xc4, 0xc2, 0xc7, 0xc8, 0xc2, 0xc5, 0xcc,
		0xcb, 0xc1, 0xcc, 0xca, 0xc4, 0xce, 0xcf, 0xc9,
		0xca, 0xc3, 0xc6, 0xcf, 0xc9, 0xc0, 0xc0, 0xc6,
		0xcc, 0xca, 0xcb, 0xc1, 0xc7, 0xcd, 0xcd, 0xc8,
		0xcf, 0xc9, 0xc1, 0xc4, 0xc3, 0xc5, 0xce, 0xcb,
		0xc5, 0xcc, 0xc2, 0xc7, 0xc8, 0xc2, 0xc4, 0xce,
		0x57, 0x5d, 0x5d, 0x58, 0x5e, 0x5b, 0x53, 0x55,
		0x50, 0x56, 0x56, 0x5f, 0x59, 0x50, 0x5a, 0x53,
		0x51, 0x54, 0x52, 0x57, 0x58, 0x52, 0x55, 0x5c,
		0x5b, 0x51, 0x5c, 0x5a, 0x54, 0x5e, 0x5f, 0x59,
		0x5a, 0x53, 0x56, 0x5f, 0x59, 0x50, 0x50, 0x56,
		0x5c, 0x5a, 0x5b, 0x51, 0x57, 0x5d, 0x5d, 0x58,
		0x5f, 0x59, 0x51, 0x54, 0x53, 0x55, 0x5e, 0x5b,
		0x55, 0x5c, 0x52, 0x57, 0x58, 0x52, 0x54, 0x5e,
		0x77, 0x7d, 0x7d, 0x78, 0x7e, 0x7b, 0x73, 0x75,
		0x70, 0x76, 0x76, 0x7f, 0x79, 0x70, 0x7a, 0x73,
		0x71, 0x74, 0x72, 0x77, 0x78, 0x72, 0x75, 0x7c,
		0x7b, 0x71, 0x7c, 0x7a, 0x74, 0x7e, 0x7f, 0x79,
		0x7a, 0x73, 0x76, 0x7f, 0x79, 0x70, 0x70, 0x76,
		0x7c, 0x7a, 0x7b, 0x71, 0x77, 0x7d, 0x7d, 0x78,
		0x7f, 0x79, 0x71, 0x74, 0x73, 0x75, 0x7e, 0x7b,
		0x75, 0x7c, 0x72, 0x77, 0x78, 0x72, 0x74, 0x7e,
		0xe7, 0xed, 0xed, 0xe8, 0xee, 0xeb, 0xe3, 0xe5,
		0xe0, 0xe6, 0xe6, 0xef, 0xe9, 0xe0, 0xea, 0xe3,
		0xe1, 0xe4, 0xe2, 0xe7, 0xe8, 0xe2, 0xe5, 0xec,
		0xeb, 0xe1, 0xec, 0xea, 0xe4, 0xee, 0xef, 0xe9,
		0xea, 0xe3, 0xe6, 0xef, 0xe9, 0xe0, 0xe0, 0xe6,
		0xec, 0xea, 0xeb, 0xe1, 0xe7, 0xed, 0xed, 0xe8,
		0xef, 0xe9, 0xe1, 0xe4, 0xe3, 0xe5, 0xee, 0xeb,
		0xe5, 0xec, 0xe2, 0xe7, 0xe8, 0xe2, 0xe4, 0xee,
		0xb7, 0xbd, 0xbd, 0xb8, 0xbe, 0xbb, 0xb3, 0xb5,
		0xb0, 0xb6, 0xb6, 0xbf, 0xb9, 0xb0, 0xba, 0xb3,
		0xb1, 0xb4, 0xb2, 0xb7, 0xb8, 0xb2, 0xb5, 0xbc,
		0xbb, 0xb1, 0xbc, 0xba, 0xb4, 0xbe, 0xbf, 0xb9,
		0xba, 0xb3, 0xb6, 0xbf, 0xb9, 0xb0, 0xb0, 0xb6,
		0xbc, 0xba, 0xbb, 0xb1, 0xb7, 0xbd, 0xbd, 0xb8,
		0xbf, 0xb9, 0xb1, 0xb4, 0xb3, 0xb5, 0xbe, 0xbb,
		0xb5, 0xbc, 0xb2, 0xb7, 0xb8, 0xb2, 0xb4, 0xbe,
		0xc7, 0xcd, 0xcd, 0xc8, 0xce, 0xcb, 0xc3, 0xc5,
		0xc0, 0xc6, 0xc6, 0xcf, 0xc9, 0xc0, 0xca, 0xc3,
		0xc1, 0xc4, 0xc2, 0xc7, 0xc8, 0xc2, 0xc5, 0xcc,
		0xcb, 0xc1, 0xcc, 0xca, 0xc4, 0xce, 0xcf, 0xc9,
		0xca, 0xc3, 0xc6, 0xcf, 0xc9, 0xc0, 0xc0, 0xc6,
		0xcc, 0xca, 0xcb, 0xc1, 0xc7, 0xcd, 0xcd, 0xc8,
		0xcf, 0xc9, 0xc1, 0xc4, 0xc3, 0xc5, 0xce, 0xcb,
		0xc5, 0xcc, 0xc2, 0xc7, 0xc8, 0xc2, 0xc4, 0xce,
		0x47, 0x4d, 0x4d, 0x48, 0x4e, 0x4b, 0x43, 0x45,
		0x40, 0x46, 0x46, 0x4f, 0x49, 0x40, 0x4a, 0x43,
		0x41, 0x44, 0x42, 0x47, 0x48, 0x42, 0x45, 0x4c,
		0x4b, 0x41, 0x4c, 0x4a, 0x44, 0x4e, 0x4f, 0x49,
		0x4a, 0x43, 0x46, 0x4f, 0x49, 0x40, 0x40, 0x46,
		0x4c, 0x4a, 0x4b, 0x41, 0x47, 0x4d, 0x4d, 0x48,
		0x4f, 0x49, 0x41, 0x44, 0x43, 0x45, 0x4e, 0x4b,
		0x45, 0x4c, 0x42, 0x47, 0x48, 0x42, 0x44, 0x4e,
		0xb7, 0xbd, 0xbd, 0xb8, 0xbe, 0xbb, 0xb3, 0xb5,
		0xb0, 0xb6, 0xb6, 0xbf, 0xb9, 0xb0, 0xba, 0xb3,
		0xb1, 0xb4, 0xb2, 0xb7, 0xb8, 0xb2, 0xb5, 0xbc,
		0xbb, 0xb1, 0xbc, 0xba, 0xb4, 0xbe, 0xbf, 0xb9,
		0xba, 0xb3, 0xb6, 0xbf, 0xb9, 0xb0, 0xb0, 0xb6,
		0xbc, 0xba, 0xbb, 0xb1, 0xb7, 0xbd, 0xbd, 0xb8,
		0xbf, 0xb9, 0xb1, 0xb4, 0xb3, 0xb5, 0xbe, 0xbb,
		0xb5, 0xbc, 0xb2, 0xb7, 0xb8, 0xb2, 0xb4, 0xbe,
		0x27, 0x2d, 0x2d, 0x28, 0x2e, 0x2b, 0x23, 0x25,
		0x20, 0x26, 0x26, 0x2f, 0x29, 0x20, 0x2a, 0x23,
		0x21, 0x24, 0x22, 0x27, 0x28, 0x22, 0x25, 0x2c,
		0x2b, 0x21, 0x2c, 0x2a, 0x24, 0x2e, 0x2f, 0x29,
		0x2a, 0x23, 0x26, 0x2f, 0x29, 0x20, 0x20, 0x26,
		0x2c, 0x2a, 0x2b, 0x21, 0x27, 0x2d, 0x2d, 0x28,
		0x2f, 0x29, 0x21, 0x24, 0x23, 0x25, 0x2e, 0x2b,
		0x25, 0x2c, 0x22, 0x27, 0x28, 0x22, 0x24, 0x2e,
		0xf7, 0xfd, 0xfd, 0xf8, 0xfe, 0xfb, 0xf3, 0xf5,
		0xf0, 0xf6, 0xf6, 0xff, 0xf9, 0xf0, 0xfa, 0xf3,
		0xf1, 0xf4, 0xf2, 0xf7, 0xf8, 0xf2, 0xf5, 0xfc,
		0xfb, 0xf1, 0xfc, 0xfa, 0xf4, 0xfe, 0xff, 0xf9,
		0xfa, 0xf3, 0xf6, 0xff, 0xf9, 0xf0, 0xf0, 0xf6,
		0xfc, 0xfa, 0xfb, 0xf1, 0xf7, 0xfd, 0xfd, 0xf8,
		0xff, 0xf9, 0xf1, 0xf4, 0xf3, 0xf5, 0xfe, 0xfb,
		0xf5, 0xfc, 0xf2, 0xf7, 0xf8, 0xf2, 0xf4, 0xfe,
		0x87, 0x8d, 0x8d, 0x88, 0x8e, 0x8b, 0x83, 0x85,
		0x80, 0x86, 0x86, 0x8f, 0x89, 0x80, 0x8a, 0x83,
		0x81, 0x84, 0x82, 0x87, 0x88, 0x82, 0x85, 0x8c,
		0x8b, 0x81, 0x8c, 0x8a, 0x84, 0x8e, 0x8f, 0x89,
		0x8a, 0x83, 0x86, 0x8f, 0x89, 0x80, 0x80, 0x86,
		0x8c, 0x8a, 0x8b, 0x81, 0x87, 0x8d, 0x8d, 0x88,
		0x8f, 0x89, 0x81, 0x84, 0x83, 0x85, 0x8e, 0x8b,
		0x85, 0x8c, 0x82, 0x87, 0x88, 0x82, 0x84, 0x8e,
		0x17, 0x1d, 0x1d, 0x18, 0x1e, 0x1b, 0x13, 0x15,
		0x10, 0x16, 0x16, 0x1f, 0x19, 0x10, 0x1a, 0x13,
		0x11, 0x14, 0x12, 0x17, 0x18, 0x12, 0x15, 0x1c,
		0x1b, 0x11, 0x1c, 0x1a, 0x14, 0x1e, 0x1f, 0x19,
		0x1a, 0x13, 0x16, 0x1f, 0x19, 0x10, 0x10, 0x16,
		0x1c, 0x1a, 0x1b, 0x11, 0x17, 0x1d, 0x1d, 0x18,
		0x1f, 0x19, 0x11, 0x14, 0x13, 0x15, 0x1e, 0x1b,
		0x15, 0x1c, 0x12, 0x17, 0x18, 0x12, 0x14, 0x1e,
		0xd7, 0xdd, 0xdd, 0xd8, 0xde, 0xdb, 0xd3, 0xd5,
		0xd0, 0xd6, 0xd6, 0xdf, 0xd9, 0xd0, 0xda, 0xd3,
		0xd1, 0xd4, 0xd2, 0xd7, 0xd8, 0xd2, 0xd5, 0xdc,
		0xdb, 0xd1, 0xdc, 0xda, 0xd4, 0xde, 0xdf, 0xd9,
		0xda, 0xd3, 0xd6, 0xdf, 0xd9, 0xd0, 0xd0, 0xd6,
		0xdc, 0xda, 0xdb, 0xd1, 0xd7, 0xdd, 0xdd, 0xd8,
		0xdf, 0xd9, 0xd1, 0xd4, 0xd3, 0xd5, 0xde, 0xdb,
		0xd5, 0xdc, 0xd2, 0xd7, 0xd8, 0xd2, 0xd4, 0xde,
		0x17, 0x1d, 0x1d, 0x18, 0x1e, 0x1b, 0x13, 0x15,
		0x10, 0x16, 0x16, 0x1f, 0x19, 0x10, 0x1a, 0x13,
		0x11, 0x14, 0x12, 0x17, 0x18, 0x12, 0x15, 0x1c,
		0x1b, 0x11, 0x1c, 0x1a, 0x14, 0x1e, 0x1f, 0x19,
		0x1a, 0x13, 0x16, 0x1f, 0x19, 0x10, 0x10, 0x16,
		0x1c, 0x1a, 0x1b, 0x11, 0x17, 0x1d, 0x1d, 0x18,
		0x1f, 0x19, 0x11, 0x14, 0x13, 0x15, 0x1e, 0x1b,
		0x15, 0x1c, 0x12, 0x17, 0x18, 0x12, 0x14, 0x1e,
		0x67, 0x6d, 0x6d, 0x68, 0x6e, 0x6b, 0x63, 0x65,
		0x60, 0x66, 0x66, 0x6f, 0x69, 0x60, 0x6a, 0x63,
		0x61, 0x64, 0x62, 0x67, 0x68, 0x62, 0x65, 0x6c,
		0x6b, 0x61, 0x6c, 0x6a, 0x64, 0x6e, 0x6f, 0x69,
		0x6a, 0x63, 0x66, 0x6f, 0x69, 0x60, 0x60, 0x66,
		0x6c, 0x6a, 0x6b, 0x61, 0x67, 0x6d, 0x6d, 0x68,
		0x6f, 0x69, 0x61, 0x64, 0x63, 0x65, 0x6e, 0x6b,
		0x65, 0x6c, 0x62, 0x67, 0x68, 0x62, 0x64, 0x6e,
		0xa7, 0xad, 0xad, 0xa8, 0xae, 0xab, 0xa3, 0xa5,
		0xa0, 0xa6, 0xa6, 0xaf, 0xa9, 0xa0, 0xaa, 0xa3,
		0xa1, 0xa4, 0xa2, 0xa7, 0xa8, 0xa2, 0xa5, 0xac,
		0xab, 0xa1, 0xac, 0xaa, 0xa4, 0xae, 0xaf, 0xa9,
		0xaa, 0xa3, 0xa6, 0xaf, 0xa9, 0xa0, 0xa0, 0xa6,
		0xac, 0xaa, 0xab, 0xa1, 0xa7, 0xad, 0xad, 0xa8,
		0xaf, 0xa9, 0xa1, 0xa4, 0xa3, 0xa5, 0xae, 0xab,
		0xa5, 0xac, 0xa2, 0xa7, 0xa8, 0xa2, 0xa4, 0xae,
		0x47, 0x4d, 0x4d, 0x48, 0x4e, 0x4b, 0x43, 0x45,
		0x40, 0x46, 0x46, 0x4f, 0x49, 0x40, 0x4a, 0x43,
		0x41, 0x44, 0x42, 0x47, 0x48, 0x42, 0x45, 0x4c,
		0x4b, 0x41, 0x4c, 0x4a, 0x44, 0x4e, 0x4f, 0x49,
		0x4a, 0x43, 0x46, 0x4f, 0x49, 0x40, 0x40, 0x46,
		0x4c, 0x4a, 0x4b, 0x41, 0x47, 0x4d, 0x4d, 0x48,
		0x4f, 0x49, 0x41, 0x44, 0x43, 0x45, 0x4e, 0x4b,
		0x45, 0x4c, 0x42, 0x47, 0x48, 0x42, 0x44, 0x4e,
		0xd7, 0xdd, 0xdd, 0xd8, 0xde, 0xdb, 0xd3, 0xd5,
		0xd0, 0xd6, 0xd6, 0xdf, 0xd9, 0xd0, 0xda, 0xd3,
		0xd1, 0xd4, 0xd2, 0xd7, 0xd8, 0xd2, 0xd5, 0xdc,
		0xdb, 0xd1, 0xdc, 0xda, 0xd4, 0xde, 0xdf, 0xd9,
		0xda, 0xd3, 0xd6, 0xdf, 0xd9, 0xd0, 0xd0, 0xd6,
		0xdc, 0xda, 0xdb, 0xd1, 0xd7, 0xdd, 0xdd, 0xd8,
		0xdf, 0xd9, 0xd1, 0xd4, 0xd3, 0xd5, 0xde, 0xdb,
		0xd5, 0xdc, 0xd2, 0xd7, 0xd8, 0xd2, 0xd4, 0xde,
		0x97, 0x9d, 0x9d, 0x98, 0x9e, 0x9b, 0x93, 0x95,
		0x90, 0x96, 0x96, 0x9f, 0x99, 0x90, 0x9a, 0x93,
		0x91, 0x94, 0x92, 0x97, 0x98, 0x92, 0x95, 0x9c,
		0x9b, 0x91, 0x9c, 0x9a, 0x94, 0x9e, 0x9f, 0x99,
		0x9a, 0x93, 0x96, 0x9f, 0x99, 0x90, 0x90, 0x96,
		0x9c, 0x9a, 0x9b, 0x91, 0x97, 0x9d, 0x9d, 0x98,
		0x9f, 0x99, 0x91, 0x94, 0x93, 0x95, 0x9e, 0x9b,
		0x95, 0x9c, 0x92, 0x97, 0x98, 0x92, 0x94, 0x9e,
		0x07, 0x0d, 0x0d, 0x08, 0x0e, 0x0b, 0x03, 0x05,
		0x00, 0x06, 0x06, 0x0f, 0x09, 0x00, 0x0a, 0x03,
		0x01, 0x04, 0x02, 0x07, 0x08, 0x02, 0x05, 0x0c,
		0x0b, 0x01, 0x0c, 0x0a, 0x04, 0x0e, 0x0f, 0x09,
		0x0a, 0x03, 0x06, 0x0f, 0x09, 0x00, 0x00, 0x06,
		0x0c, 0x0a, 0x0b, 0x01, 0x07, 0x0d, 0x0d, 0x08,
		0x0f, 0x09, 0x01, 0x04, 0x03, 0x05, 0x0e, 0x0b,
		0x05, 0x0c, 0x02, 0x07, 0x08, 0x02, 0x04, 0x0e,
		0x87, 0x8d, 0x8d, 0x88, 0x8e, 0x8b, 0x83, 0x85,
		0x80, 0x86, 0x86, 0x8f, 0x89, 0x80, 0x8a, 0x83,
		0x81, 0x84, 0x82, 0x87, 0x88, 0x82, 0x85, 0x8c,
		0x8b, 0x81, 0x8c, 0x8a, 0x84, 0x8e, 0x8f, 0x89,
		0x8a, 0x83, 0x86, 0x8f, 0x89, 0x80, 0x80, 0x86,
		0x8c, 0x8a, 0x8b, 0x81, 0x87, 0x8d, 0x8d, 0x88,
		0x8f, 0x89, 0x81, 0x84, 0x83, 0x85, 0x8e, 0x8b,
		0x85, 0x8c, 0x82, 0x87, 0x88, 0x82, 0x84, 0x8e,
		0x67, 0x6d, 0x6d, 0x68, 0x6e, 0x6b, 0x63, 0x65,
		0x60, 0x66, 0x66, 0x6f, 0x69, 0x60, 0x6a, 0x63,
		0x61, 0x64, 0x62, 0x67, 0x68, 0x62, 0x65, 0x6c,
		0x6b, 0x61, 0x6c, 0x6a, 0x64, 0x6e, 0x6f, 0x69,
		0x6a, 0x63, 0x66, 0x6f, 0x69, 0x60, 0x60, 0x66,
		0x6c, 0x6a, 0x6b, 0x61, 0x67, 0x6d, 0x6d, 0x68,
		0x6f, 0x69, 0x61, 0x64, 0x63, 0x65, 0x6e, 0x6b,
		0x65, 0x6c, 0x62, 0x67, 0x68, 0x62, 0x64, 0x6e,
		0xf7, 0xfd, 0xfd, 0xf8, 0xfe, 0xfb, 0xf3, 0xf5,
		0xf0, 0xf6, 0xf6, 0xff, 0xf9, 0xf0, 0xfa, 0xf3,
		0xf1, 0xf4, 0xf2, 0xf7, 0xf8, 0xf2, 0xf5, 0xfc,
		0xfb, 0xf1, 0xfc, 0xfa, 0xf4, 0xfe, 0xff, 0xf9,
		0xfa, 0xf3, 0xf6, 0xff, 0xf9, 0xf0, 0xf0, 0xf6,
		0xfc, 0xfa, 0xfb, 0xf1, 0xf7, 0xfd, 0xfd, 0xf8,
		0xff, 0xf9, 0xf1, 0xf4, 0xf3, 0xf5, 0xfe, 0xfb,
		0xf5, 0xfc, 0xf2, 0xf7, 0xf8, 0xf2, 0xf4, 0xfe,
		0x97, 0x9d, 0x9d, 0x98, 0x9e, 0x9b, 0x93, 0x95,
		0x90, 0x96, 0x96, 0x9f, 0x99, 0x90, 0x9a, 0x93,
		0x91, 0x94, 0x92, 0x97, 0x98, 0x92, 0x95, 0x9c,
		0x9b, 0x91, 0x9c, 0x9a, 0x94, 0x9e, 0x9f, 0x99,
		0x9a, 0x93, 0x96, 0x9f, 0x99, 0x90, 0x90, 0x96,
		0x9c, 0x9a, 0x9b, 0x91, 0x97, 0x9d, 0x9d, 0x98,
		0x9f, 0x99, 0x91, 0x94, 0x93, 0x95, 0x9e, 0x9b,
		0x95, 0x9c, 0x92, 0x97, 0x98, 0x92, 0x94, 0x9e,
		0x37, 0x3d, 0x3d, 0x38, 0x3e, 0x3b, 0x33, 0x35,
		0x30, 0x36, 0x36, 0x3f, 0x39, 0x30, 0x3a, 0x33,
		0x31, 0x34, 0x32, 0x37, 0x38, 0x32, 0x35, 0x3c,
		0x3b, 0x31, 0x3c, 0x3a, 0x34, 0x3e, 0x3f, 0x39,
		0x3a, 0x33, 0x36, 0x3f, 0x39, 0x30, 0x30, 0x36,
		0x3c, 0x3a, 0x3b, 0x31, 0x37, 0x3d, 0x3d, 0x38,
		0x3f, 0x39, 0x31, 0x34, 0x33, 0x35, 0x3e, 0x3b,
		0x35, 0x3c, 0x32, 0x37, 0x38, 0x32, 0x34, 0x3e,
		0x87, 0x8d, 0x8d, 0x88, 0x8e, 0x8b, 0x83, 0x85,
		0x80, 0x86, 0x86, 0x8f, 0x89, 0x80, 0x8a, 0x83,
		0x81, 0x84, 0x82, 0x87, 0x88, 0x82, 0x85, 0x8c,
		0x8b, 0x81, 0x8c, 0x8a, 0x84, 0x8e, 0x8f, 0x89,
		0x8a, 0x83, 0x86, 0x8f, 0x89, 0x80, 0x80, 0x86,
		0x8c, 0x8a, 0x8b, 0x81, 0x87, 0x8d, 0x8d, 0x88,
		0x8f, 0x89, 0x81, 0x84, 0x83, 0x85, 0x8e, 0x8b,
		0x85, 0x8c, 0x82, 0x87, 0x88, 0x82, 0x84, 0x8e,
		0x07, 0x0d, 0x0d, 0x08, 0x0e, 0x0b, 0x03, 0x05,
		0x00, 0x06, 0x06, 0x0f, 0x09, 0x00, 0x0a, 0x03,
		0x01, 0x04, 0x02, 0x07, 0x08, 0x02, 0x05, 0x0c,
		0x0b, 0x01, 0x0c, 0x0a, 0x04, 0x0e, 0x0f, 0x09,
		0x0a, 0x03, 0x06, 0x0f, 0x09, 0x00, 0x00, 0x06,
		0x0c, 0x0a, 0x0b, 0x01, 0x07, 0x0d, 0x0d, 0x08,
		0x0f, 0x09, 0x01, 0x04, 0x03, 0x05, 0x0e, 0x0b,
		0x05, 0x0c, 0x02, 0x07, 0x08, 0x02, 0x04, 0x0e,
		0x77, 0x7d, 0x7d, 0x78, 0x7e, 0x7b, 0x73, 0x75,
		0x70, 0x76, 0x76, 0x7f, 0x79, 0x70, 0x7a, 0x73,
		0x71, 0x74, 0x72, 0x77, 0x78, 0x72, 0x75, 0x7c,
		0x7b, 0x71, 0x7c, 0x7a, 0x74, 0x7e, 0x7f, 0x79,
		0x7a, 0x73, 0x76, 0x7f, 0x79, 0x70, 0x70, 0x76,
		0x7c, 0x7a, 0x7b, 0x71, 0x77, 0x7d, 0x7d, 0x78,
		0x7f, 0x79, 0x71, 0x74, 0x73, 0x75, 0x7e, 0x7b,
		0x75, 0x7c, 0x72, 0x77, 0x78, 0x72, 0x74, 0x7e,
		0xb7, 0xbd, 0xbd, 0xb8, 0xbe, 0xbb, 0xb3, 0xb5,
		0xb0, 0xb6, 0xb6, 0xbf, 0xb9, 0xb0, 0xba, 0xb3,
		0xb1, 0xb4, 0xb2, 0xb7, 0xb8, 0xb2, 0xb5, 0xbc,
		0xbb, 0xb1, 0xbc, 0xba, 0xb4, 0xbe, 0xbf, 0xb9,
		0xba, 0xb3, 0xb6, 0xbf, 0xb9, 0xb0, 0xb0, 0xb6,
		0xbc, 0xba, 0xbb, 0xb1, 0xb7, 0xbd, 0xbd, 0xb8,
		0xbf, 0xb9, 0xb1, 0xb4, 0xb3, 0xb5, 0xbe, 0xbb,
		0xb5, 0xbc, 0xb2, 0xb7, 0xb8, 0xb2, 0xb4, 0xbe,
		0x47, 0x4d, 0x4d, 0x48, 0x4e, 0x4b, 0x43, 0x45,
		0x40, 0x46, 0x46, 0x4f, 0x49, 0x40, 0x4a, 0x43,
		0x41, 0x44, 0x42, 0x47, 0x48, 0x42, 0x45, 0x4c,
		0x4b, 0x41, 0x4c, 0x4a, 0x44, 0x4e, 0x4f, 0x49,
		0x4a, 0x43, 0x46, 0x4f, 0x49, 0x40, 0x40, 0x46,
		0x4c, 0x4a, 0x4b, 0x41, 0x47, 0x4d, 0x4d, 0x48,
		0x4f, 0x49, 0x41, 0x44, 0x43, 0x45, 0x4e, 0x4b,
		0x45, 0x4c, 0x42, 0x47, 0x48, 0x42, 0x44, 0x4e,
		0x17, 0x1d, 0x1d, 0x18, 0x1e, 0x1b, 0x13, 0x15,
		0x10, 0x16, 0x16, 0x1f, 0x19, 0x10, 0x1a, 0x13,
		0x11, 0x14, 0x12, 0x17, 0x18, 0x12, 0x15, 0x1c,
		0x1b, 0x11, 0x1c, 0x1a, 0x14, 0x1e, 0x1f, 0x19,
		0x1a, 0x13, 0x16, 0x1f, 0x19, 0x10, 0x10, 0x16,
		0x1c, 0x1a, 0x1b, 0x11, 0x17, 0x1d, 0x1d, 0x18,
		0x1f, 0x19, 0x11, 0x14, 0x13, 0x15, 0x1e, 0x1b,
		0x15, 0x1c, 0x12, 0x17, 0x18, 0x12, 0x14, 0x1e,
		0xf7, 0xfd, 0xfd, 0xf8, 0xfe, 0xfb, 0xf3, 0xf5,
		0xf0, 0xf6, 0xf6, 0xff, 0xf9, 0xf0, 0xfa, 0xf3,
		0xf1, 0xf4, 0xf2, 0xf7, 0xf8, 0xf2, 0xf5, 0xfc,
		0xfb, 0xf1, 0xfc, 0xfa, 0xf4, 0xfe, 0xff, 0xf9,
		0xfa, 0xf3, 0xf6, 0xff, 0xf9, 0xf0, 0xf0, 0xf6,
		0xfc, 0xfa, 0xfb, 0xf1, 0xf7, 0xfd, 0xfd, 0xf8,
		0xff, 0xf9, 0xf1, 0xf4, 0xf3, 0xf5, 0xfe, 0xfb,
		0xf5, 0xfc, 0xf2, 0xf7, 0xf8, 0xf2, 0xf4, 0xfe,
		0x27, 0x2d, 0x2d, 0x28, 0x2e, 0x2b, 0x23, 0x25,
		0x20, 0x26, 0x26, 0x2f, 0x29, 0x20, 0x2a, 0x23,
		0x21, 0x24, 0x22, 0x27, 0x28, 0x22, 0x25, 0x2c,
		0x2b, 0x21, 0x2c, 0x2a, 0x24, 0x2e, 0x2f, 0x29,
		0x2a, 0x23, 0x26, 0x2f, 0x29, 0x20, 0x20, 0x26,
		0x2c, 0x2a, 0x2b, 0x21, 0x27, 0x2d, 0x2d, 0x28,
		0x2f, 0x29, 0x21, 0x24, 0x23, 0x25, 0x2e, 0x2b,
		0x25, 0x2c, 0x22, 0x27, 0x28, 0x22, 0x24, 0x2e,
		0xe7, 0xed, 0xed, 0xe8, 0xee, 0xeb, 0xe3, 0xe5,
		0xe0, 0xe6, 0xe6, 0xef, 0xe9, 0xe0, 0xea, 0xe3,
		0xe1, 0xe4, 0xe2, 0xe7, 0xe8, 0xe2, 0xe5, 0xec,
		0xeb, 0xe1, 0xec, 0xea, 0xe4, 0xee, 0xef, 0xe9,
		0xea, 0xe3, 0xe6, 0xef, 0xe9, 0xe0, 0xe0, 0xe6,
		0xec, 0xea, 0xeb, 0xe1, 0xe7, 0xed, 0xed, 0xe8,
		0xef, 0xe9, 0xe1, 0xe4, 0xe3, 0xe5, 0xee, 0xeb,
		0xe5, 0xec, 0xe2, 0xe7, 0xe8, 0xe2, 0xe4, 0xee,
		0xc7, 0xcd, 0xcd, 0xc8, 0xce, 0xcb, 0xc3, 0xc5,
		0xc0, 0xc6, 0xc6, 0xcf, 0xc9, 0xc0, 0xca, 0xc3,
		0xc1, 0xc4, 0xc2, 0xc7, 0xc8, 0xc2, 0xc5, 0xcc,
		0xcb, 0xc1, 0xcc, 0xca, 0xc4, 0xce, 0xcf, 0xc9,
		0xca, 0xc3, 0xc6, 0xcf, 0xc9, 0xc0, 0xc0, 0xc6,
		0xcc, 0xca, 0xcb, 0xc1, 0xc7, 0xcd, 0xcd, 0xc8,
		0xcf, 0xc9, 0xc1, 0xc4, 0xc3, 0xc5, 0xce, 0xcb,
		0xc5, 0xcc, 0xc2, 0xc7, 0xc8, 0xc2, 0xc4, 0xce,
		0x37, 0x3d, 0x3d, 0x38, 0x3e, 0x3b, 0x33, 0x35,
		0x30, 0x36, 0x36, 0x3f, 0x39, 0x30, 0x3a, 0x33,
		0x31, 0x34, 0x32, 0x37, 0x38, 0x32, 0x35, 0x3c,
		0x3b, 0x31, 0x3c, 0x3a, 0x34, 0x3e, 0x3f, 0x39,
		0x3a, 0x33, 0x36, 0x3f, 0x39, 0x30, 0x30, 0x36,
		0x3c, 0x3a, 0x3b, 0x31, 0x37, 0x3d, 0x3d, 0x38,
		0x3f, 0x39, 0x31, 0x34, 0x33, 0x35, 0x3e, 0x3b,
		0x35, 0x3c, 0x32, 0x37, 0x38, 0x32, 0x34, 0x3e,
		0x57, 0x5d, 0x5d, 0x58, 0x5e, 0x5b, 0x53, 0x55,
		0x50, 0x56, 0x56, 0x5f, 0x59, 0x50, 0x5a, 0x53,
		0x51, 0x54, 0x52, 0x57, 0x58, 0x52, 0x55, 0x5c,
		0x5b, 0x51, 0x5c, 0x5a, 0x54, 0x5e, 0x5f, 0x59,
		0x5a, 0x53, 0x56, 0x5f, 0x59, 0x50, 0x50, 0x56,
		0x5c, 0x5a, 0x5b, 0x51, 0x57, 0x5d, 0x5d, 0x58,
		0x5f, 0x59, 0x51, 0x54, 0x53, 0x55, 0x5e, 0x5b,
		0x55, 0x5c, 0x52, 0x57, 0x58, 0x52, 0x54, 0x5e,
		0xb7, 0xbd, 0xbd, 0xb8, 0xbe, 0xbb, 0xb3, 0xb5,
		0xb0, 0xb6, 0xb6, 0xbf, 0xb9, 0xb0, 0xba, 0xb3,
		0xb1, 0xb4, 0xb2, 0xb7, 0xb8, 0xb2, 0xb5, 0xbc,
		0xbb, 0xb1, 0xbc, 0xba, 0xb4, 0xbe, 0xbf, 0xb9,
		0xba, 0xb3, 0xb6, 0xbf, 0xb9, 0xb0, 0xb0, 0xb6,
		0xbc, 0xba, 0xbb, 0xb1, 0xb7, 0xbd, 0xbd, 0xb8,
		0xbf, 0xb9, 0xb1, 0xb4, 0xb3, 0xb5, 0xbe, 0xbb,
		0xb5, 0xbc, 0xb2, 0xb7, 0xb8, 0xb2, 0xb4, 0xbe,
		0xa7, 0xad, 0xad, 0xa8, 0xae, 0xab, 0xa3, 0xa5,
		0xa0, 0xa6, 0xa6, 0xaf, 0xa9, 0xa0, 0xaa, 0xa3,
		0xa1, 0xa4, 0xa2, 0xa7, 0xa8, 0xa2, 0xa5, 0xac,
		0xab, 0xa1, 0xac, 0xaa, 0xa4, 0xae, 0xaf, 0xa9,
		0xaa, 0xa3, 0xa6, 0xaf, 0xa9, 0xa0, 0xa0, 0xa6,
		0xac, 0xaa, 0xab, 0xa1, 0xa7, 0xad, 0xad, 0xa8,
		0xaf, 0xa9, 0xa1, 0xa4, 0xa3, 0xa5, 0xae, 0xab,
		0xa5, 0xac, 0xa2, 0xa7, 0xa8, 0xa2, 0xa4, 0xae,
		0x57, 0x5d, 0x5d, 0x58, 0x5e, 0x5b, 0x53, 0x55,
		0x50, 0x56, 0x56, 0x5f, 0x59, 0x50, 0x5a, 0x53,
		0x51, 0x54, 0x52, 0x57, 0x58, 0x52, 0x55, 0x5c,
		0x5b, 0x51, 0x5c, 0x5a, 0x54, 0x5e, 0x5f, 0x59,
		0x5a, 0x53, 0x56, 0x5f, 0x59, 0x50, 0x50, 0x56,
		0x5c, 0x5a, 0x5b, 0x51, 0x57, 0x5d, 0x5d, 0x58,
		0x5f, 0x59, 0x51, 0x54, 0x53, 0x55, 0x5e, 0x5b,
		0x55, 0x5c, 0x52, 0x57, 0x58, 0x52, 0x54, 0x5e,
		0xe7, 0xed, 0xed, 0xe8, 0xee, 0xeb, 0xe3, 0xe5,
		0xe0, 0xe6, 0xe6, 0xef, 0xe9, 0xe0, 0xea, 0xe3,
		0xe1, 0xe4, 0xe2, 0xe7, 0xe8, 0xe2, 0xe5, 0xec,
		0xeb, 0xe1, 0xec, 0xea, 0xe4, 0xee, 0xef, 0xe9,
		0xea, 0xe3, 0xe6, 0xef, 0xe9, 0xe0, 0xe0, 0xe6,
		0xec, 0xea, 0xeb, 0xe1, 0xe7, 0xed, 0xed, 0xe8,
		0xef, 0xe9, 0xe1, 0xe4, 0xe3, 0xe5, 0xee, 0xeb,
		0xe5, 0xec, 0xe2, 0xe7, 0xe8, 0xe2, 0xe4, 0xee,
		0x27, 0x2d, 0x2d, 0x28, 0x2e, 0x2b, 0x23, 0x25,
		0x20, 0x26, 0x26, 0x2f, 0x29, 0x20, 0x2a, 0x23,
		0x21, 0x24, 0x22, 0x27, 0x28, 0x22, 0x25, 0x2c,
		0x2b, 0x21, 0x2c, 0x2a, 0x24, 0x2e, 0x2f, 0x29,
		0x2a, 0x23, 0x26, 0x2f, 0x29, 0x20, 0x20, 0x26,
		0x2c, 0x2a, 0x2b, 0x21, 0x27, 0x2d, 0x2d, 0x28,
		0x2f, 0x29, 0x21, 0x24, 0x23, 0x25, 0x2e, 0x2b,
		0x25, 0x2c, 0x22, 0x27, 0x28, 0x22, 0x24, 0x2e,
		0x77, 0x7d, 0x7d, 0x78, 0x7e, 0x7b, 0x73, 0x75,
		0x70, 0x76, 0x76, 0x7f, 0x79, 0x70, 0x7a, 0x73,
		0x71, 0x74, 0x72, 0x77, 0x78, 0x72, 0x75, 0x7c,
		0x7b, 0x71, 0x7c, 0x7a, 0x74, 0x7e, 0x7f, 0x79,
		0x7a, 0x73, 0x76, 0x7f, 0x79, 0x70, 0x70, 0x76,
		0x7c, 0x7a, 0x7b, 0x71, 0x77, 0x7d, 0x7d, 0x78,
		0x7f, 0x79, 0x71, 0x74, 0x73, 0x75, 0x7e, 0x7b,
		0x75, 0x7c, 0x72, 0x77, 0x78, 0x72, 0x74, 0x7e,
		0xc7, 0xcd, 0xcd, 0xc8, 0xce, 0xcb, 0xc3, 0xc5,
		0xc0, 0xc6, 0xc6, 0xcf, 0xc9, 0xc0, 0xca, 0xc3,
		0xc1, 0xc4, 0xc2, 0xc7, 0xc8, 0xc2, 0xc5, 0xcc,
		0xcb, 0xc1, 0xcc, 0xca, 0xc4, 0xce, 0xcf, 0xc9,
		0xca, 0xc3, 0xc6, 0xcf, 0xc9, 0xc0, 0xc0, 0xc6,
		0xcc, 0xca, 0xcb, 0xc1, 0xc7, 0xcd, 0xcd, 0xc8,
		0xcf, 0xc9, 0xc1, 0xc4, 0xc3, 0xc5, 0xce, 0xcb,
		0xc5, 0xcc, 0xc2, 0xc7, 0xc8, 0xc2, 0xc4, 0xce,
	},{
		0x2c, 0x2a, 0x21, 0x2f, 0x2a, 0x24, 0x2f, 0x22,
		0x29, 0x27, 0x22, 0x2c, 0x26, 0x29, 0x28, 0x25,
		0x20, 0x26, 0x2d, 0x21, 0x23, 0x2d, 0x24, 0x2e,
		0x2e, 0x20, 0x27, 0x2b, 0x25, 0x23, 0x2b, 0x28,
		0x29, 0x24, 0x2e, 0x23, 0x2f, 0x22, 0x25, 0x2c,
		0x22, 0x29, 0x28, 0x25, 0x2c, 0x2f, 0x23, 0x2a,
		0x27, 0x2b, 0x20, 0x2e, 0x24, 0x21, 0x2a, 0x27,
		0x21, 0x26, 0x2d, 0x20, 0x2b, 0x28, 0x26, 0x2d,
		0xec, 0xea, 0xe1, 0xef, 0xea, 0xe4, 0xef, 0xe2,
		0xe9, 0xe7, 0xe2, 0xec, 0xe6, 0xe9, 0xe8, 0xe5,
		0xe0, 0xe6, 0xed, 0xe1, 0xe3, 0xed, 0xe4, 0xee,
		0xee, 0xe0, 0xe7, 0xeb, 0xe5, 0xe3, 0xeb, 0xe8,
		0xe9, 0xe4, 0xee, 0xe3, 0xef, 0xe2, 0xe5, 0xec,
		0xe2, 0xe9, 0xe8, 0xe5, 0xec, 0xef, 0xe3, 0xea,
		0xe7, 0xeb, 0xe0, 0xee, 0xe4, 0xe1, 0xea, 0xe7,
		0xe1, 0xe6, 0xed, 0xe0, 0xeb, 0xe8, 0xe6, 0xed,
		0xcc, 0xca, 0xc1, 0xcf, 0xca, 0xc4, 0xcf, 0xc2,
		0xc9, 0xc7, 0xc2, 0xcc, 0xc6, 0xc9, 0xc8, 0xc5,
		0xc0, 0xc6, 0xcd, 0xc1, 0xc3, 0xcd, 0xc4, 0xce,
		0xce, 0xc0, 0xc7, 0xcb, 0xc5, 0xc3, 0xcb, 0xc8,
		0xc9, 0xc4, 0xce, 0xc3, 0xcf, 0xc2, 0xc5, 0xcc,
		0xc2, 0xc9, 0xc8, 0xc5, 0xcc, 0xcf, 0xc3, 0xca,
		0xc7, 0xcb, 0xc0, 0xce, 0xc4, 0xc1, 0xca, 0xc7,
		0xc1, 0xc6, 0xcd, 0xc0, 0xcb, 0xc8, 0xc6, 0xcd,
		0xbc, 0xba, 0xb1, 0xbf, 0xba, 0xb4, 0xbf, 0xb2,
		0xb9, 0xb7, 0xb2, 0xbc, 0xb6, 0xb9, 0xb8, 0xb5,
		0xb0, 0xb6, 0xbd, 0xb1, 0xb3, 0xbd, 0xb4, 0xbe,
		0xbe, 0xb0, 0xb7, 0xbb, 0xb5, 0xb3, 0xbb, 0xb8,
		0xb9, 0xb4, 0xbe, 0xb3, 0xbf, 0xb2, 0xb5, 0xbc,
		0xb2, 0xb9, 0xb8, 0xb5, 0xbc, 0xbf, 0xb3, 0xba,
		0xb7, 0xbb, 0xb0, 0xbe, 0xb4, 0xb1, 0xba, 0xb7,
		0xb1, 0xb6, 0xbd, 0xb0, 0xbb, 0xb8, 0xb6, 0xbd,
		0x4c, 0x4a, 0x41, 0x4f, 0x4a, 0x44, 0x4f, 0x42,
		0x49, 0x47, 0x42, 0x4c, 0x46, 0x49, 0x48, 0x45,
		0x40, 0x46, 0x4d, 0x41, 0x43, 0x4d, 0x44, 0x4e,
		0x4e, 0x40, 0x47, 0x4b, 0x45, 0x43, 0x4b, 0x48,
		0x49, 0x44, 0x4e, 0x43, 0x4f, 0x42, 0x45, 0x4c,
		0x42, 0x49, 0x48, 0x45, 0x4c, 0x4f, 0x43, 0x4a,
		0x47, 0x4b, 0x40, 0x4e, 0x44, 0x41, 0x4a, 0x47,
		0x41, 0x46, 0x4d, 0x40, 0x4b, 0x48, 0x46, 0x4d,
		0x2c, 0x2a, 0x21, 0x2f, 0x2a, 0x24, 0x2f, 0x22,
		0x29, 0x27, 0x22, 0x2c, 0x26, 0x29, 0x28, 0x25,
		0x20, 0x26, 0x2d, 0x21, 0x23, 0x2d, 0x24, 0x2e,
		0x2e, 0x20, 0x27, 0x2b, 0x25, 0x23, 0x2b, 0x28,
		0x29, 0x24, 0x2e, 0x23, 0x2f, 0x22, 0x25, 0x2c,
		0x22, 0x29, 0x28, 0x25, 0x2c, 0x2f, 0x23, 0x2a,
		0x27, 0x2b, 0x20, 0x2e, 0x24, 0x21, 0x2a, 0x27,
		0x21, 0x26, 0x2d, 0x20, 0x2b, 0x28, 0x26, 0x2d,
		0x1c, 0x1a, 0x11, 0x1f, 0x1a, 0x14, 0x1f, 0x12,
		0x19, 0x17, 0x12, 0x1c, 0x16, 0x19, 0x18, 0x15,
		0x10, 0x16, 0x1d, 0x11, 0x13, 0x1d, 0x14, 0x1e,
		0x1e, 0x10, 0x17, 0x1b, 0x15, 0x13, 0x1b, 0x18,
		0x19, 0x14, 0x1e, 0x13, 0x1f, 0x12, 0x15, 0x1c,
		0x12, 0x19, 0x18, 0x15, 0x1c, 0x1f, 0x13, 0x1a,
		0x17, 0x1b, 0x10, 0x1e, 0x14, 0x11, 0x1a, 0x17,
		0x11, 0x16, 0x1d, 0x10, 0x1b, 0x18, 0x16, 0x1d,
		0xcc, 0xca, 0xc1, 0xcf, 0xca, 0xc4, 0xcf, 0xc2,
		0xc9, 0xc7, 0xc2, 0xcc, 0xc6, 0xc9, 0xc8, 0xc5,
		0xc0, 0xc6, 0xcd, 0xc1, 0xc3, 0xcd, 0xc4, 0xce,
		0xce, 0xc0, 0xc7, 0xcb, 0xc5, 0xc3, 0xcb, 0xc8,
		0xc9, 0xc4, 0xce, 0xc3, 0xcf, 0xc2, 0xc5, 0xcc,
		0xc2, 0xc9, 0xc8, 0xc5, 0xcc, 0xcf, 0xc3, 0xca,
		0xc7, 0xcb, 0xc0, 0xce, 0xc4, 0xc1, 0xca, 0xc7,
		0xc1, 0xc6, 0xcd, 0xc0, 0xcb, 0xc8, 0xc6, 0xcd,
		0x7c, 0x7a, 0x71, 0x7f, 0x7a, 0x74, 0x7f, 0x72,
		0x79, 0x77, 0x72, 0x7c, 0x76, 0x79, 0x78, 0x75,
		0x70, 0x76, 0x7d, 0x71, 0x73, 0x7d, 0x74, 0x7e,
		0x7e, 0x70, 0x77, 0x7b, 0x75, 0x73, 0x7b, 0x78,
		0x79, 0x74, 0x7e, 0x73, 0x7f, 0x72, 0x75, 0x7c,
		0x72, 0x79, 0x78, 0x75, 0x7c, 0x7f, 0x73, 0x7a,
		0x77, 0x7b, 0x70, 0x7e, 0x74, 0x71, 0x7a, 0x77,
		0x71, 0x76, 0x7d, 0x70, 0x7b, 0x78, 0x76, 0x7d,
		0x4c, 0x4a, 0x41, 0x4f, 0x4a, 0x44, 0x4f, 0x42,
		0x49, 0x47, 0x42, 0x4c, 0x46, 0x49, 0x48, 0x45,
		0x40, 0x46, 0x4d, 0x41, 0x43, 0x4d, 0x44, 0x4e,
		0x4e, 0x40, 0x47, 0x4b, 0x45, 0x43, 0x4b, 0x48,
		0x49, 0x44, 0x4e, 0x43, 0x4f, 0x42, 0x45, 0x4c,
		0x42, 0x49, 0x48, 0x45, 0x4c, 0x4f, 0x43, 0x4a,
		0x47, 0x4b, 0x40, 0x4e, 0x44, 0x41, 0x4a, 0x47,
		0x41, 0x46, 0x4d, 0x40, 0x4b, 0x48, 0x46, 0x4d,
		0xac, 0xaa, 0xa1, 0xaf, 0xaa, 0xa4, 0xaf, 0xa2,
		0xa9, 0xa7, 0xa2, 0xac, 0xa6, 0xa9, 0xa8, 0xa5,
		0xa0, 0xa6, 0xad, 0xa1, 0xa3, 0xad, 0xa4, 0xae,
		0xae, 0xa0, 0xa7, 0xab, 0xa5, 0xa3, 0xab, 0xa8,
		0xa9, 0xa4, 0xae, 0xa3, 0xaf, 0xa2, 0xa5, 0xac,
		0xa2, 0xa9, 0xa8, 0xa5, 0xac, 0xaf, 0xa3, 0xaa,
		0xa7, 0xab, 0xa0, 0xae, 0xa4, 0xa1, 0xaa, 0xa7,
		0xa1, 0xa6, 0xad, 0xa0, 0xab, 0xa8, 0xa6, 0xad,
		0x7c, 0x7a, 0x71, 0x7f, 0x7a, 0x74, 0x7f, 0x72,
		0x79, 0x77, 0x72, 0x7c, 0x76, 0x79, 0x78, 0x75,
		0x70, 0x76, 0x7d, 0x71, 0x73, 0x7d, 0x74, 0x7e,
		0x7e, 0x70, 0x77, 0x7b, 0x75, 0x73, 0x7b, 0x78,
		0x79, 0x74, 0x7e, 0x73, 0x7f, 0x72, 0x75, 0x7c,
		0x72, 0x79, 0x78, 0x75, 0x7c, 0x7f, 0x73, 0x7a,
		0x77, 0x7b, 0x70, 0x7e, 0x74, 0x71, 0x7a, 0x77,
		0x71, 0x76, 0x7d, 0x70, 0x7b, 0x78, 0x76, 0x7d,
		0xbc, 0xba, 0xb1, 0xbf, 0xba, 0xb4, 0xbf, 0xb2,
		0xb9, 0xb7, 0xb2, 0xbc, 0xb6, 0xb9, 0xb8, 0xb5,
		0xb0, 0xb6, 0xbd, 0xb1, 0xb3, 0xbd, 0xb4, 0xbe,
		0xbe, 0xb0, 0xb7, 0xbb, 0xb5, 0xb3, 0xbb, 0xb8,
		0xb9, 0xb4, 0xbe, 0xb3, 0xbf, 0xb2, 0xb5, 0xbc,
		0xb2, 0xb9, 0xb8, 0xb5, 0xbc, 0xbf, 0xb3, 0xba,
		0xb7, 0xbb, 0xb0, 0xbe, 0xb4, 0xb1, 0xba, 0xb7,
		0xb1, 0xb6, 0xbd, 0xb0, 0xbb, 0xb8, 0xb6, 0xbd,
		0xdc, 0xda, 0xd1, 0xdf, 0xda, 0xd4, 0xdf, 0xd2,
		0xd9, 0xd7, 0xd2, 0xdc, 0xd6, 0xd9, 0xd8, 0xd5,
		0xd0, 0xd6, 0xdd, 0xd1, 0xd3, 0xdd, 0xd4, 0xde,
		0xde, 0xd0, 0xd7, 0xdb, 0xd5, 0xd3, 0xdb, 0xd8,
		0xd9, 0xd4, 0xde, 0xd3, 0xdf, 0xd2, 0xd5, 0xdc,
		0xd2, 0xd9, 0xd8, 0xd5, 0xdc, 0xdf, 0xd3, 0xda,
		0xd7, 0xdb, 0xd0, 0xde, 0xd4, 0xd1, 0xda, 0xd7,
		0xd1, 0xd6, 0xdd, 0xd0, 0xdb, 0xd8, 0xd6, 0xdd,
		0x6c, 0x6a, 0x61, 0x6f, 0x6a, 0x64, 0x6f, 0x62,
		0x69, 0x67, 0x62, 0x6c, 0x66, 0x69, 0x68, 0x65,
		0x60, 0x66, 0x6d, 0x61, 0x63, 0x6d, 0x64, 0x6e,
		0x6e, 0x60, 0x67, 0x6b, 0x65, 0x63, 0x6b, 0x68,
		0x69, 0x64, 0x6e, 0x63, 0x6f, 0x62, 0x65, 0x6c,
		0x62, 0x69, 0x68, 0x65, 0x6c, 0x6f, 0x63, 0x6a,
		0x67, 0x6b, 0x60, 0x6e, 0x64, 0x61, 0x6a, 0x67,
		0x61, 0x66, 0x6d, 0x60, 0x6b, 0x68, 0x66, 0x6d,
		0x1c, 0x1a, 0x11, 0x1f, 0x1a, 0x14, 0x1f, 0x12,
		0x19, 0x17, 0x12, 0x1c, 0x16, 0x19, 0x18, 0x15,
		0x10, 0x16, 0x1d, 0x11, 0x13, 0x1d, 0x14, 0x1e,
		0x1e, 0x10, 0x17, 0x1b, 0x15, 0x13, 0x1b, 0x18,
		0x19, 0x14, 0x1e, 0x13, 0x1f, 0x12, 0x15, 0x1c,
		0x12, 0x19, 0x18, 0x15, 0x1c, 0x1f, 0x13, 0x1a,
		0x17, 0x1b, 0x10, 0x1e, 0x14, 0x11, 0x1a, 0x17,
		0x11, 0x16, 0x1d, 0x10, 0x1b, 0x18, 0x16, 0x1d,
		0x8c, 0x8a, 0x81, 0x8f, 0x8a, 0x84, 0x8f, 0x82,
		0x89, 0x87, 0x82, 0x8c, 0x86, 0x89, 0x88, 0x85,
		0x80, 0x86, 0x8d, 0x81, 0x83, 0x8d, 0x84, 0x8e,
		0x8e, 0x80, 0x87, 0x8b, 0x85, 0x83, 0x8b, 0x88,
		0x89, 0x84, 0x8e, 0x83, 0x8f, 0x82, 0x85, 0x8c,
		0x82, 0x89, 0x88, 0x85, 0x8c, 0x8f, 0x83, 0x8a,
		0x87, 0x8b, 0x80, 0x8e, 0x84, 0x81, 0x8a, 0x87,
		0x81, 0x86, 0x8d, 0x80, 0x8b, 0x88, 0x86, 0x8d,
		0x5c, 0x5a, 0x51, 0x5f, 0x5a, 0x54, 0x5f, 0x52,
		0x59, 0x57, 0x52, 0x5c, 0x56, 0x59, 0x58, 0x55,
		0x50, 0x56, 0x5d, 0x51, 0x53, 0x5d, 0x54, 0x5e,
		0x5e, 0x50, 0x57, 0x5b, 0x55, 0x53, 0x5b, 0x58,
		0x59, 0x54, 0x5e, 0x53, 0x5f, 0x52, 0x55, 0x5c,
		0x52, 0x59, 0x58, 0x55, 0x5c, 0x5f, 0x53, 0x5a,
		0x57, 0x5b, 0x50, 0x5e, 0x54, 0x51, 0x5a, 0x57,
		0x51, 0x56, 0x5d, 0x50, 0x5b, 0x58, 0x56, 0x5d,
		0x5c, 0x5a, 0x51, 0x5f, 0x5a, 0x54, 0x5f, 0x52,
		0x59, 0x57, 0x52, 0x5c, 0x56, 0x59, 0x58, 0x55,
		0x50, 0x56, 0x5d, 0x51, 0x53, 0x5d, 0x54, 0x5e,
		0x5e, 0x50, 0x57, 0x5b, 0x55, 0x53, 0x5b, 0x58,
		0x59, 0x54, 0x5e, 0x53, 0x5f, 0x52, 0x55, 0x5c,
		0x52, 0x59, 0x58, 0x55, 0x5c, 0x5f, 0x53, 0x5a,
		0x57, 0x5b, 0x50, 0x5e, 0x54, 0x51, 0x5a, 0x57,
		0x51, 0x56, 0x5d, 0x50, 0x5b, 0x58, 0x56, 0x5d,
		0x0c, 0x0a, 0x01, 0x0f, 0x0a, 0x04, 0x0f, 0x02,
		0x09, 0x07, 0x02, 0x0c, 0x06, 0x09, 0x08, 0x05,
		0x00, 0x06, 0x0d, 0x01, 0x03, 0x0d, 0x04, 0x0e,
		0x0e, 0x00, 0x07, 0x0b, 0x05, 0x03, 0x0b, 0x08,
		0x09, 0x04, 0x0e, 0x03, 0x0f, 0x02, 0x05, 0x0c,
		0x02, 0x09, 0x08, 0x05, 0x0c, 0x0f, 0x03, 0x0a,
		0x07, 0x0b, 0x00, 0x0e, 0x04, 0x01, 0x0a, 0x07,
		0x01, 0x06, 0x0d, 0x00, 0x0b, 0x08, 0x06, 0x0d,
		0x3c, 0x3a, 0x31, 0x3f, 0x3a, 0x34, 0x3f, 0x32,
		0x39, 0x37, 0x32, 0x3c, 0x36, 0x39, 0x38, 0x35,
		0x30, 0x36, 0x3d, 0x31, 0x33, 0x3d, 0x34, 0x3e,
		0x3e, 0x30, 0x37, 0x3b, 0x35, 0x33, 0x3b, 0x38,
		0x39, 0x34, 0x3e, 0x33, 0x3f, 0x32, 0x35, 0x3c,
		0x32, 0x39, 0x38, 0x35, 0x3c, 0x3f, 0x33, 0x3a,
		0x37, 0x3b, 0x30, 0x3e, 0x34, 0x31, 0x3a, 0x37,
		0x31, 0x36, 0x3d, 0x30, 0x3b, 0x38, 0x36, 0x3d,
		0xfc, 0xfa, 0xf1, 0xff, 0xfa, 0xf4, 0xff, 0xf2,
		0xf9, 0xf7, 0xf2, 0xfc, 0xf6, 0xf9, 0xf8, 0xf5,
		0xf0, 0xf6, 0xfd, 0xf1, 0xf3, 0xfd, 0xf4, 0xfe,
		0xfe, 0xf0, 0xf7, 0xfb, 0xf5, 0xf3, 0xfb, 0xf8,
		0xf9, 0xf4, 0xfe, 0xf3, 0xff, 0xf2, 0xf5, 0xfc,
		0xf2, 0xf9, 0xf8, 0xf5, 0xfc, 0xff, 0xf3, 0xfa,
		0xf7, 0xfb, 0xf0, 0xfe, 0xf4, 0xf1, 0xfa, 0xf7,
		0xf1, 0xf6, 0xfd, 0xf0, 0xfb, 0xf8, 0xf6, 0xfd,
		0xfc, 0xfa, 0xf1, 0xff, 0xfa, 0xf4, 0xff, 0xf2,
		0xf9, 0xf7, 0xf2, 0xfc, 0xf6, 0xf9, 0xf8, 0xf5,
		0xf0, 0xf6, 0xfd, 0xf1, 0xf3, 0xfd, 0xf4, 0xfe,
		0xfe, 0xf0, 0xf7, 0xfb, 0xf5, 0xf3, 0xfb, 0xf8,
		0xf9, 0xf4, 0xfe, 0xf3, 0xff, 0xf2, 0xf5, 0xfc,
		0xf2, 0xf9, 0xf8, 0xf5, 0xfc, 0xff, 0xf3, 0xfa,
		0xf7, 0xfb, 0xf0, 0xfe, 0xf4, 0xf1, 0xfa, 0xf7,
		0xf1, 0xf6, 0xfd, 0xf0, 0xfb, 0xf8, 0xf6, 0xfd,
		0xac, 0xaa, 0xa1, 0xaf, 0xaa, 0xa4, 0xaf, 0xa2,
		0xa9, 0xa7, 0xa2, 0xac, 0xa6, 0xa9, 0xa8, 0xa5,
		0xa0, 0xa6, 0xad, 0xa1, 0xa3, 0xad, 0xa4, 0xae,
		0xae, 0xa0, 0xa7, 0xab, 0xa5, 0xa3, 0xab, 0xa8,
		0xa9, 0xa4, 0xae, 0xa3, 0xaf, 0xa2, 0xa5, 0xac,
		0xa2, 0xa9, 0xa8, 0xa5, 0xac, 0xaf, 0xa3, 0xaa,
		0xa7, 0xab, 0xa0, 0xae, 0xa4, 0xa1, 0xaa, 0xa7,
		0xa1, 0xa6, 0xad, 0xa0, 0xab, 0xa8, 0xa6, 0xad,
		0xdc, 0xda, 0xd1, 0xdf, 0xda, 0xd4, 0xdf, 0xd2,
		0xd9, 0xd7, 0xd2, 0xdc, 0xd6, 0xd9, 0xd8, 0xd5,
		0xd0, 0xd6, 0xdd, 0xd1, 0xd3, 0xdd, 0xd4, 0xde,
		0xde, 0xd0, 0xd7, 0xdb, 0xd5, 0xd3, 0xdb, 0xd8,
		0xd9, 0xd4, 0xde, 0xd3, 0xdf, 0xd2, 0xd5, 0xdc,
		0xd2, 0xd9, 0xd8, 0xd5, 0xdc, 0xdf, 0xd3, 0xda,
		0xd7, 0xdb, 0xd0, 0xde, 0xd4, 0xd1, 0xda, 0xd7,
		0xd1, 0xd6, 0xdd, 0xd0, 0xdb, 0xd8, 0xd6, 0xdd,
		0x3c, 0x3a, 0x31, 0x3f, 0x3a, 0x34, 0x3f, 0x32,
		0x39, 0x37, 0x32, 0x3c, 0x36, 0x39, 0x38, 0x35,
		0x30, 0x36, 0x3d, 0x31, 0x33, 0x3d, 0x34, 0x3e,
		0x3e, 0x30, 0x37, 0x3b, 0x35, 0x33, 0x3b, 0x38,
		0x39, 0x34, 0x3e, 0x33, 0x3f, 0x32, 0x35, 0x3c,
		0x32, 0x39, 0x38, 0x35, 0x3c, 0x3f, 0x33, 0x3a,
		0x37, 0x3b, 0x30, 0x3e, 0x34, 0x31, 0x3a, 0x37,
		0x31, 0x36, 0x3d, 0x30, 0x3b, 0x38, 0x36, 0x3d,
		0x0c, 0x0a, 0x01, 0x0f, 0x0a, 0x04, 0x0f, 0x02,
		0x09, 0x07, 0x02, 0x0c, 0x06, 0x09, 0x08, 0x05,
		0x00, 0x06, 0x0d, 0x01, 0x03, 0x0d, 0x04, 0x0e,
		0x0e, 0x00, 0x07, 0x0b, 0x05, 0x03, 0x0b, 0x08,
		0x09, 0x04, 0x0e, 0x03, 0x0f, 0x02, 0x05, 0x0c,
		0x02, 0x09, 0x08, 0x05, 0x0c, 0x0f, 0x03, 0x0a,
		0x07, 0x0b, 0x00, 0x0e, 0x04, 0x01, 0x0a, 0x07,
		0x01, 0x06, 0x0d, 0x00, 0x0b, 0x08, 0x06, 0x0d,
		0x9c, 0x9a, 0x91, 0x9f, 0x9a, 0x94, 0x9f, 0x92,
		0x99, 0x97, 0x92, 0x9c, 0x96, 0x99, 0x98, 0x95,
		0x90, 0x96, 0x9d, 0x91, 0x93, 0x9d, 0x94, 0x9e,
		0x9e, 0x90, 0x97, 0x9b, 0x95, 0x93, 0x9b, 0x98,
		0x99, 0x94, 0x9e, 0x93, 0x9f, 0x92, 0x95, 0x9c,
		0x92, 0x99, 0x98, 0x95, 0x9c, 0x9f, 0x93, 0x9a,
		0x97, 0x9b, 0x90, 0x9e, 0x94, 0x91, 0x9a, 0x97,
		0x91, 0x96, 0x9d, 0x90, 0x9b, 0x98, 0x96, 0x9d,
		0xec, 0xea, 0xe1, 0xef, 0xea, 0xe4, 0xef, 0xe2,
		0xe9, 0xe7, 0xe2, 0xec, 0xe6, 0xe9, 0xe8, 0xe5,
		0xe0, 0xe6, 0xed, 0xe1, 0xe3, 0xed, 0xe4, 0xee,
		0xee, 0xe0, 0xe7, 0xeb, 0xe5, 0xe3, 0xeb, 0xe8,
		0xe9, 0xe4, 0xee, 0xe3, 0xef, 0xe2, 0xe5, 0xec,
		0xe2, 0xe9, 0xe8, 0xe5, 0xec, 0xef, 0xe3, 0xea,
		0xe7, 0xeb, 0xe0, 0xee, 0xe4, 0xe1, 0xea, 0xe7,
		0xe1, 0xe6, 0xed, 0xe0, 0xeb, 0xe8, 0xe6, 0xed,
		0x8c, 0x8a, 0x81, 0x8f, 0x8a, 0x84, 0x8f, 0x82,
		0x89, 0x87, 0x82, 0x8c, 0x86, 0x89, 0x88, 0x85,
		0x80, 0x86, 0x8d, 0x81, 0x83, 0x8d, 0x84, 0x8e,
		0x8e, 0x80, 0x87, 0x8b, 0x85, 0x83, 0x8b, 0x88,
		0x89, 0x84, 0x8e, 0x83, 0x8f, 0x82, 0x85, 0x8c,
		0x82, 0x89, 0x88, 0x85, 0x8c, 0x8f, 0x83, 0x8a,
		0x87, 0x8b, 0x80, 0x8e, 0x84, 0x81, 0x8a, 0x87,
		0x81, 0x86, 0x8d, 0x80, 0x8b, 0x88, 0x86, 0x8d,
		0x9c, 0x9a, 0x91, 0x9f, 0x9a, 0x94, 0x9f, 0x92,
		0x99, 0x97, 0x92, 0x9c, 0x96, 0x99, 0x98, 0x95,
		0x90, 0x96, 0x9d, 0x91, 0x93, 0x9d, 0x94, 0x9e,
		0x9e, 0x90, 0x97, 0x9b, 0x95, 0x93, 0x9b, 0x98,
		0x99, 0x94, 0x9e, 0x93, 0x9f, 0x92, 0x95, 0x9c,
		0x92, 0x99, 0x98, 0x95, 0x9c, 0x9f, 0x93, 0x9a,
		0x97, 0x9b, 0x90, 0x9e, 0x94, 0x91, 0x9a, 0x97,
		0x91, 0x96, 0x9d, 0x90, 0x9b, 0x98, 0x96, 0x9d,
		0x6c, 0x6a, 0x61, 0x6f, 0x6a, 0x64, 0x6f, 0x62,
		0x69, 0x67, 0x62, 0x6c, 0x66, 0x69, 0x68, 0x65,
		0x60, 0x66, 0x6d, 0x61, 0x63, 0x6d, 0x64, 0x6e,
		0x6e, 0x60, 0x67, 0x6b, 0x65, 0x63, 0x6b, 0x68,
		0x69, 0x64, 0x6e, 0x63, 0x6f, 0x62, 0x65, 0x6c,
		0x62, 0x69, 0x68, 0x65, 0x6c, 0x6f, 0x63, 0x6a,
		0x67, 0x6b, 0x60, 0x6e, 0x64, 0x61, 0x6a, 0x67,
		0x61, 0x66, 0x6d, 0x60, 0x6b, 0x68, 0x66, 0x6d,
		0x4c, 0x4a, 0x41, 0x4f, 0x4a, 0x44, 0x4f, 0x42,
		0x49, 0x47, 0x42, 0x4c, 0x46, 0x49, 0x48, 0x45,
		0x40, 0x46, 0x4d, 0x41, 0x43, 0x4d, 0x44, 0x4e,
		0x4e, 0x40, 0x47, 0x4b, 0x45, 0x43, 0x4b, 0x48,
		0x49, 0x44, 0x4e, 0x43, 0x4f, 0x42, 0x45, 0x4c,
		0x42, 0x49, 0x48, 0x45, 0x4c, 0x4f, 0x43, 0x4a,
		0x47, 0x4b, 0x40, 0x4e, 0x44, 0x41, 0x4a, 0x47,
		0x41, 0x46, 0x4d, 0x40, 0x4b, 0x48, 0x46, 0x4d,
		0xbc, 0xba, 0xb1, 0xbf, 0xba, 0xb4, 0xbf, 0xb2,
		0xb9, 0xb7, 0xb2, 0xbc, 0xb6, 0xb9, 0xb8, 0xb5,
		0xb0, 0xb6, 0xbd, 0xb1, 0xb3, 0xbd, 0xb4, 0xbe,
		0xbe, 0xb0, 0xb7, 0xbb, 0xb5, 0xb3, 0xbb, 0xb8,
		0xb9, 0xb4, 0xbe, 0xb3, 0xbf, 0xb2, 0xb5, 0xbc,
		0xb2, 0xb9, 0xb8, 0xb5, 0xbc, 0xbf, 0xb3, 0xba,
		0xb7, 0xbb, 0xb0, 0xbe, 0xb4, 0xb1, 0xba, 0xb7,
		0xb1, 0xb6, 0xbd, 0xb0, 0xbb, 0xb8, 0xb6, 0xbd,
		0x2c, 0x2a, 0x21, 0x2f, 0x2a, 0x24, 0x2f, 0x22,
		0x29, 0x27, 0x22, 0x2c, 0x26, 0x29, 0x28, 0x25,
		0x20, 0x26, 0x2d, 0x21, 0x23, 0x2d, 0x24, 0x2e,
		0x2e, 0x20, 0x27, 0x2b, 0x25, 0x23, 0x2b, 0x28,
		0x29, 0x24, 0x2e, 0x23, 0x2f, 0x22, 0x25, 0x2c,
		0x22, 0x29, 0x28, 0x25, 0x2c, 0x2f, 0x23, 0x2a,
		0x27, 0x2b, 0x20, 0x2e, 0x24, 0x21, 0x2a, 0x27,
		0x21, 0x26, 0x2d, 0x20, 0x2b, 0x28, 0x26, 0x2d,
		0x8c, 0x8a, 0x81, 0x8f, 0x8a, 0x84, 0x8f, 0x82,
		0x89, 0x87, 0x82, 0x8c, 0x86, 0x89, 0x88, 0x85,
		0x80, 0x86, 0x8d, 0x81, 0x83, 0x8d, 0x84, 0x8e,
		0x8e, 0x80, 0x87, 0x8b, 0x85, 0x83, 0x8b, 0x88,
		0x89, 0x84, 0x8e, 0x83, 0x8f, 0x82, 0x85, 0x8c,
		0x82, 0x89, 0x88, 0x85, 0x8c, 0x8f, 0x83, 0x8a,
		0x87, 0x8b, 0x80, 0x8e, 0x84, 0x81, 0x8a, 0x87,
		0x81, 0x86, 0x8d, 0x80, 0x8b, 0x88, 0x86, 0x8d,
		0x1c, 0x1a, 0x11, 0x1f, 0x1a, 0x14, 0x1f, 0x12,
		0x19, 0x17, 0x12, 0x1c, 0x16, 0x19, 0x18, 0x15,
		0x10, 0x16, 0x1d, 0x11, 0x13, 0x1d, 0x14, 0x1e,
		0x1e, 0x10, 0x17, 0x1b, 0x15, 0x13, 0x1b, 0x18,
		0x19, 0x14, 0x1e, 0x13, 0x1f, 0x12, 0x15, 0x1c,
		0x12, 0x19, 0x18, 0x15, 0x1c, 0x1f, 0x13, 0x1a,
		0x17, 0x1b, 0x10, 0x1e, 0x14, 0x11, 0x1a, 0x17,
		0x11, 0x16, 0x1d, 0x10, 0x1b, 0x18, 0x16, 0x1d,
		0xcc, 0xca, 0xc1, 0xcf, 0xca, 0xc4, 0xcf, 0xc2,
		0xc9, 0xc7, 0xc2, 0xcc, 0xc6, 0xc9, 0xc8, 0xc5,
		0xc0, 0xc6, 0xcd, 0xc1, 0xc3, 0xcd, 0xc4, 0xce,
		0xce, 0xc0, 0xc7, 0xcb, 0xc5, 0xc3, 0xcb, 0xc8,
		0xc9, 0xc4, 0xce, 0xc3, 0xcf, 0xc2, 0xc5, 0xcc,
		0xc2, 0xc9, 0xc8, 0xc5, 0xcc, 0xcf, 0xc3, 0xca,
		0xc7, 0xcb, 0xc0, 0xce, 0xc4, 0xc1, 0xca, 0xc7,
		0xc1, 0xc6, 0xcd, 0xc0, 0xcb, 0xc8, 0xc6, 0xcd,
		0xbc, 0xba, 0xb1, 0xbf, 0xba, 0xb4, 0xbf, 0xb2,
		0xb9, 0xb7, 0xb2, 0xbc, 0xb6, 0xb9, 0xb8, 0xb5,
		0xb0, 0xb6, 0xbd, 0xb1, 0xb3, 0xbd, 0xb4, 0xbe,
		0xbe, 0xb0, 0xb7, 0xbb, 0xb5, 0xb3, 0xbb, 0xb8,
		0xb9, 0xb4, 0xbe, 0xb3, 0xbf, 0xb2, 0xb5, 0xbc,
		0xb2, 0xb9, 0xb8, 0xb5, 0xbc, 0xbf, 0xb3, 0xba,
		0xb7, 0xbb, 0xb0, 0xbe, 0xb4, 0xb1, 0xba, 0xb7,
		0xb1, 0xb6, 0xbd, 0xb0, 0xbb, 0xb8, 0xb6, 0xbd,
		0x7c, 0x7a, 0x71, 0x7f, 0x7a, 0x74, 0x7f, 0x72,
		0x79, 0x77, 0x72, 0x7c, 0x76, 0x79, 0x78, 0x75,
		0x70, 0x76, 0x7d, 0x71, 0x73, 0x7d, 0x74, 0x7e,
		0x7e, 0x70, 0x77, 0x7b, 0x75, 0x73, 0x7b, 0x78,
		0x79, 0x74, 0x7e, 0x73, 0x7f, 0x72, 0x75, 0x7c,
		0x72, 0x79, 0x78, 0x75, 0x7c, 0x7f, 0x73, 0x7a,
		0x77, 0x7b, 0x70, 0x7e, 0x74, 0x71, 0x7a, 0x77,
		0x71, 0x76, 0x7d, 0x70, 0x7b, 0x78, 0x76, 0x7d,
		0xac, 0xaa, 0xa1, 0xaf, 0xaa, 0xa4, 0xaf, 0xa2,
		0xa9, 0xa7, 0xa2, 0xac, 0xa6, 0xa9, 0xa8, 0xa5,
		0xa0, 0xa6, 0xad, 0xa1, 0xa3, 0xad, 0xa4, 0xae,
		0xae, 0xa0, 0xa7, 0xab, 0xa5, 0xa3, 0xab, 0xa8,
		0xa9, 0xa4, 0xae, 0xa3, 0xaf, 0xa2, 0xa5, 0xac,
		0xa2, 0xa9, 0xa8, 0xa5, 0xac, 0xaf, 0xa3, 0xaa,
		0xa7, 0xab, 0xa0, 0xae, 0xa4, 0xa1, 0xaa, 0xa7,
		0xa1, 0xa6, 0xad, 0xa0, 0xab, 0xa8, 0xa6, 0xad,
		0x1c, 0x1a, 0x11, 0x1f, 0x1a, 0x14, 0x1f, 0x12,
		0x19, 0x17, 0x12, 0x1c, 0x16, 0x19, 0x18, 0x15,
		0x10, 0x16, 0x1d, 0x11, 0x13, 0x1d, 0x14, 0x1e,
		0x1e, 0x10, 0x17, 0x1b, 0x15, 0x13, 0x1b, 0x18,
		0x19, 0x14, 0x1e, 0x13, 0x1f, 0x12, 0x15, 0x1c,
		0x12, 0x19, 0x18, 0x15, 0x1c, 0x1f, 0x13, 0x1a,
		0x17, 0x1b, 0x10, 0x1e, 0x14, 0x11, 0x1a, 0x17,
		0x11, 0x16, 0x1d, 0x10, 0x1b, 0x18, 0x16, 0x1d,
		0xdc, 0xda, 0xd1, 0xdf, 0xda, 0xd4, 0xdf, 0xd2,
		0xd9, 0xd7, 0xd2, 0xdc, 0xd6, 0xd9, 0xd8, 0xd5,
		0xd0, 0xd6, 0xdd, 0xd1, 0xd3, 0xdd, 0xd4, 0xde,
		0xde, 0xd0, 0xd7, 0xdb, 0xd5, 0xd3, 0xdb, 0xd8,
		0xd9, 0xd4, 0xde, 0xd3, 0xdf, 0xd2, 0xd5, 0xdc,
		0xd2, 0xd9, 0xd8, 0xd5, 0xdc, 0xdf, 0xd3, 0xda,
		0xd7, 0xdb, 0xd0, 0xde, 0xd4, 0xd1, 0xda, 0xd7,
		0xd1, 0xd6, 0xdd, 0xd0, 0xdb, 0xd8, 0xd6, 0xdd,
		0xec, 0xea, 0xe1, 0xef, 0xea, 0xe4, 0xef, 0xe2,
		0xe9, 0xe7, 0xe2, 0xec, 0xe6, 0xe9, 0xe8, 0xe5,
		0xe0, 0xe6, 0xed, 0xe1, 0xe3, 0xed, 0xe4, 0xee,
		0xee, 0xe0, 0xe7, 0xeb, 0xe5, 0xe3, 0xeb, 0xe8,
		0xe9, 0xe4, 0xee, 0xe3, 0xef, 0xe2, 0xe5, 0xec,
		0xe2, 0xe9, 0xe8, 0xe5, 0xec, 0xef, 0xe3, 0xea,
		0xe7, 0xeb, 0xe0, 0xee, 0xe4, 0xe1, 0xea, 0xe7,
		0xe1, 0xe6, 0xed, 0xe0, 0xeb, 0xe8, 0xe6, 0xed,
		0x7c, 0x7a, 0x71, 0x7f, 0x7a, 0x74, 0x7f, 0x72,
		0x79, 0x77, 0x72, 0x7c, 0x76, 0x79, 0x78, 0x75,
		0x70, 0x76, 0x7d, 0x71, 0x73, 0x7d, 0x74, 0x7e,
		0x7e, 0x70, 0x77, 0x7b, 0x75, 0x73, 0x7b, 0x78,
		0x79, 0x74, 0x7e, 0x73, 0x7f, 0x72, 0x75, 0x7c,
		0x72, 0x79, 0x78, 0x75, 0x7c, 0x7f, 0x73, 0x7a,
		0x77, 0x7b, 0x70, 0x7e, 0x74, 0x71, 0x7a, 0x77,
		0x71, 0x76, 0x7d, 0x70, 0x7b, 0x78, 0x76, 0x7d,
		0x2c, 0x2a, 0x21, 0x2f, 0x2a, 0x24, 0x2f, 0x22,
		0x29, 0x27, 0x22, 0x2c, 0x26, 0x29, 0x28, 0x25,
		0x20, 0x26, 0x2d, 0x21, 0x23, 0x2d, 0x24, 0x2e,
		0x2e, 0x20, 0x27, 0x2b, 0x25, 0x23, 0x2b, 0x28,
		0x29, 0x24, 0x2e, 0x23, 0x2f, 0x22, 0x25, 0x2c,
		0x22, 0x29, 0x28, 0x25, 0x2c, 0x2f, 0x23, 0x2a,
		0x27, 0x2b, 0x20, 0x2e, 0x24, 0x21, 0x2a, 0x27,
		0x21, 0x26, 0x2d, 0x20, 0x2b, 0x28, 0x26, 0x2d,
		0x8c, 0x8a, 0x81, 0x8f, 0x8a, 0x84, 0x8f, 0x82,
		0x89, 0x87, 0x82, 0x8c, 0x86, 0x89, 0x88, 0x85,
		0x80, 0x86, 0x8d, 0x81, 0x83, 0x8d, 0x84, 0x8e,
		0x8e, 0x80, 0x87, 0x8b, 0x85, 0x83, 0x8b, 0x88,
		0x89, 0x84, 0x8e, 0x83, 0x8f, 0x82, 0x85, 0x8c,
		0x82, 0x89, 0x88, 0x85, 0x8c, 0x8f, 0x83, 0x8a,
		0x87, 0x8b, 0x80, 0x8e, 0x84, 0x81, 0x8a, 0x87,
		0x81, 0x86, 0x8d, 0x80, 0x8b, 0x88, 0x86, 0x8d,
		0xdc, 0xda, 0xd1, 0xdf, 0xda, 0xd4, 0xdf, 0xd2,
		0xd9, 0xd7, 0xd2, 0xdc, 0xd6, 0xd9, 0xd8, 0xd5,
		0xd0, 0xd6, 0xdd, 0xd1, 0xd3, 0xdd, 0xd4, 0xde,
		0xde, 0xd0, 0xd7, 0xdb, 0xd5, 0xd3, 0xdb, 0xd8,
		0xd9, 0xd4, 0xde, 0xd3, 0xdf, 0xd2, 0xd5, 0xdc,
		0xd2, 0xd9, 0xd8, 0xd5, 0xdc, 0xdf, 0xd3, 0xda,
		0xd7, 0xdb, 0xd0, 0xde, 0xd4, 0xd1, 0xda, 0xd7,
		0xd1, 0xd6, 0xdd, 0xd0, 0xdb, 0xd8, 0xd6, 0xdd,
		0xfc, 0xfa, 0xf1, 0xff, 0xfa, 0xf4, 0xff, 0xf2,
		0xf9, 0xf7, 0xf2, 0xfc, 0xf6, 0xf9, 0xf8, 0xf5,
		0xf0, 0xf6, 0xfd, 0xf1, 0xf3, 0xfd, 0xf4, 0xfe,
		0xfe, 0xf0, 0xf7, 0xfb, 0xf5, 0xf3, 0xfb, 0xf8,
		0xf9, 0xf4, 0xfe, 0xf3, 0xff, 0xf2, 0xf5, 0xfc,
		0xf2, 0xf9, 0xf8, 0xf5, 0xfc, 0xff, 0xf3, 0xfa,
		0xf7, 0xfb, 0xf0, 0xfe, 0xf4, 0xf1, 0xfa, 0xf7,
		0xf1, 0xf6, 0xfd, 0xf0, 0xfb, 0xf8, 0xf6, 0xfd,
		0x6c, 0x6a, 0x61, 0x6f, 0x6a, 0x64, 0x6f, 0x62,
		0x69, 0x67, 0x62, 0x6c, 0x66, 0x69, 0x68, 0x65,
		0x60, 0x66, 0x6d, 0x61, 0x63, 0x6d, 0x64, 0x6e,
		0x6e, 0x60, 0x67, 0x6b, 0x65, 0x63, 0x6b, 0x68,
		0x69, 0x64, 0x6e, 0x63, 0x6f, 0x62, 0x65, 0x6c,
		0x62, 0x69, 0x68, 0x65, 0x6c, 0x6f, 0x63, 0x6a,
		0x67, 0x6b, 0x60, 0x6e, 0x64, 0x61, 0x6a, 0x67,
		0x61, 0x66, 0x6d, 0x60, 0x6b, 0x68, 0x66, 0x6d,
		0x9c, 0x9a, 0x91, 0x9f, 0x9a, 0x94, 0x9f, 0x92,
		0x99, 0x97, 0x92, 0x9c, 0x96, 0x99, 0x98, 0x95,
		0x90, 0x96, 0x9d, 0x91, 0x93, 0x9d, 0x94, 0x9e,
		0x9e, 0x90, 0x97, 0x9b, 0x95, 0x93, 0x9b, 0x98,
		0x99, 0x94, 0x9e, 0x93, 0x9f, 0x92, 0x95, 0x9c,
		0x92, 0x99, 0x98, 0x95, 0x9c, 0x9f, 0x93, 0x9a,
		0x97, 0x9b, 0x90, 0x9e, 0x94, 0x91, 0x9a, 0x97,
		0x91, 0x96, 0x9d, 0x90, 0x9b, 0x98, 0x96, 0x9d,
		0xfc, 0xfa, 0xf1, 0xff, 0xfa, 0xf4, 0xff, 0xf2,
		0xf9, 0xf7, 0xf2, 0xfc, 0xf6, 0xf9, 0xf8, 0xf5,
		0xf0, 0xf6, 0xfd, 0xf1, 0xf3, 0xfd, 0xf4, 0xfe,
		0xfe, 0xf0, 0xf7, 0xfb, 0xf5, 0xf3, 0xfb, 0xf8,
		0xf9, 0xf4, 0xfe, 0xf3, 0xff, 0xf2, 0xf5, 0xfc,
		0xf2, 0xf9, 0xf8, 0xf5, 0xfc, 0xff, 0xf3, 0xfa,
		0xf7, 0xfb, 0xf0, 0xfe, 0xf4, 0xf1, 0xfa, 0xf7,
		0xf1, 0xf6, 0xfd, 0xf0, 0xfb, 0xf8, 0xf6, 0xfd,
		0xcc, 0xca, 0xc1, 0xcf, 0xca, 0xc4, 0xcf, 0xc2,
		0xc9, 0xc7, 0xc2, 0xcc, 0xc6, 0xc9, 0xc8, 0xc5,
		0xc0, 0xc6, 0xcd, 0xc1, 0xc3, 0xcd, 0xc4, 0xce,
		0xce, 0xc0, 0xc7, 0xcb, 0xc5, 0xc3, 0xcb, 0xc8,
		0xc9, 0xc4, 0xce, 0xc3, 0xcf, 0xc2, 0xc5, 0xcc,
		0xc2, 0xc9, 0xc8, 0xc5, 0xcc, 0xcf, 0xc3, 0xca,
		0xc7, 0xcb, 0xc0, 0xce, 0xc4, 0xc1, 0xca, 0xc7,
		0xc1, 0xc6, 0xcd, 0xc0, 0xcb, 0xc8, 0xc6, 0xcd,
		0x0c, 0x0a, 0x01, 0x0f, 0x0a, 0x04, 0x0f, 0x02,
		0x09, 0x07, 0x02, 0x0c, 0x06, 0x09, 0x08, 0x05,
		0x00, 0x06, 0x0d, 0x01, 0x03, 0x0d, 0x04, 0x0e,
		0x0e, 0x00, 0x07, 0x0b, 0x05, 0x03, 0x0b, 0x08,
		0x09, 0x04, 0x0e, 0x03, 0x0f, 0x02, 0x05, 0x0c,
		0x02, 0x09, 0x08, 0x05, 0x0c, 0x0f, 0x03, 0x0a,
		0x07, 0x0b, 0x00, 0x0e, 0x04, 0x01, 0x0a, 0x07,
		0x01, 0x06, 0x0d, 0x00, 0x0b, 0x08, 0x06, 0x0d,
		0x5c, 0x5a, 0x51, 0x5f, 0x5a, 0x54, 0x5f, 0x52,
		0x59, 0x57, 0x52, 0x5c, 0x56, 0x59, 0x58, 0x55,
		0x50, 0x56, 0x5d, 0x51, 0x53, 0x5d, 0x54, 0x5e,
		0x5e, 0x50, 0x57, 0x5b, 0x55, 0x53, 0x5b, 0x58,
		0x59, 0x54, 0x5e, 0x53, 0x5f, 0x52, 0x55, 0x5c,
		0x52, 0x59, 0x58, 0x55, 0x5c, 0x5f, 0x53, 0x5a,
		0x57, 0x5b, 0x50, 0x5e, 0x54, 0x51, 0x5a, 0x57,
		0x51, 0x56, 0x5d, 0x50, 0x5b, 0x58, 0x56, 0x5d,
		0x9c, 0x9a, 0x91, 0x9f, 0x9a, 0x94, 0x9f, 0x92,
		0x99, 0x97, 0x92, 0x9c, 0x96, 0x99, 0x98, 0x95,
		0x90, 0x96, 0x9d, 0x91, 0x93, 0x9d, 0x94, 0x9e,
		0x9e, 0x90, 0x97, 0x9b, 0x95, 0x93, 0x9b, 0x98,
		0x99, 0x94, 0x9e, 0x93, 0x9f, 0x92, 0x95, 0x9c,
		0x92, 0x99, 0x98, 0x95, 0x9c, 0x9f, 0x93, 0x9a,
		0x97, 0x9b, 0x90, 0x9e, 0x94, 0x91, 0x9a, 0x97,
		0x91, 0x96, 0x9d, 0x90, 0x9b, 0x98, 0x96, 0x9d,
		0x6c, 0x6a, 0x61, 0x6f, 0x6a, 0x64, 0x6f, 0x62,
		0x69, 0x67, 0x62, 0x6c, 0x66, 0x69, 0x68, 0x65,
		0x60, 0x66, 0x6d, 0x61, 0x63, 0x6d, 0x64, 0x6e,
		0x6e, 0x60, 0x67, 0x6b, 0x65, 0x63, 0x6b, 0x68,
		0x69, 0x64, 0x6e, 0x63, 0x6f, 0x62, 0x65, 0x6c,
		0x62, 0x69, 0x68, 0x65, 0x6c, 0x6f, 0x63, 0x6a,
		0x67, 0x6b, 0x60, 0x6e, 0x64, 0x61, 0x6a, 0x67,
		0x61, 0x66, 0x6d, 0x60, 0x6b, 0x68, 0x66, 0x6d,
		0xac, 0xaa, 0xa1, 0xaf, 0xaa, 0xa4, 0xaf, 0xa2,
		0xa9, 0xa7, 0xa2, 0xac, 0xa6, 0xa9, 0xa8, 0xa5,
		0xa0, 0xa6, 0xad, 0xa1, 0xa3, 0xad, 0xa4, 0xae,
		0xae, 0xa0, 0xa7, 0xab, 0xa5, 0xa3, 0xab, 0xa8,
		0xa9, 0xa4, 0xae, 0xa3, 0xaf, 0xa2, 0xa5, 0xac,
		0xa2, 0xa9, 0xa8, 0xa5, 0xac, 0xaf, 0xa3, 0xaa,
		0xa7, 0xab, 0xa0, 0xae, 0xa4, 0xa1, 0xaa, 0xa7,
		0xa1, 0xa6, 0xad, 0xa0, 0xab, 0xa8, 0xa6, 0xad,
		0x3c, 0x3a, 0x31, 0x3f, 0x3a, 0x34, 0x3f, 0x32,
		0x39, 0x37, 0x32, 0x3c, 0x36, 0x39, 0x38, 0x35,
		0x30, 0x36, 0x3d, 0x31, 0x33, 0x3d, 0x34, 0x3e,
		0x3e, 0x30, 0x37, 0x3b, 0x35, 0x33, 0x3b, 0x38,
		0x39, 0x34, 0x3e, 0x33, 0x3f, 0x32, 0x35, 0x3c,
		0x32, 0x39, 0x38, 0x35, 0x3c, 0x3f, 0x33, 0x3a,
		0x37, 0x3b, 0x30, 0x3e, 0x34, 0x31, 0x3a, 0x37,
		0x31, 0x36, 0x3d, 0x30, 0x3b, 0x38, 0x36, 0x3d,
		0x4c, 0x4a, 0x41, 0x4f, 0x4a, 0x44, 0x4f, 0x42,
		0x49, 0x47, 0x42, 0x4c, 0x46, 0x49, 0x48, 0x45,
		0x40, 0x46, 0x4d, 0x41, 0x43, 0x4d, 0x44, 0x4e,
		0x4e, 0x40, 0x47, 0x4b, 0x45, 0x43, 0x4b, 0x48,
		0x49, 0x44, 0x4e, 0x43, 0x4f, 0x42, 0x45, 0x4c,
		0x42, 0x49, 0x48, 0x45, 0x4c, 0x4f, 0x43, 0x4a,
		0x47, 0x4b, 0x40, 0x4e, 0x44, 0x41, 0x4a, 0x47,
		0x41, 0x46, 0x4d, 0x40, 0x4b, 0x48, 0x46, 0x4d,
		0x0c, 0x0a, 0x01, 0x0f, 0x0a, 0x04, 0x0f, 0x02,
		0x09, 0x07, 0x02, 0x0c, 0x06, 0x09, 0x08, 0x05,
		0x00, 0x06, 0x0d, 0x01, 0x03, 0x0d, 0x04, 0x0e,
		0x0e, 0x00, 0x07, 0x0b, 0x05, 0x03, 0x0b, 0x08,
		0x09, 0x04, 0x0e, 0x03, 0x0f, 0x02, 0x05, 0x0c,
		0x02, 0x09, 0x08, 0x05, 0x0c, 0x0f, 0x03, 0x0a,
		0x07, 0x0b, 0x00, 0x0e, 0x04, 0x01, 0x0a, 0x07,
		0x01, 0x06, 0x0d, 0x00, 0x0b, 0x08, 0x06, 0x0d,
		0x5c, 0x5a, 0x51, 0x5f, 0x5a, 0x54, 0x5f, 0x52,
		0x59, 0x57, 0x52, 0x5c, 0x56, 0x59, 0x58, 0x55,
		0x50, 0x56, 0x5d, 0x51, 0x53, 0x5d, 0x54, 0x5e,
		0x5e, 0x50, 0x57, 0x5b, 0x55, 0x53, 0x5b, 0x58,
		0x59, 0x54, 0x5e, 0x53, 0x5f, 0x52, 0x55, 0x5c,
		0x52, 0x59, 0x58, 0x55, 0x5c, 0x5f, 0x53, 0x5a,
		0x57, 0x5b, 0x50, 0x5e, 0x54, 0x51, 0x5a, 0x57,
		0x51, 0x56, 0x5d, 0x50, 0x5b, 0x58, 0x56, 0x5d,
		0xec, 0xea, 0xe1, 0xef, 0xea, 0xe4, 0xef, 0xe2,
		0xe9, 0xe7, 0xe2, 0xec, 0xe6, 0xe9, 0xe8, 0xe5,
		0xe0, 0xe6, 0xed, 0xe1, 0xe3, 0xed, 0xe4, 0xee,
		0xee, 0xe0, 0xe7, 0xeb, 0xe5, 0xe3, 0xeb, 0xe8,
		0xe9, 0xe4, 0xee, 0xe3, 0xef, 0xe2, 0xe5, 0xec,
		0xe2, 0xe9, 0xe8, 0xe5, 0xec, 0xef, 0xe3, 0xea,
		0xe7, 0xeb, 0xe0, 0xee, 0xe4, 0xe1, 0xea, 0xe7,
		0xe1, 0xe6, 0xed, 0xe0, 0xeb, 0xe8, 0xe6, 0xed,
		0x3c, 0x3a, 0x31, 0x3f, 0x3a, 0x34, 0x3f, 0x32,
		0x39, 0x37, 0x32, 0x3c, 0x36, 0x39, 0x38, 0x35,
		0x30, 0x36, 0x3d, 0x31, 0x33, 0x3d, 0x34, 0x3e,
		0x3e, 0x30, 0x37, 0x3b, 0x35, 0x33, 0x3b, 0x38,
		0x39, 0x34, 0x3e, 0x33, 0x3f, 0x32, 0x35, 0x3c,
		0x32, 0x39, 0x38, 0x35, 0x3c, 0x3f, 0x33, 0x3a,
		0x37, 0x3b, 0x30, 0x3e, 0x34, 0x31, 0x3a, 0x37,
		0x31, 0x36, 0x3d, 0x30, 0x3b, 0x38, 0x36, 0x3d,
	},{
		0x4d, 0x41, 0x42, 0x4f, 0x48, 0x4d, 0x44, 0x48,
		0x46, 0x4a, 0x4f, 0x43, 0x4b, 0x47, 0x41, 0x44,
		0x4a, 0x4c, 0x49, 0x45, 0x43, 0x46, 0x4e, 0x4b,
		0x45, 0x40, 0x40, 0x4e, 0x4c, 0x49, 0x47, 0x42,
		0x47, 0x42, 0x4b, 0x41, 0x44, 0x4e, 0x41, 0x47,
		0x49, 0x44, 0x4c, 0x4a, 0x4e, 0x48, 0x42, 0x4d,
		0x40, 0x4f, 0x46, 0x4c, 0x4a, 0x49, 0x4d, 0x40,
		0x4f, 0x43, 0x43, 0x45, 0x45, 0x46, 0x48, 0x4b,
		0xdd, 0xd1, 0xd2, 0xdf, 0xd8, 0xdd, 0xd4, 0xd8,
		0xd6, 0xda, 0xdf, 0xd3, 0xdb, 0xd7, 0xd1, 0xd4,
		0xda, 0xdc, 0xd9, 0xd5, 0xd3, 0xd6, 0xde, 0xdb,
		0xd5, 0xd0, 0xd0, 0xde, 0xdc, 0xd9, 0xd7, 0xd2,
		0xd7, 0xd2, 0xdb, 0xd1, 0xd4, 0xde, 0xd1, 0xd7,
		0xd9, 0xd4, 0xdc, 0xda, 0xde, 0xd8, 0xd2, 0xdd,
		0xd0, 0xdf, 0xd6, 0xdc, 0xda, 0xd9, 0xdd, 0xd0,
		0xdf, 0xd3, 0xd3, 0xd5, 0xd5, 0xd6, 0xd8, 0xdb,
		0xbd, 0xb1, 0xb2, 0xbf, 0xb8, 0xbd, 0xb4, 0xb8,
		0xb6, 0xba, 0xbf, 0xb3, 0xbb, 0xb7, 0xb1, 0xb4,
		0xba, 0xbc, 0xb9, 0xb5, 0xb3, 0xb6, 0xbe, 0xbb,
		0xb5, 0xb0, 0xb0, 0xbe, 0xbc, 0xb9, 0xb7, 0xb2,
		0xb7, 0xb2, 0xbb, 0xb1, 0xb4, 0xbe, 0xb1, 0xb7,
		0xb9, 0xb4, 0xbc, 0xba, 0xbe, 0xb8, 0xb2, 0xbd,
		0xb0, 0xbf, 0xb6, 0xbc, 0xba, 0xb9, 0xbd, 0xb0,
		0xbf, 0xb3, 0xb3, 0xb5, 0xb5, 0xb6, 0xb8, 0xbb,
		0x0d, 0x01, 0x02, 0x0f, 0x08, 0x0d, 0x04, 0x08,
		0x06, 0x0a, 0x0f, 0x03, 0x0b, 0x07, 0x01, 0x04,
		0x0a, 0x0c, 0x09, 0x05, 0x03, 0x06, 0x0e, 0x0b,
		0x05, 0x00, 0x00, 0x0e, 0x0c, 0x09, 0x07, 0x02,
		0x07, 0x02, 0x0b, 0x01, 0x04, 0x0e, 0x01, 0x07,
		0x09, 0x04, 0x0c, 0x0a, 0x0e, 0x08, 0x02, 0x0d,
		0x00, 0x0f, 0x06, 0x0c, 0x0a, 0x09, 0x0d, 0x00,
		0x0f, 0x03, 0x03, 0x05, 0x05, 0x06, 0x08, 0x0b,
		0x2d, 0x21, 0x22, 0x2f, 0x28, 0x2d, 0x24, 0x28,
		0x26, 0x2a, 0x2f, 0x23, 0x2b, 0x27, 0x21, 0x24,
		0x2a, 0x2c, 0x29, 0x25, 0x23, 0x26, 0x2e, 0x2b,
		0x25, 0x20, 0x20, 0x2e, 0x2c, 0x29, 0x27, 0x22,
		0x27, 0x22, 0x2b, 0x21, 0x24, 0x2e, 0x21, 0x27,
		0x29, 0x24, 0x2c, 0x2a, 0x2e, 0x28, 0x22, 0x2d,
		0x20, 0x2f, 0x26, 0x2c, 0x2a, 0x29, 0x2d, 0x20,
		0x2f, 0x23, 0x23, 0x25, 0x25, 0x26, 0x28, 0x2b,
		0xbd, 0xb1, 0xb2, 0xbf, 0xb8, 0xbd, 0xb4, 0xb8,
		0xb6, 0xba, 0xbf, 0xb3, 0xbb, 0xb7, 0xb1, 0xb4,
		0xba, 0xbc, 0xb9, 0xb5, 0xb3, 0xb6, 0xbe, 0xbb,
		0xb5, 0xb0, 0xb0, 0xbe, 0xbc, 0xb9, 0xb7, 0xb2,
		0xb7, 0xb2, 0xbb, 0xb1, 0xb4, 0xbe, 0xb1, 0xb7,
		0xb9, 0xb4, 0xbc, 0xba, 0xbe, 0xb8, 0xb2, 0xbd,
		0xb0, 0xbf, 0xb6, 0xbc, 0xba, 0xb9, 0xbd, 0xb0,
		0xbf, 0xb3, 0xb3, 0xb5, 0xb5, 0xb6, 0xb8, 0xbb,
		0xed, 0xe1, 0xe2, 0xef, 0xe8, 0xed, 0xe4, 0xe8,
		0xe6, 0xea, 0xef, 0xe3, 0xeb, 0xe7, 0xe1, 0xe4,
		0xea, 0xec, 0xe9, 0xe5, 0xe3, 0xe6, 0xee, 0xeb,
		0xe5, 0xe0, 0xe0, 0xee, 0xec, 0xe9, 0xe7, 0xe2,
		0xe7, 0xe2, 0xeb, 0xe1, 0xe4, 0xee, 0xe1, 0xe7,
		0xe9, 0xe4, 0xec, 0xea, 0xee, 0xe8, 0xe2, 0xed,
		0xe0, 0xef, 0xe6, 0xec, 0xea, 0xe9, 0xed, 0xe0,
		0xef, 0xe3, 0xe3, 0xe5, 0xe5, 0xe6, 0xe8, 0xeb,
		0x7d, 0x71, 0x72, 0x7f, 0x78, 0x7d, 0x74, 0x78,
		0x76, 0x7a, 0x7f, 0x73, 0x7b, 0x77, 0x71, 0x74,
		0x7a, 0x7c, 0x79, 0x75, 0x73, 0x76, 0x7e, 0x7b,
		0x75, 0x70, 0x70, 0x7e, 0x7c, 0x79, 0x77, 0x72,
		0x77, 0x72, 0x7b, 0x71, 0x74, 0x7e, 0x71, 0x77,
		0x79, 0x74, 0x7c, 0x7a, 0x7e, 0x78, 0x72, 0x7d,
		0x70, 0x7f, 0x76, 0x7c, 0x7a, 0x79, 0x7d, 0x70,
		0x7f, 0x73, 0x73, 0x75, 0x75, 0x76, 0x78, 0x7b,
		0xfd, 0xf1, 0xf2, 0xff, 0xf8, 0xfd, 0xf4, 0xf8,
		0xf6, 0xfa, 0xff, 0xf3, 0xfb, 0xf7, 0xf1, 0xf4,
		0xfa, 0xfc, 0xf9, 0xf5, 0xf3, 0xf6, 0xfe, 0xfb,
		0xf5, 0xf0, 0xf0, 0xfe, 0xfc, 0xf9, 0xf7, 0xf2,
		0xf7, 0xf2, 0xfb, 0xf1, 0xf4, 0xfe, 0xf1, 0xf7,
		0xf9, 0xf4, 0xfc, 0xfa, 0xfe, 0xf8, 0xf2, 0xfd,
		0xf0, 0xff, 0xf6, 0xfc, 0xfa, 0xf9, 0xfd, 0xf0,
		0xff, 0xf3, 0xf3, 0xf5, 0xf5, 0xf6, 0xf8, 0xfb,
		0x4d, 0x41, 0x42, 0x4f, 0x48, 0x4d, 0x44, 0x48,
		0x46, 0x4a, 0x4f, 0x43, 0x4b, 0x47, 0x41, 0x44,
		0x4a, 0x4c, 0x49, 0x45, 0x43, 0x46, 0x4e, 0x4b,
		0x45, 0x40, 0x40, 0x4e, 0x4c, 0x49, 0x47, 0x42,
		0x47, 0x42, 0x4b, 0x41, 0x44, 0x4e, 0x41, 0x47,
		0x49, 0x44, 0x4c, 0x4a, 0x4e, 0x48, 0x42, 0x4d,
		0x40, 0x4f, 0x46, 0x4c, 0x4a, 0x49, 0x4d, 0x40,
		0x4f, 0x43, 0x43, 0x45, 0x45, 0x46, 0x48, 0x4b,
		0x0d, 0x01, 0x02, 0x0f, 0x08, 0x0d, 0x04, 0x08,
		0x06, 0x0a, 0x0f, 0x03, 0x0b, 0x07, 0x01, 0x04,
		0x0a, 0x0c, 0x09, 0x05, 0x03, 0x06, 0x0e, 0x0b,
		0x05, 0x00, 0x00, 0x0e, 0x0c, 0x09, 0x07, 0x02,
		0x07, 0x02, 0x0b, 0x01, 0x04, 0x0e, 0x01, 0x07,
		0x09, 0x04, 0x0c, 0x0a, 0x0e, 0x08, 0x02, 0x0d,
		0x00, 0x0f, 0x06, 0x0c, 0x0a, 0x09, 0x0d, 0x00,
		0x0f, 0x03, 0x03, 0x05, 0x05, 0x06, 0x08, 0x0b,
		0x9d, 0x91, 0x92, 0x9f, 0x98, 0x9d, 0x94, 0x98,
		0x96, 0x9a, 0x9f, 0x93, 0x9b, 0x97, 0x91, 0x94,
		0x9a, 0x9c, 0x99, 0x95, 0x93, 0x96, 0x9e, 0x9b,
		0x95, 0x90, 0x90, 0x9e, 0x9c, 0x99, 0x97, 0x92,
		0x97, 0x92, 0x9b, 0x91, 0x94, 0x9e, 0x91, 0x97,
		0x99, 0x94, 0x9c, 0x9a, 0x9e, 0x98, 0x92, 0x9d,
		0x90, 0x9f, 0x96, 0x9c, 0x9a, 0x99, 0x9d, 0x90,
		0x9f, 0x93, 0x93, 0x95, 0x95, 0x96, 0x98, 0x9b,
		0x8d, 0x81, 0x82, 0x8f, 0x88, 0x8d, 0x84, 0x88,
		0x86, 0x8a, 0x8f, 0x83, 0x8b, 0x87, 0x81, 0x84,
		0x8a, 0x8c, 0x89, 0x85, 0x83, 0x86, 0x8e, 0x8b,
		0x85, 0x80, 0x80, 0x8e, 0x8c, 0x89, 0x87, 0x82,
		0x87, 0x82, 0x8b, 0x81, 0x84, 0x8e, 0x81, 0x87,
		0x89, 0x84, 0x8c, 0x8a, 0x8e, 0x88, 0x82, 0x8d,
		0x80, 0x8f, 0x86, 0x8c, 0x8a, 0x89, 0x8d, 0x80,
		0x8f, 0x83, 0x83, 0x85, 0x85, 0x86, 0x88, 0x8b,
		0x1d, 0x11, 0x12, 0x1f, 0x18, 0x1d, 0x14, 0x18,
		0x16, 0x1a, 0x1f, 0x13, 0x1b, 0x17, 0x11, 0x14,
		0x1a, 0x1c, 0x19, 0x15, 0x13, 0x16, 0x1e, 0x1b,
		0x15, 0x10, 0x10, 0x1e, 0x1c, 0x19, 0x17, 0x12,
		0x17, 0x12, 0x1b, 0x11, 0x14, 0x1e, 0x11, 0x17,
		0x19, 0x14, 0x1c, 0x1a, 0x1e, 0x18, 0x12, 0x1d,
		0x10, 0x1f, 0x16, 0x1c, 0x1a, 0x19, 0x1d, 0x10,
		0x1f, 0x13, 0x13, 0x15, 0x15, 0x16, 0x18, 0x1b,
		0xdd, 0xd1, 0xd2, 0xdf, 0xd8, 0xdd, 0xd4, 0xd8,
		0xd6, 0xda, 0xdf, 0xd3, 0xdb, 0xd7, 0xd1, 0xd4,
		0xda, 0xdc, 0xd9, 0xd5, 0xd3, 0xd6, 0xde, 0xdb,
		0xd5, 0xd0, 0xd0, 0xde, 0xdc, 0xd9, 0xd7, 0xd2,
		0xd7, 0xd2, 0xdb, 0xd1, 0xd4, 0xde, 0xd1, 0xd7,
		0xd9, 0xd4, 0xdc, 0xda, 0xde, 0xd8, 0xd2, 0xdd,
		0xd0, 0xdf, 0xd6, 0xdc, 0xda, 0xd9, 0xdd, 0xd0,
		0xdf, 0xd3, 0xd3, 0xd5, 0xd5, 0xd6, 0xd8, 0xdb,
		0xad, 0xa1, 0xa2, 0xaf, 0xa8, 0xad, 0xa4, 0xa8,
		0xa6, 0xaa, 0xaf, 0xa3, 0xab, 0xa7, 0xa1, 0xa4,
		0xaa, 0xac, 0xa9, 0xa5, 0xa3, 0xa6, 0xae, 0xab,
		0xa5, 0xa0, 0xa0, 0xae, 0xac, 0xa9, 0xa7, 0xa2,
		0xa7, 0xa2, 0xab, 0xa1, 0xa4, 0xae, 0xa1, 0xa7,
		0xa9, 0xa4, 0xac, 0xaa, 0xae, 0xa8, 0xa2, 0xad,
		0xa0, 0xaf, 0xa6, 0xac, 0xaa, 0xa9, 0xad, 0xa0,
		0xaf, 0xa3, 0xa3, 0xa5, 0xa5, 0xa6, 0xa8, 0xab,
		0x3d, 0x31, 0x32, 0x3f, 0x38, 0x3d, 0x34, 0x38,
		0x36, 0x3a, 0x3f, 0x33, 0x3b, 0x37, 0x31, 0x34,
		0x3a, 0x3c, 0x39, 0x35, 0x33, 0x36, 0x3e, 0x3b,
		0x35, 0x30, 0x30, 0x3e, 0x3c, 0x39, 0x37, 0x32,
		0x37, 0x32, 0x3b, 0x31, 0x34, 0x3e, 0x31, 0x37,
		0x39, 0x34, 0x3c, 0x3a, 0x3e, 0x38, 0x32, 0x3d,
		0x30, 0x3f, 0x36, 0x3c, 0x3a, 0x39, 0x3d, 0x30,
		0x3f, 0x33, 0x33, 0x35, 0x35, 0x36, 0x38, 0x3b,
		0xed, 0xe1, 0xe2, 0xef, 0xe8, 0xed, 0xe4, 0xe8,
		0xe6, 0xea, 0xef, 0xe3, 0xeb, 0xe7, 0xe1, 0xe4,
		0xea, 0xec, 0xe9, 0xe5, 0xe3, 0xe6, 0xee, 0xeb,
		0xe5, 0xe0, 0xe0, 0xee, 0xec, 0xe9, 0xe7, 0xe2,
		0xe7, 0xe2, 0xeb, 0xe1, 0xe4, 0xee, 0xe1, 0xe7,
		0xe9, 0xe4, 0xec, 0xea, 0xee, 0xe8, 0xe2, 0xed,
		0xe0, 0xef, 0xe6, 0xec, 0xea, 0xe9, 0xed, 0xe0,
		0xef, 0xe3, 0xe3, 0xe5, 0xe5, 0xe6, 0xe8, 0xeb,
		0xcd, 0xc1, 0xc2, 0xcf, 0xc8, 0xcd, 0xc4, 0xc8,
		0xc6, 0xca, 0xcf, 0xc3, 0xcb, 0xc7, 0xc1, 0xc4,
		0xca, 0xcc, 0xc9, 0xc5, 0xc3, 0xc6, 0xce, 0xcb,
		0xc5, 0xc0, 0xc0, 0xce, 0xcc, 0xc9, 0xc7, 0xc2,
		0xc7, 0xc2, 0xcb, 0xc1, 0xc4, 0xce, 0xc1, 0xc7,
		0xc9, 0xc4, 0xcc, 0xca, 0xce, 0xc8, 0xc2, 0xcd,
		0xc0, 0xcf, 0xc6, 0xcc, 0xca, 0xc9, 0xcd, 0xc0,
		0xcf, 0xc3, 0xc3, 0xc5, 0xc5, 0xc6, 0xc8, 0xcb,
		0x3d, 0x31, 0x32, 0x3f, 0x38, 0x3d, 0x34, 0x38,
		0x36, 0x3a, 0x3f, 0x33, 0x3b, 0x37, 0x31, 0x34,
		0x3a, 0x3c, 0x39, 0x35, 0x33, 0x36, 0x3e, 0x3b,
		0x35, 0x30, 0x30, 0x3e, 0x3c, 0x39, 0x37, 0x32,
		0x37, 0x32, 0x3b, 0x31, 0x34, 0x3e, 0x31, 0x37,
		0x39, 0x34, 0x3c, 0x3a, 0x3e, 0x38, 0x32, 0x3d,
		0x30, 0x3f, 0x36, 0x3c, 0x3a, 0x39, 0x3d, 0x30,
		0x3f, 0x33, 0x33, 0x35, 0x35, 0x36, 0x38, 0x3b,
		0x9d, 0x91, 0x92, 0x9f, 0x98, 0x9d, 0x94, 0x98,
		0x96, 0x9a, 0x9f, 0x93, 0x9b, 0x97, 0x91, 0x94,
		0x9a, 0x9c, 0x99, 0x95, 0x93, 0x96, 0x9e, 0x9b,
		0x95, 0x90, 0x90, 0x9e, 0x9c, 0x99, 0x97, 0x92,
		0x97, 0x92, 0x9b, 0x91, 0x94, 0x9e, 0x91, 0x97,
		0x99, 0x94, 0x9c, 0x9a, 0x9e, 0x98, 0x92, 0x9d,
		0x90, 0x9f, 0x96, 0x9c, 0x9a, 0x99, 0x9d, 0x90,
		0x9f, 0x93, 0x93, 0x95, 0x95, 0x96, 0x98, 0x9b,
		0x5d, 0x51, 0x52, 0x5f, 0x58, 0x5d, 0x54, 0x58,
		0x56, 0x5a, 0x5f, 0x53, 0x5b, 0x57, 0x51, 0x54,
		0x5a, 0x5c, 0x59, 0x55, 0x53, 0x56, 0x5e, 0x5b,
		0x55, 0x50, 0x50, 0x5e, 0x5c, 0x59, 0x57, 0x52,
		0x57, 0x52, 0x5b, 0x51, 0x54, 0x5e, 0x51, 0x57,
		0x59, 0x54, 0x5c, 0x5a, 0x5e, 0x58, 0x52, 0x5d,
		0x50, 0x5f, 0x56, 0x5c, 0x5a, 0x59, 0x5d, 0x50,
		0x5f, 0x53, 0x53, 0x55, 0x55, 0x56, 0x58, 0x5b,
		0x7d, 0x71, 0x72, 0x7f, 0x78, 0x7d, 0x74, 0x78,
		0x76, 0x7a, 0x7f, 0x73, 0x7b, 0x77, 0x71, 0x74,
		0x7a, 0x7c, 0x79, 0x75, 0x73, 0x76, 0x7e, 0x7b,
		0x75, 0x70, 0x70, 0x7e, 0x7c, 0x79, 0x77, 0x72,
		0x77, 0x72, 0x7b, 0x71, 0x74, 0x7e, 0x71, 0x77,
		0x79, 0x74, 0x7c, 0x7a, 0x7e, 0x78, 0x72, 0x7d,
		0x70, 0x7f, 0x76, 0x7c, 0x7a, 0x79, 0x7d, 0x70,
		0x7f, 0x73, 0x73, 0x75, 0x75, 0x76, 0x78, 0x7b,
		0xcd, 0xc1, 0xc2, 0xcf, 0xc8, 0xcd, 0xc4, 0xc8,
		0xc6, 0xca, 0xcf, 0xc3, 0xcb, 0xc7, 0xc1, 0xc4,
		0xca, 0xcc, 0xc9, 0xc5, 0xc3, 0xc6, 0xce, 0xcb,
		0xc5, 0xc0, 0xc0, 0xce, 0xcc, 0xc9, 0xc7, 0xc2,
		0xc7, 0xc2, 0xcb, 0xc1, 0xc4, 0xce, 0xc1, 0xc7,
		0xc9, 0xc4, 0xcc, 0xca, 0xce, 0xc8, 0xc2, 0xcd,
		0xc0, 0xcf, 0xc6, 0xcc, 0xca, 0xc9, 0xcd, 0xc0,
		0xcf, 0xc3, 0xc3, 0xc5, 0xc5, 0xc6, 0xc8, 0xcb,
		0x5d, 0x51, 0x52, 0x5f, 0x58, 0x5d, 0x54, 0x58,
		0x56, 0x5a, 0x5f, 0x53, 0x5b, 0x57, 0x51, 0x54,
		0x5a, 0x5c, 0x59, 0x55, 0x53, 0x56, 0x5e, 0x5b,
		0x55, 0x50, 0x50, 0x5e, 0x5c, 0x59, 0x57, 0x52,
		0x57, 0x52, 0x5b, 0x51, 0x54, 0x5e, 0x51, 0x57,
		0x59, 0x54, 0x5c, 0x5a, 0x5e, 0x58, 0x52, 0x5d,
		0x50, 0x5f, 0x56, 0x5c, 0x5a, 0x59, 0x5d, 0x50,
		0x5f, 0x53, 0x53, 0x55, 0x55, 0x56, 0x58, 0x5b,
		0x2d, 0x21, 0x22, 0x2f, 0x28, 0x2d, 0x24, 0x28,
		0x26, 0x2a, 0x2f, 0x23, 0x2b, 0x27, 0x21, 0x24,
		0x2a, 0x2c, 0x29, 0x25, 0x23, 0x26, 0x2e, 0x2b,
		0x25, 0x20, 0x20, 0x2e, 0x2c, 0x29, 0x27, 0x22,
		0x27, 0x22, 0x2b, 0x21, 0x24, 0x2e, 0x21, 0x27,
		0x29, 0x24, 0x2c, 0x2a, 0x2e, 0x28, 0x22, 0x2d,
		0x20, 0x2f, 0x26, 0x2c, 0x2a, 0x29, 0x2d, 0x20,
		0x2f, 0x23, 0x23, 0x25, 0x25, 0x26, 0x28, 0x2b,
		0xad, 0xa1, 0xa2, 0xaf, 0xa8, 0xad, 0xa4, 0xa8,
		0xa6, 0xaa, 0xaf, 0xa3, 0xab, 0xa7, 0xa1, 0xa4,
		0xaa, 0xac, 0xa9, 0xa5, 0xa3, 0xa6, 0xae, 0xab,
		0xa5, 0xa0, 0xa0, 0xae, 0xac, 0xa9, 0xa7, 0xa2,
		0xa7, 0xa2, 0xab, 0xa1, 0xa4, 0xae, 0xa1, 0xa7,
		0xa9, 0xa4, 0xac, 0xaa, 0xae, 0xa8, 0xa2, 0xad,
		0xa0, 0xaf, 0xa6, 0xac, 0xaa, 0xa9, 0xad, 0xa0,
		0xaf, 0xa3, 0xa3, 0xa5, 0xa5, 0xa6, 0xa8, 0xab,
		0xfd, 0xf1, 0xf2, 0xff, 0xf8, 0xfd, 0xf4, 0xf8,
		0xf6, 0xfa, 0xff, 0xf3, 0xfb, 0xf7, 0xf1, 0xf4,
		0xfa, 0xfc, 0xf9, 0xf5, 0xf3, 0xf6, 0xfe, 0xfb,
		0xf5, 0xf0, 0xf0, 0xfe, 0xfc, 0xf9, 0xf7, 0xf2,
		0xf7, 0xf2, 0xfb, 0xf1, 0xf4, 0xfe, 0xf1, 0xf7,
		0xf9, 0xf4, 0xfc, 0xfa, 0xfe, 0xf8, 0xf2, 0xfd,
		0xf0, 0xff, 0xf6, 0xfc, 0xfa, 0xf9, 0xfd, 0xf0,
		0xff, 0xf3, 0xf3, 0xf5, 0xf5, 0xf6, 0xf8, 0xfb,
		0x6d, 0x61, 0x62, 0x6f, 0x68, 0x6d, 0x64, 0x68,
		0x66, 0x6a, 0x6f, 0x63, 0x6b, 0x67, 0x61, 0x64,
		0x6a, 0x6c, 0x69, 0x65, 0x63, 0x66, 0x6e, 0x6b,
		0x65, 0x60, 0x60, 0x6e, 0x6c, 0x69, 0x67, 0x62,
		0x67, 0x62, 0x6b, 0x61, 0x64, 0x6e, 0x61, 0x67,
		0x69, 0x64, 0x6c, 0x6a, 0x6e, 0x68, 0x62, 0x6d,
		0x60, 0x6f, 0x66, 0x6c, 0x6a, 0x69, 0x6d, 0x60,
		0x6f, 0x63, 0x63, 0x65, 0x65, 0x66, 0x68, 0x6b,
		0x8d, 0x81, 0x82, 0x8f, 0x88, 0x8d, 0x84, 0x88,
		0x86, 0x8a, 0x8f, 0x83, 0x8b, 0x87, 0x81, 0x84,
		0x8a, 0x8c, 0x89, 0x85, 0x83, 0x86, 0x8e, 0x8b,
		0x85, 0x80, 0x80, 0x8e, 0x8c, 0x89, 0x87, 0x82,
		0x87, 0x82, 0x8b, 0x81, 0x84, 0x8e, 0x81, 0x87,
		0x89, 0x84, 0x8c, 0x8a, 0x8e, 0x88, 0x82, 0x8d,
		0x80, 0x8f, 0x86, 0x8c, 0x8a, 0x89, 0x8d, 0x80,
		0x8f, 0x83, 0x83, 0x85, 0x85, 0x86, 0x88, 0x8b,
		0x1d, 0x11, 0x12, 0x1f, 0x18, 0x1d, 0x14, 0x18,
		0x16, 0x1a, 0x1f, 0x13, 0x1b, 0x17, 0x11, 0x14,
		0x1a, 0x1c, 0x19, 0x15, 0x13, 0x16, 0x1e, 0x1b,
		0x15, 0x10, 0x10, 0x1e, 0x1c, 0x19, 0x17, 0x12,
		0x17, 0x12, 0x1b, 0x11, 0x14, 0x1e, 0x11, 0x17,
		0x19, 0x14, 0x1c, 0x1a, 0x1e, 0x18, 0x12, 0x1d,
		0x10, 0x1f, 0x16, 0x1c, 0x1a, 0x19, 0x1d, 0x10,
		0x1f, 0x13, 0x13, 0x15, 0x15, 0x16, 0x18, 0x1b,
		0x6d, 0x61, 0x62, 0x6f, 0x68, 0x6d, 0x64, 0x68,
		0x66, 0x6a, 0x6f, 0x63, 0x6b, 0x67, 0x61, 0x64,
		0x6a, 0x6c, 0x69, 0x65, 0x63, 0x66, 0x6e, 0x6b,
		0x65, 0x60, 0x60, 0x6e, 0x6c, 0x69, 0x67, 0x62,
		0x67, 0x62, 0x6b, 0x61, 0x64, 0x6e, 0x61, 0x67,
		0x69, 0x64, 0x6c, 0x6a, 0x6e, 0x68, 0x62, 0x6d,
		0x60, 0x6f, 0x66, 0x6c, 0x6a, 0x69, 0x6d, 0x60,
		0x6f, 0x63, 0x63, 0x65, 0x65, 0x66, 0x68, 0x6b,
		0x1d, 0x11, 0x12, 0x1f, 0x18, 0x1d, 0x14, 0x18,
		0x16, 0x1a, 0x1f, 0x13, 0x1b, 0x17, 0x11, 0x14,
		0x1a, 0x1c, 0x19, 0x15, 0x13, 0x16, 0x1e, 0x1b,
		0x15, 0x10, 0x10, 0x1e, 0x1c, 0x19, 0x17, 0x12,
		0x17, 0x12, 0x1b, 0x11, 0x14, 0x1e, 0x11, 0x17,
		0x19, 0x14, 0x1c, 0x1a, 0x1e, 0x18, 0x12, 0x1d,
		0x10, 0x1f, 0x16, 0x1c, 0x1a, 0x19, 0x1d, 0x10,
		0x1f, 0x13, 0x13, 0x15, 0x15, 0x16, 0x18, 0x1b,
		0x6d, 0x61, 0x62, 0x6f, 0x68, 0x6d, 0x64, 0x68,
		0x66, 0x6a, 0x6f, 0x63, 0x6b, 0x67, 0x61, 0x64,
		0x6a, 0x6c, 0x69, 0x65, 0x63, 0x66, 0x6e, 0x6b,
		0x65, 0x60, 0x60, 0x6e, 0x6c, 0x69, 0x67, 0x62,
		0x67, 0x62, 0x6b, 0x61, 0x64, 0x6e, 0x61, 0x67,
		0x69, 0x64, 0x6c, 0x6a, 0x6e, 0x68, 0x62, 0x6d,
		0x60, 0x6f, 0x66, 0x6c, 0x6a, 0x69, 0x6d, 0x60,
		0x6f, 0x63, 0x63, 0x65, 0x65, 0x66, 0x68, 0x6b,
		0x4d, 0x41, 0x42, 0x4f, 0x48, 0x4d, 0x44, 0x48,
		0x46, 0x4a, 0x4f, 0x43, 0x4b, 0x47, 0x41, 0x44,
		0x4a, 0x4c, 0x49, 0x45, 0x43, 0x46, 0x4e, 0x4b,
		0x45, 0x40, 0x40, 0x4e, 0x4c, 0x49, 0x47, 0x42,
		0x47, 0x42, 0x4b, 0x41, 0x44, 0x4e, 0x41, 0x47,
		0x49, 0x44, 0x4c, 0x4a, 0x4e, 0x48, 0x42, 0x4d,
		0x40, 0x4f, 0x46, 0x4c, 0x4a, 0x49, 0x4d, 0x40,
		0x4f, 0x43, 0x43, 0x45, 0x45, 0x46, 0x48, 0x4b,
		0xbd, 0xb1, 0xb2, 0xbf, 0xb8, 0xbd, 0xb4, 0xb8,
		0xb6, 0xba, 0xbf, 0xb3, 0xbb, 0xb7, 0xb1, 0xb4,
		0xba, 0xbc, 0xb9, 0xb5, 0xb3, 0xb6, 0xbe, 0xbb,
		0xb5, 0xb0, 0xb0, 0xbe, 0xbc, 0xb9, 0xb7, 0xb2,
		0xb7, 0xb2, 0xbb, 0xb1, 0xb4, 0xbe, 0xb1, 0xb7,
		0xb9, 0xb4, 0xbc, 0xba, 0xbe, 0xb8, 0xb2, 0xbd,
		0xb0, 0xbf, 0xb6, 0xbc, 0xba, 0xb9, 0xbd, 0xb0,
		0xbf, 0xb3, 0xb3, 0xb5, 0xb5, 0xb6, 0xb8, 0xbb,
		0xbd, 0xb1, 0xb2, 0xbf, 0xb8, 0xbd, 0xb4, 0xb8,
		0xb6, 0xba, 0xbf, 0xb3, 0xbb, 0xb7, 0xb1, 0xb4,
		0xba, 0xbc, 0xb9, 0xb5, 0xb3, 0xb6, 0xbe, 0xbb,
		0xb5, 0xb0, 0xb0, 0xbe, 0xbc, 0xb9, 0xb7, 0xb2,
		0xb7, 0xb2, 0xbb, 0xb1, 0xb4, 0xbe, 0xb1, 0xb7,
		0xb9, 0xb4, 0xbc, 0xba, 0xbe, 0xb8, 0xb2, 0xbd,
		0xb0, 0xbf, 0xb6, 0xbc, 0xba, 0xb9, 0xbd, 0xb0,
		0xbf, 0xb3, 0xb3, 0xb5, 0xb5, 0xb6, 0xb8, 0xbb,
		0xdd, 0xd1, 0xd2, 0xdf, 0xd8, 0xdd, 0xd4, 0xd8,
		0xd6, 0xda, 0xdf, 0xd3, 0xdb, 0xd7, 0xd1, 0xd4,
		0xda, 0xdc, 0xd9, 0xd5, 0xd3, 0xd6, 0xde, 0xdb,
		0xd5, 0xd0, 0xd0, 0xde, 0xdc, 0xd9, 0xd7, 0xd2,
		0xd7, 0xd2, 0xdb, 0xd1, 0xd4, 0xde, 0xd1, 0xd7,
		0xd9, 0xd4, 0xdc, 0xda, 0xde, 0xd8, 0xd2, 0xdd,
		0xd0, 0xdf, 0xd6, 0xdc, 0xda, 0xd9, 0xdd, 0xd0,
		0xdf, 0xd3, 0xd3, 0xd5, 0xd5, 0xd6, 0xd8, 0xdb,
		0xdd, 0xd1, 0xd2, 0xdf, 0xd8, 0xdd, 0xd4, 0xd8,
		0xd6, 0xda, 0xdf, 0xd3, 0xdb, 0xd7, 0xd1, 0xd4,
		0xda, 0xdc, 0xd9, 0xd5, 0xd3, 0xd6, 0xde, 0xdb,
		0xd5, 0xd0, 0xd0, 0xde, 0xdc, 0xd9, 0xd7, 0xd2,
		0xd7, 0xd2, 0xdb, 0xd1, 0xd4, 0xde, 0xd1, 0xd7,
		0xd9, 0xd4, 0xdc, 0xda, 0xde, 0xd8, 0xd2, 0xdd,
		0xd0, 0xdf, 0xd6, 0xdc, 0xda, 0xd9, 0xdd, 0xd0,
		0xdf, 0xd3, 0xd3, 0xd5, 0xd5, 0xd6, 0xd8, 0xdb,
		0x8d, 0x81, 0x82, 0x8f, 0x88, 0x8d, 0x84, 0x88,
		0x86, 0x8a, 0x8f, 0x83, 0x8b, 0x87, 0x81, 0x84,
		0x8a, 0x8c, 0x89, 0x85, 0x83, 0x86, 0x8e, 0x8b,
		0x85, 0x80, 0x80, 0x8e, 0x8c, 0x89, 0x87, 0x82,
		0x87, 0x82, 0x8b, 0x81, 0x84, 0x8e, 0x81, 0x87,
		0x89, 0x84, 0x8c, 0x8a, 0x8e, 0x88, 0x82, 0x8d,
		0x80, 0x8f, 0x86, 0x8c, 0x8a, 0x89, 0x8d, 0x80,
		0x8f, 0x83, 0x83, 0x85, 0x85, 0x86, 0x88, 0x8b,
		0xcd, 0xc1, 0xc2, 0xcf, 0xc8, 0xcd, 0xc4, 0xc8,
		0xc6, 0xca, 0xcf, 0xc3, 0xcb, 0xc7, 0xc1, 0xc4,
		0xca, 0xcc, 0xc9, 0xc5, 0xc3, 0xc6, 0xce, 0xcb,
		0xc5, 0xc0, 0xc0, 0xce, 0xcc, 0xc9, 0xc7, 0xc2,
		0xc7, 0xc2, 0xcb, 0xc1, 0xc4, 0xce, 0xc1, 0xc7,
		0xc9, 0xc4, 0xcc, 0xca, 0xce, 0xc8, 0xc2, 0xcd,
		0xc0, 0xcf, 0xc6, 0xcc, 0xca, 0xc9, 0xcd, 0xc0,
		0xcf, 0xc3, 0xc3, 0xc5, 0xc5, 0xc6, 0xc8, 0xcb,
		0x1d, 0x11, 0x12, 0x1f, 0x18, 0x1d, 0x14, 0x18,
		0x16, 0x1a, 0x1f, 0x13, 0x1b, 0x17, 0x11, 0x14,
		0x1a, 0x1c, 0x19, 0x15, 0x13, 0x16, 0x1e, 0x1b,
		0x15, 0x10, 0x10, 0x1e, 0x1c, 0x19, 0x17, 0x12,
		0x17, 0x12, 0x1b, 0x11, 0x14, 0x1e, 0x11, 0x17,
		0x19, 0x14, 0x1c, 0x1a, 0x1e, 0x18, 0x12, 0x1d,
		0x10, 0x1f, 0x16, 0x1c, 0x1a, 0x19, 0x1d, 0x10,
		0x1f, 0x13, 0x13, 0x15, 0x15, 0x16, 0x18, 0x1b,
		0x3d, 0x31, 0x32, 0x3f, 0x38, 0x3d, 0x34, 0x38,
		0x36, 0x3a, 0x3f, 0x33, 0x3b, 0x37, 0x31, 0x34,
		0x3a, 0x3c, 0x39, 0x35, 0x33, 0x36, 0x3e, 0x3b,
		0x35, 0x30, 0x30, 0x3e, 0x3c, 0x39, 0x37, 0x32,
		0x37, 0x32, 0x3b, 0x31, 0x34, 0x3e, 0x31, 0x37,
		0x39, 0x34, 0x3c, 0x3a, 0x3e, 0x38, 0x32, 0x3d,
		0x30, 0x3f, 0x36, 0x3c, 0x3a, 0x39, 0x3d, 0x30,
		0x3f, 0x33, 0x33, 0x35, 0x35, 0x36, 0x38, 0x3b,
		0x4d, 0x41, 0x42, 0x4f, 0x48, 0x4d, 0x44, 0x48,
		0x46, 0x4a, 0x4f, 0x43, 0x4b, 0x47, 0x41, 0x44,
		0x4a, 0x4c, 0x49, 0x45, 0x43, 0x46, 0x4e, 0x4b,
		0x45, 0x40, 0x40, 0x4e, 0x4c, 0x49, 0x47, 0x42,
		0x47, 0x42, 0x4b, 0x41, 0x44, 0x4e, 0x41, 0x47,
		0x49, 0x44, 0x4c, 0x4a, 0x4e, 0x48, 0x42, 0x4d,
		0x40, 0x4f, 0x46, 0x4c, 0x4a, 0x49, 0x4d, 0x40,
		0x4f, 0x43, 0x43, 0x45, 0x45, 0x46, 0x48, 0x4b,
		0x7d, 0x71, 0x72, 0x7f, 0x78, 0x7d, 0x74, 0x78,
		0x76, 0x7a, 0x7f, 0x73, 0x7b, 0x77, 0x71, 0x74,
		0x7a, 0x7c, 0x79, 0x75, 0x73, 0x76, 0x7e, 0x7b,
		0x75, 0x70, 0x70, 0x7e, 0x7c, 0x79, 0x77, 0x72,
		0x77, 0x72, 0x7b, 0x71, 0x74, 0x7e, 0x71, 0x77,
		0x79, 0x74, 0x7c, 0x7a, 0x7e, 0x78, 0x72, 0x7d,
		0x70, 0x7f, 0x76, 0x7c, 0x7a, 0x79, 0x7d, 0x70,
		0x7f, 0x73, 0x73, 0x75, 0x75, 0x76, 0x78, 0x7b,
		0xad, 0xa1, 0xa2, 0xaf, 0xa8, 0xad, 0xa4, 0xa8,
		0xa6, 0xaa, 0xaf, 0xa3, 0xab, 0xa7, 0xa1, 0xa4,
		0xaa, 0xac, 0xa9, 0xa5, 0xa3, 0xa6, 0xae, 0xab,
		0xa5, 0xa0, 0xa0, 0xae, 0xac, 0xa9, 0xa7, 0xa2,
		0xa7, 0xa2, 0xab, 0xa1, 0xa4, 0xae, 0xa1, 0xa7,
		0xa9, 0xa4, 0xac, 0xaa, 0xae, 0xa8, 0xa2, 0xad,
		0xa0, 0xaf, 0xa6, 0xac, 0xaa, 0xa9, 0xad, 0xa0,
		0xaf, 0xa3, 0xa3, 0xa5, 0xa5, 0xa6, 0xa8, 0xab,
		0xed, 0xe1, 0xe2, 0xef, 0xe8, 0xed, 0xe4, 0xe8,
		0xe6, 0xea, 0xef, 0xe3, 0xeb, 0xe7, 0xe1, 0xe4,
		0xea, 0xec, 0xe9, 0xe5, 0xe3, 0xe6, 0xee, 0xeb,
		0xe5, 0xe0, 0xe0, 0xee, 0xec, 0xe9, 0xe7, 0xe2,
		0xe7, 0xe2, 0xeb, 0xe1, 0xe4, 0xee, 0xe1, 0xe7,
		0xe9, 0xe4, 0xec, 0xea, 0xee, 0xe8, 0xe2, 0xed,
		0xe0, 0xef, 0xe6, 0xec, 0xea, 0xe9, 0xed, 0xe0,
		0xef, 0xe3, 0xe3, 0xe5, 0xe5, 0xe6, 0xe8, 0xeb,
		0x7d, 0x71, 0x72, 0x7f, 0x78, 0x7d, 0x74, 0x78,
		0x76, 0x7a, 0x7f, 0x73, 0x7b, 0x77, 0x71, 0x74,
		0x7a, 0x7c, 0x79, 0x75, 0x73, 0x76, 0x7e, 0x7b,
		0x75, 0x70, 0x70, 0x7e, 0x7c, 0x79, 0x77, 0x72,
		0x77, 0x72, 0x7b, 0x71, 0x74, 0x7e, 0x71, 0x77,
		0x79, 0x74, 0x7c, 0x7a, 0x7e, 0x78, 0x72, 0x7d,
		0x70, 0x7f, 0x76, 0x7c, 0x7a, 0x79, 0x7d, 0x70,
		0x7f, 0x73, 0x73, 0x75, 0x75, 0x76, 0x78, 0x7b,
		0xad, 0xa1, 0xa2, 0xaf, 0xa8, 0xad, 0xa4, 0xa8,
		0xa6, 0xaa, 0xaf, 0xa3, 0xab, 0xa7, 0xa1, 0xa4,
		0xaa, 0xac, 0xa9, 0xa5, 0xa3, 0xa6, 0xae, 0xab,
		0xa5, 0xa0, 0xa0, 0xae, 0xac, 0xa9, 0xa7, 0xa2,
		0xa7, 0xa2, 0xab, 0xa1, 0xa4, 0xae, 0xa1, 0xa7,
		0xa9, 0xa4, 0xac, 0xaa, 0xae, 0xa8, 0xa2, 0xad,
		0xa0, 0xaf, 0xa6, 0xac, 0xaa, 0xa9, 0xad, 0xa0,
		0xaf, 0xa3, 0xa3, 0xa5, 0xa5, 0xa6, 0xa8, 0xab,
		0x9d, 0x91, 0x92, 0x9f, 0x98, 0x9d, 0x94, 0x98,
		0x96, 0x9a, 0x9f, 0x93, 0x9b, 0x97, 0x91, 0x94,
		0x9a, 0x9c, 0x99, 0x95, 0x93, 0x96, 0x9e, 0x9b,
		0x95, 0x90, 0x90, 0x9e, 0x9c, 0x99, 0x97, 0x92,
		0x97, 0x92, 0x9b, 0x91, 0x94, 0x9e, 0x91, 0x97,
		0x99, 0x94, 0x9c, 0x9a, 0x9e, 0x98, 0x92, 0x9d,
		0x90, 0x9f, 0x96, 0x9c, 0x9a, 0x99, 0x9d, 0x90,
		0x9f, 0x93, 0x93, 0x95, 0x95, 0x96, 0x98, 0x9b,
		0xfd, 0xf1, 0xf2, 0xff, 0xf8, 0xfd, 0xf4, 0xf8,
		0xf6, 0xfa, 0xff, 0xf3, 0xfb, 0xf7, 0xf1, 0xf4,
		0xfa, 0xfc, 0xf9, 0xf5, 0xf3, 0xf6, 0xfe, 0xfb,
		0xf5, 0xf0, 0xf0, 0xfe, 0xfc, 0xf9, 0xf7, 0xf2,
		0xf7, 0xf2, 0xfb, 0xf1, 0xf4, 0xfe, 0xf1, 0xf7,
		0xf9, 0xf4, 0xfc, 0xfa, 0xfe, 0xf8, 0xf2, 0xfd,
		0xf0, 0xff, 0xf6, 0xfc, 0xfa, 0xf9, 0xfd, 0xf0,
		0xff, 0xf3, 0xf3, 0xf5, 0xf5, 0xf6, 0xf8, 0xfb,
		0x5d, 0x51, 0x52, 0x5f, 0x58, 0x5d, 0x54, 0x58,
		0x56, 0x5a, 0x5f, 0x53, 0x5b, 0x57, 0x51, 0x54,
		0x5a, 0x5c, 0x59, 0x55, 0x53, 0x56, 0x5e, 0x5b,
		0x55, 0x50, 0x50, 0x5e, 0x5c, 0x59, 0x57, 0x52,
		0x57, 0x52, 0x5b, 0x51, 0x54, 0x5e, 0x51, 0x57,
		0x59, 0x54, 0x5c, 0x5a, 0x5e, 0x58, 0x52, 0x5d,
		0x50, 0x5f, 0x56, 0x5c, 0x5a, 0x59, 0x5d, 0x50,
		0x5f, 0x53, 0x53, 0x55, 0x55, 0x56, 0x58, 0x5b,
		0x6d, 0x61, 0x62, 0x6f, 0x68, 0x6d, 0x64, 0x68,
		0x66, 0x6a, 0x6f, 0x63, 0x6b, 0x67, 0x61, 0x64,
		0x6a, 0x6c, 0x69, 0x65, 0x63, 0x66, 0x6e, 0x6b,
		0x65, 0x60, 0x60, 0x6e, 0x6c, 0x69, 0x67, 0x62,
		0x67, 0x62, 0x6b, 0x61, 0x64, 0x6e, 0x61, 0x67,
		0x69, 0x64, 0x6c, 0x6a, 0x6e, 0x68, 0x62, 0x6d,
		0x60, 0x6f, 0x66, 0x6c, 0x6a, 0x69, 0x6d, 0x60,
		0x6f, 0x63, 0x63, 0x65, 0x65, 0x66, 0x68, 0x6b,
		0x0d, 0x01, 0x02, 0x0f, 0x08, 0x0d, 0x04, 0x08,
		0x06, 0x0a, 0x0f, 0x03, 0x0b, 0x07, 0x01, 0x04,
		0x0a, 0x0c, 0x09, 0x05, 0x03, 0x06, 0x0e, 0x0b,
		0x05, 0x00, 0x00, 0x0e, 0x0c, 0x09, 0x07, 0x02,
		0x07, 0x02, 0x0b, 0x01, 0x04, 0x0e, 0x01, 0x07,
		0x09, 0x04, 0x0c, 0x0a, 0x0e, 0x08, 0x02, 0x0d,
		0x00, 0x0f, 0x06, 0x0c, 0x0a, 0x09, 0x0d, 0x00,
		0x0f, 0x03, 0x03, 0x05, 0x05, 0x06, 0x08, 0x0b,
		0x8d, 0x81, 0x82, 0x8f, 0x88, 0x8d, 0x84, 0x88,
		0x86, 0x8a, 0x8f, 0x83, 0x8b, 0x87, 0x81, 0x84,
		0x8a, 0x8c, 0x89, 0x85, 0x83, 0x86, 0x8e, 0x8b,
		0x85, 0x80, 0x80, 0x8e, 0x8c, 0x89, 0x87, 0x82,
		0x87, 0x82, 0x8b, 0x81, 0x84, 0x8e, 0x81, 0x87,
		0x89, 0x84, 0x8c, 0x8a, 0x8e, 0x88, 0x82, 0x8d,
		0x80, 0x8f, 0x86, 0x8c, 0x8a, 0x89, 0x8d, 0x80,
		0x8f, 0x83, 0x83, 0x85, 0x85, 0x86, 0x88, 0x8b,
		0xfd, 0xf1, 0xf2, 0xff, 0xf8, 0xfd, 0xf4, 0xf8,
		0xf6, 0xfa, 0xff, 0xf3, 0xfb, 0xf7, 0xf1, 0xf4,
		0xfa, 0xfc, 0xf9, 0xf5, 0xf3, 0xf6, 0xfe, 0xfb,
		0xf5, 0xf0, 0xf0, 0xfe, 0xfc, 0xf9, 0xf7, 0xf2,
		0xf7, 0xf2, 0xfb, 0xf1, 0xf4, 0xfe, 0xf1, 0xf7,
		0xf9, 0xf4, 0xfc, 0xfa, 0xfe, 0xf8, 0xf2, 0xfd,
		0xf0, 0xff, 0xf6, 0xfc, 0xfa, 0xf9, 0xfd, 0xf0,
		0xff, 0xf3, 0xf3, 0xf5, 0xf5, 0xf6, 0xf8, 0xfb,
		0x0d, 0x01, 0x02, 0x0f, 0x08, 0x0d, 0x04, 0x08,
		0x06, 0x0a, 0x0f, 0x03, 0x0b, 0x07, 0x01, 0x04,
		0x0a, 0x0c, 0x09, 0x05, 0x03, 0x06, 0x0e, 0x0b,
		0x05, 0x00, 0x00, 0x0e, 0x0c, 0x09, 0x07, 0x02,
		0x07, 0x02, 0x0b, 0x01, 0x04, 0x0e, 0x01, 0x07,
		0x09, 0x04, 0x0c, 0x0a, 0x0e, 0x08, 0x02, 0x0d,
		0x00, 0x0f, 0x06, 0x0c, 0x0a, 0x09, 0x0d, 0x00,
		0x0f, 0x03, 0x03, 0x05, 0x05, 0x06, 0x08, 0x0b,
		0xed, 0xe1, 0xe2, 0xef, 0xe8, 0xed, 0xe4, 0xe8,
		0xe6, 0xea, 0xef, 0xe3, 0xeb, 0xe7, 0xe1, 0xe4,
		0xea, 0xec, 0xe9, 0xe5, 0xe3, 0xe6, 0xee, 0xeb,
		0xe5, 0xe0, 0xe0, 0xee, 0xec, 0xe9, 0xe7, 0xe2,
		0xe7, 0xe2, 0xeb, 0xe1, 0xe4, 0xee, 0xe1, 0xe7,
		0xe9, 0xe4, 0xec, 0xea, 0xee, 0xe8, 0xe2, 0xed,
		0xe0, 0xef, 0xe6, 0xec, 0xea, 0xe9, 0xed, 0xe0,
		0xef, 0xe3, 0xe3, 0xe5, 0xe5, 0xe6, 0xe8, 0xeb,
		0x5d, 0x51, 0x52, 0x5f, 0x58, 0x5d, 0x54, 0x58,
		0x56, 0x5a, 0x5f, 0x53, 0x5b, 0x57, 0x51, 0x54,
		0x5a, 0x5c, 0x59, 0x55, 0x53, 0x56, 0x5e, 0x5b,
		0x55, 0x50, 0x50, 0x5e, 0x5c, 0x59, 0x57, 0x52,
		0x57, 0x52, 0x5b, 0x51, 0x54, 0x5e, 0x51, 0x57,
		0x59, 0x54, 0x5c, 0x5a, 0x5e, 0x58, 0x52, 0x5d,
		0x50, 0x5f, 0x56, 0x5c, 0x5a, 0x59, 0x5d, 0x50,
		0x5f, 0x53, 0x53, 0x55, 0x55, 0x56, 0x58, 0x5b,
		0x2d, 0x21, 0x22, 0x2f, 0x28, 0x2d, 0x24, 0x28,
		0x26, 0x2a, 0x2f, 0x23, 0x2b, 0x27, 0x21, 0x24,
		0x2a, 0x2c, 0x29, 0x25, 0x23, 0x26, 0x2e, 0x2b,
		0x25, 0x20, 0x20, 0x2e, 0x2c, 0x29, 0x27, 0x22,
		0x27, 0x22, 0x2b, 0x21, 0x24, 0x2e, 0x21, 0x27,
		0x29, 0x24, 0x2c, 0x2a, 0x2e, 0x28, 0x22, 0x2d,
		0x20, 0x2f, 0x26, 0x2c, 0x2a, 0x29, 0x2d, 0x20,
		0x2f, 0x23, 0x23, 0x25, 0x25, 0x26, 0x28, 0x2b,
		0x9d, 0x91, 0x92, 0x9f, 0x98, 0x9d, 0x94, 0x98,
		0x96, 0x9a, 0x9f, 0x93, 0x9b, 0x97, 0x91, 0x94,
		0x9a, 0x9c, 0x99, 0x95, 0x93, 0x96, 0x9e, 0x9b,
		0x95, 0x90, 0x90, 0x9e, 0x9c, 0x99, 0x97, 0x92,
		0x97, 0x92, 0x9b, 0x91, 0x94, 0x9e, 0x91, 0x97,
		0x99, 0x94, 0x9c, 0x9a, 0x9e, 0x98, 0x92, 0x9d,
		0x90, 0x9f, 0x96, 0x9c, 0x9a, 0x99, 0x9d, 0x90,
		0x9f, 0x93, 0x93, 0x95, 0x95, 0x96, 0x98, 0x9b,
		0x3d, 0x31, 0x32, 0x3f, 0x38, 0x3d, 0x34, 0x38,
		0x36, 0x3a, 0x3f, 0x33, 0x3b, 0x37, 0x31, 0x34,
		0x3a, 0x3c, 0x39, 0x35, 0x33, 0x36, 0x3e, 0x3b,
		0x35, 0x30, 0x30, 0x3e, 0x3c, 0x39, 0x37, 0x32,
		0x37, 0x32, 0x3b, 0x31, 0x34, 0x3e, 0x31, 0x37,
		0x39, 0x34, 0x3c, 0x3a, 0x3e, 0x38, 0x32, 0x3d,
		0x30, 0x3f, 0x36, 0x3c, 0x3a, 0x39, 0x3d, 0x30,
		0x3f, 0x33, 0x33, 0x35, 0x35, 0x36, 0x38, 0x3b,
		0x2d, 0x21, 0x22, 0x2f, 0x28, 0x2d, 0x24, 0x28,
		0x26, 0x2a, 0x2f, 0x23, 0x2b, 0x27, 0x21, 0x24,
		0x2a, 0x2c, 0x29, 0x25, 0x23, 0x26, 0x2e, 0x2b,
		0x25, 0x20, 0x20, 0x2e, 0x2c, 0x29, 0x27, 0x22,
		0x27, 0x22, 0x2b, 0x21, 0x24, 0x2e, 0x21, 0x27,
		0x29, 0x24, 0x2c, 0x2a, 0x2e, 0x28, 0x22, 0x2d,
		0x20, 0x2f, 0x26, 0x2c, 0x2a, 0x29, 0x2d, 0x20,
		0x2f, 0x23, 0x23, 0x25, 0x25, 0x26, 0x28, 0x2b,
		0xcd, 0xc1, 0xc2, 0xcf, 0xc8, 0xcd, 0xc4, 0xc8,
		0xc6, 0xca, 0xcf, 0xc3, 0xcb, 0xc7, 0xc1, 0xc4,
		0xca, 0xcc, 0xc9, 0xc5, 0xc3, 0xc6, 0xce, 0xcb,
		0xc5, 0xc0, 0xc0, 0xce, 0xcc, 0xc9, 0xc7, 0xc2,
		0xc7, 0xc2, 0xcb, 0xc1, 0xc4, 0xce, 0xc1, 0xc7,
		0xc9, 0xc4, 0xcc, 0xca, 0xce, 0xc8, 0xc2, 0xcd,
		0xc0, 0xcf, 0xc6, 0xcc, 0xca, 0xc9, 0xcd, 0xc0,
		0xcf, 0xc3, 0xc3, 0xc5, 0xc5, 0xc6, 0xc8, 0xcb,
	},
};

static const uint8_t key_perm[56] = {
	57, 49, 41, 33, 25, 17,  9,  1, 58, 50, 42, 34, 26, 18,
	10,  2, 59, 51, 43, 35, 27, 19, 11,  3, 60, 52, 44, 36,
	63, 55, 47, 39, 31, 23, 15,  7, 62, 54, 46, 38, 30, 22,
	14,  6, 61, 53, 45, 37, 29, 21, 13,  5, 28, 20, 12,  4
};

static const uint8_t key_shifts[16] = {
	1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1
};

static const uint8_t comp_perm[48] = {
	14, 17, 11, 24,  1,  5,  3, 28, 15,  6, 21, 10,
	23, 19, 12,  4, 26,  8, 16,  7, 27, 20, 13,  2,
	41, 52, 31, 37, 47, 55, 30, 40, 51, 45, 33, 48,
	44, 49, 39, 56, 34, 53, 46, 42, 50, 36, 29, 32
};

static const uint8_t pbox[32] = {
	16,  7, 20, 21, 29, 12, 28, 17,  1, 15, 23, 26,  5, 18, 31, 10,
	 2,  8, 24, 14, 32, 27,  3,  9, 19, 13, 30,  6, 22, 11,  4, 25
};

static int
ascii_to_bin(char ch)
{
	if (ch > 'z')
		return 0;
	if (ch >= 'a')
		return (ch - 'a' + 38);
	if (ch > 'Z')
		return 0;
	if (ch >= 'A')
		return (ch - 'A' + 12);
	if (ch > '9')
		return 0;
	if (ch >= '.')
		return (ch - '.');
	return 0;
}



struct des_ctx {
	uint32_t saltbits; /* referenced 5 times */
	uint8_t	un_pbox[32]; /* 2 times */
	uint8_t	inv_comp_perm[56]; /* 3 times */
	uint8_t	inv_key_perm[64]; /* 3 times */
	uint32_t en_keysl[16], en_keysr[16]; /* 2 times each */
	uint32_t fp_maskl[8][256], fp_maskr[8][256]; /* 9 times each */
	uint32_t key_perm_maskl[8][128], key_perm_maskr[8][128]; /* 9 times */
	uint32_t comp_maskl[8][128], comp_maskr[8][128]; /* 9 times each */
	uint32_t psbox[4][256]; /* 5 times */
};
#define D (*ctx)
#define saltbits        (D.saltbits       )
#define old_salt        (D.old_salt       )
#define old_rawkey0     (D.old_rawkey0    )
#define old_rawkey1     (D.old_rawkey1    )
#define un_pbox         (D.un_pbox        )
#define inv_comp_perm   (D.inv_comp_perm  )
#define inv_key_perm    (D.inv_key_perm   )
#define en_keysl        (D.en_keysl       )
#define en_keysr        (D.en_keysr       )
#define de_keysl        (D.de_keysl       )
#define de_keysr        (D.de_keysr       )
#define ip_maskl        (D.ip_maskl       )
#define ip_maskr        (D.ip_maskr       )
#define fp_maskl        (D.fp_maskl       )
#define fp_maskr        (D.fp_maskr       )
#define key_perm_maskl  (D.key_perm_maskl )
#define key_perm_maskr  (D.key_perm_maskr )
#define comp_maskl      (D.comp_maskl     )
#define comp_maskr      (D.comp_maskr     )
#define psbox           (D.psbox          )

static struct des_ctx *des_init(struct des_ctx *ctx)
{
	int i, j, b, k, inbit, obit;
	uint32_t p;

	saltbits = 0;

	/* Initialise the inverted key permutation. */
	for (i = 0; i < 64; i++) {
		inv_key_perm[i] = 255;
	}

	/*
	 * Invert the key permutation and initialise the inverted key
	 * compression permutation.
	 */
	for (i = 0; i < 56; i++) {
		inv_key_perm[key_perm[i] - 1] = (uint8_t)i;
		inv_comp_perm[i] = 255;
	}

	/* Invert the key compression permutation. */
	for (i = 0; i < 48; i++) {
		inv_comp_perm[comp_perm[i] - 1] = (uint8_t)i;
	}

	/*
	 * Set up the OR-mask arrays for the initial and final permutations,
	 * and for the key initial and compression permutations.
	 */
	for (k = 0; k < 8; k++) {
		uint32_t il, ir;
		uint32_t fl, fr;
		for (i = 0; i < 256; i++) {
			fl = 0;
			fr = 0;
			for (j = 0; j < 8; j++) {
				inbit = 8 * k + j;
				if (i & (0x80>>j)) {
					obit = final_perm[inbit];
					if (obit < 32)
						fl |= 0x80000000UL >> obit;
					else
						fr |= 0x80000000UL >> obit-32;
				}
			}
			fp_maskl[k][i] = fl;
			fp_maskr[k][i] = fr;
		}
		for (i = 0; i < 128; i++) {
			il = 0;
			ir = 0;
			for (j = 0; j < 7; j++) {
				inbit = 8 * k + j;
				if (i & (0x80 >> j + 1)) {
					obit = inv_key_perm[inbit];
					if (obit == 255)
						continue;
					if (obit < 28)
						il |= 0x8000000UL >> obit;
					else
						ir |= 0x8000000UL >> obit-28;
				}
			}
			key_perm_maskl[k][i] = il;
			key_perm_maskr[k][i] = ir;
			il = 0;
			ir = 0;
			for (j = 0; j < 7; j++) {
				inbit = 7 * k + j;
				if (i & (0x80 >> j + 1)) {
					obit = inv_comp_perm[inbit];
					if (obit == 255)
						continue;
					if (obit < 24)
						il |= 0x800000UL >> obit;
					else
						ir |= 0x800000UL >> obit-24;
				}
			}
			comp_maskl[k][i] = il;
			comp_maskr[k][i] = ir;
		}
	}

	/*
	 * Invert the P-box permutation, and convert into OR-masks for
	 * handling the output of the S-box arrays setup above.
	 */
	for (i = 0; i < 32; i++)
		un_pbox[pbox[i] - 1] = (uint8_t)i;

	for (b = 0; b < 4; b++) {
		for (i = 0; i < 256; i++) {
			p = 0;
			for (j = 0; j < 8; j++) {
				if (i & (0x80 >> j))
					p |= 0x80000000>>un_pbox[8 * b + j];
			}
			psbox[b][i] = p;
		}
	}

	return ctx;
}


static void setup_salt(struct des_ctx *ctx, uint32_t salt)
{
	uint32_t obit, saltbit;
	int i;

	saltbits = 0;
	saltbit = 1;
	obit = 0x800000;
	for (i = 0; i < 24; i++) {
		if (salt & saltbit)
			saltbits |= obit;
		saltbit <<= 1;
		obit >>= 1;
	}
}

static void des_setkey(struct des_ctx *ctx, const unsigned char *key)
{
	uint32_t k0, k1, rawkey0, rawkey1;
	int shifts, round;

	rawkey0 = key[0]<<24 | key[1]<<16 | key[2]<<8 | key[3];
	rawkey1 = key[4]<<24 | key[5]<<16 | key[6]<<8 | key[7];

	/*
	 * Do key permutation and split into two 28-bit subkeys.
	 */
	k0 = key_perm_maskl[0][rawkey0 >> 25]
	   | key_perm_maskl[1][(rawkey0 >> 17) & 0x7f]
	   | key_perm_maskl[2][(rawkey0 >> 9) & 0x7f]
	   | key_perm_maskl[3][(rawkey0 >> 1) & 0x7f]
	   | key_perm_maskl[4][rawkey1 >> 25]
	   | key_perm_maskl[5][(rawkey1 >> 17) & 0x7f]
	   | key_perm_maskl[6][(rawkey1 >> 9) & 0x7f]
	   | key_perm_maskl[7][(rawkey1 >> 1) & 0x7f];
	k1 = key_perm_maskr[0][rawkey0 >> 25]
	   | key_perm_maskr[1][(rawkey0 >> 17) & 0x7f]
	   | key_perm_maskr[2][(rawkey0 >> 9) & 0x7f]
	   | key_perm_maskr[3][(rawkey0 >> 1) & 0x7f]
	   | key_perm_maskr[4][rawkey1 >> 25]
	   | key_perm_maskr[5][(rawkey1 >> 17) & 0x7f]
	   | key_perm_maskr[6][(rawkey1 >> 9) & 0x7f]
	   | key_perm_maskr[7][(rawkey1 >> 1) & 0x7f];
	/*
	 * Rotate subkeys and do compression permutation.
	 */
	shifts = 0;
	for (round = 0; round < 16; round++) {
		uint32_t t0, t1;

		shifts += key_shifts[round];

		t0 = (k0 << shifts) | (k0 >> (28 - shifts));
		t1 = (k1 << shifts) | (k1 >> (28 - shifts));

		en_keysl[round] = comp_maskl[0][(t0 >> 21) & 0x7f]
				| comp_maskl[1][(t0 >> 14) & 0x7f]
				| comp_maskl[2][(t0 >> 7) & 0x7f]
				| comp_maskl[3][t0 & 0x7f]
				| comp_maskl[4][(t1 >> 21) & 0x7f]
				| comp_maskl[5][(t1 >> 14) & 0x7f]
				| comp_maskl[6][(t1 >> 7) & 0x7f]
				| comp_maskl[7][t1 & 0x7f];

		en_keysr[round] = comp_maskr[0][(t0 >> 21) & 0x7f]
				| comp_maskr[1][(t0 >> 14) & 0x7f]
				| comp_maskr[2][(t0 >> 7) & 0x7f]
				| comp_maskr[3][t0 & 0x7f]
				| comp_maskr[4][(t1 >> 21) & 0x7f]
				| comp_maskr[5][(t1 >> 14) & 0x7f]
				| comp_maskr[6][(t1 >> 7) & 0x7f]
				| comp_maskr[7][t1 & 0x7f];
	}
}


static void do_des(struct des_ctx *ctx, uint32_t *l_out, uint32_t *r_out, int count)
{
	uint32_t l, r, *kl, *kr;
	uint32_t f;
	uint32_t r48l, r48r;
	int round;

	/* Do initial permutation (IP). */
	l = r = 0;

	do {
		/* Do each round. */
		kl = en_keysl;
		kr = en_keysr;
		round = 16;
		do {
			/* Expand R to 48 bits (simulate the E-box). */
			r48l	= ((r & 0x00000001) << 23)
				| ((r & 0xf8000000) >> 9)
				| ((r & 0x1f800000) >> 11)
				| ((r & 0x01f80000) >> 13)
				| ((r & 0x001f8000) >> 15);

			r48r	= ((r & 0x0001f800) << 7)
				| ((r & 0x00001f80) << 5)
				| ((r & 0x000001f8) << 3)
				| ((r & 0x0000001f) << 1)
				| ((r & 0x80000000) >> 31);
			/*
			 * Do salting for crypt() and friends, and
			 * XOR with the permuted key.
			 */
			f = (r48l ^ r48r) & saltbits;
			r48l ^= f ^ *kl++;
			r48r ^= f ^ *kr++;
			/*
			 * Do sbox lookups (which shrink it back to 32 bits)
			 * and do the pbox permutation at the same time.
			 */
			f = psbox[0][m_sbox[0][r48l >> 12]]
			  | psbox[1][m_sbox[1][r48l & 0xfff]]
			  | psbox[2][m_sbox[2][r48r >> 12]]
			  | psbox[3][m_sbox[3][r48r & 0xfff]];
			/* Now that we've permuted things, complete f(). */
			f ^= l;
			l = r;
			r = f;
		} while (--round);
		r = l;
		l = f;
	} while (--count);

	/* Do final permutation (inverse of IP). */
	*l_out	= fp_maskl[0][l >> 24]
		| fp_maskl[1][(l >> 16) & 0xff]
		| fp_maskl[2][(l >> 8) & 0xff]
		| fp_maskl[3][l & 0xff]
		| fp_maskl[4][r >> 24]
		| fp_maskl[5][(r >> 16) & 0xff]
		| fp_maskl[6][(r >> 8) & 0xff]
		| fp_maskl[7][r & 0xff];
	*r_out	= fp_maskr[0][l >> 24]
		| fp_maskr[1][(l >> 16) & 0xff]
		| fp_maskr[2][(l >> 8) & 0xff]
		| fp_maskr[3][l & 0xff]
		| fp_maskr[4][r >> 24]
		| fp_maskr[5][(r >> 16) & 0xff]
		| fp_maskr[6][(r >> 8) & 0xff]
		| fp_maskr[7][r & 0xff];
}

#define DES_OUT_BUFSIZE 21

static void
to64_msb_first(char *s, unsigned v)
{
	*s++ = i64c(v >> 18); /* bits 23..18 */
	*s++ = i64c(v >> 12); /* bits 17..12 */
	*s++ = i64c(v >> 6); /* bits 11..6 */
	*s   = i64c(v); /* bits 5..0 */
}

static char *
des_crypt(struct des_ctx *ctx, char output[DES_OUT_BUFSIZE],
		const unsigned char *key, const unsigned char *setting)
{
	uint32_t salt, r0, r1, keybuf[2];
	uint8_t *q;

	/*
	 * Copy the key, shifting each character up by one bit
	 * and padding with zeros.
	 */
	q = (uint8_t *)keybuf;
	while (q - (uint8_t *)keybuf != 8) {
		*q = *key << 1;
		if (*q)
			key++;
		q++;
	}
	des_setkey(ctx, (void *)keybuf);

	/*
	 * setting - 2 bytes of salt
	 * key - up to 8 characters
	 */
	salt = (ascii_to_bin(setting[1]) << 6)
	     |  ascii_to_bin(setting[0]);

	output[0] = setting[0];
	/*
	 * If the encrypted password that the salt was extracted from
	 * is only 1 character long, the salt will be corrupted.  We
	 * need to ensure that the output string doesn't have an extra
	 * NUL in it!
	 */
	output[1] = setting[1] ? setting[1] : output[0];

	setup_salt(ctx, salt);
	/* Do it. */
	do_des(ctx, /*0, 0,*/ &r0, &r1, 25 /* count */);

	/* Now encode the result. */
	/* Each call takes low-order 24 bits and stores 4 chars */
	/* bits 31..8 of r0 */
	to64_msb_first(output + 2, (r0 >> 8));
	/* bits 7..0 of r0 and 31..16 of r1 */
	to64_msb_first(output + 6, (r0 << 16) | (r1 >> 16));
	/* bits 15..0 of r1 and two zero bits (plus extra zero byte) */
	to64_msb_first(output + 10, (r1 << 8));
	/* extra zero byte is encoded as '.', fixing it */
	output[13] = '\0';

	return output;
}

#undef C
#undef init_perm
#undef final_perm
#undef m_sbox
#undef D
#undef const_ctx
#undef saltbits
#undef old_salt
#undef old_rawkey0
#undef old_rawkey1
#undef un_pbox
#undef inv_comp_perm
#undef inv_key_perm
#undef en_keysl
#undef en_keysr
#undef de_keysl
#undef de_keysr
#undef ip_maskl
#undef ip_maskr
#undef fp_maskl
#undef fp_maskr
#undef key_perm_maskl
#undef key_perm_maskr
#undef comp_maskl
#undef comp_maskr
#undef psbox

struct crypt_data;

char *__crypt_r(const char *clear, const char *salt, struct crypt_data *data)
{
	struct des_ctx des_ctx = { 0 };

#if 0
	/* MD5 or SHA? */
	if (salt[0] == '$' && salt[1] && salt[2] == '$') {
		if (salt[1] == '1')
			return md5_crypt((char *)data, clear, salt);
	}
#endif

	des_init(&des_ctx);
	return des_crypt(&des_ctx, (char *)data, (void *)clear, (void *)salt);
}

weak_alias(__crypt_r, crypt_r);

char *crypt(const char *clear, const char *salt)
{
	static char buf[128];
	return __crypt_r(clear, salt, (void *)buf);
}
